/*
   (c) Copyright 2000  convergence integrated media GmbH.
   All rights reserved.

   Written by Denis Oliver Kropp <dok@convergence.de> and
              Andreas Hundt <andi@convergence.de>.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include <directfb.h>

#include <core/coredefs.h>
#include <core/coretypes.h>

#include <core/state.h>
#include <core/gfxcard.h>
#include <core/surfaces.h>
#include <core/fbdev/fbdev.h>

#include <gfx/convert.h>

extern GraphicsDevice *emma_device;
extern FBDev *dfb_fbdev; 
#include "emma.h" 

static unsigned int format_conv[4][4] = { // format-3 , 8bpp, 16bpp , 32bpp , 15bpp
   	 { 0xff,0x03,0x04,0xff},   // 8bpp source
   	 { 0x00,0xff,0x08,0x07},   // 16bpp source
   	 { 0x00,0x0a,0xff,0x09},   // 32bpp source
   	 { 0x00,0x05,0x06,0xff}    // 15bpp source
};
static int sformat=4;
static int dformat=4;

#if 0
int is_field_display_phys(unsigned int adr)
{
    static unsigned int fb_start = 0;
    static unsigned int fb_end;
    static unsigned int fb_half;
    static unsigned int line_len;
    static unsigned int two_line_len;
    static unsigned int field_len;
    int inbottom;

    if(!dfb_fbdev->shared->modes->laced)
	return 0;
    if(!fb_start) {
	int bpp;
	bpp = ((dfb_fbdev->shared->current_var.bits_per_pixel+7)/8);
	fb_start = dfb_gfxcard_memory_physical(emma_device,0) & 0x3ffffff;
	line_len = dfb_fbdev->shared->current_var.xres * bpp;
	fb_end = fb_start
	    + dfb_fbdev->shared->current_var.yres_virtual
	    * line_len;

	two_line_len = line_len * 2;
	field_len = (dfb_fbdev->shared->current_var.yres/2)
	    * line_len;

	fb_half = fb_start + field_len;
    }
    if(adr>=fb_start && adr<fb_end) {
	int x,y;
	return dfb_fbdev->shared->current_var.yres/2;
    }
    return 0;
}
#endif

inline void emma_set_destination(
       EMMADriverData *drv,
       EMMADeviceData *emma,
       CardState *state
	)
{
     int format;
     int lformat;
     int shift=0;
     CoreSurface *dest;
     dest = state->destination;

     switch (state->destination->format) {
          case DSPF_LUT8:		
	       format = 3;
	       lformat = 2;
	       shift=0;
               break;          
          case DSPF_RGB15:		
	       format = 6;
	       lformat = 2;
	       shift=1;
               break;          
          case DSPF_RGB16:
	       format = 4;
	       lformat = 2;
	       shift=1;
               break;
          case DSPF_RGB24:
               break;
          case DSPF_RGB32:
	       format = 5;
	       lformat = 4;
	       shift=2;
          case DSPF_ARGB:
	       format = 5;
	       lformat = 4;
	       shift=2;

               break;
          default:
               BUG( "unexpected pixelformat!" );
               break;
     }
     dformat = format;
     mmio_copy->GBLT_MODE &= ~(7<<8);
     mmio_copy->GBLT_MODE |= format<<8; 
     mmio_copy->GBLT_MODE &= ~(0x0f<<24);
     if(sformat >= 3 && dformat >= 3) {
	 mmio_copy->GBLT_MODE &= ~(0x0f<<24);
	 mmio_copy->GBLT_MODE |= format_conv[sformat-3][dformat-3] << 24;
     }
     mmio_copy->GBLT_LDRW_RL &= ~(7<<2);
     mmio_copy->GBLT_LDRW_RL |= lformat << 2;
     mmio_copy->GBLT_DST0TOP = state->destination->back_buffer->video.offset + FBMEM_OFFSET;
     if(state->destination->caps & DSCAPS_SEPARATED)
	 emma->dest_fields = state->destination->height/2;
     else 
	 emma->dest_fields = 0; 
     emma->dest_pitch = state->destination->back_buffer->video.pitch;
     if(emma->dest_fields) {
	 mmio_copy->GBLT_DST0BOT = mmio_copy->GBLT_DST0TOP +
	     	emma->dest_fields * state->destination->back_buffer->video.pitch;
     } else {
	 mmio_copy->GBLT_DST0BOT = mmio_copy->GBLT_DST0TOP +
	     	state->destination->back_buffer->video.pitch;
	 mmio_copy->GBLT_DST_PITCH <<= 1;
	 emma->dest_pitch <<=1;
     }
     mmio_copy->GBLT_DST_PITCH = emma->dest_pitch >> shift;
     emma->system_dst0top = 
	dfb_gfxcard_memory_virtual(emma_device,0) + mmio_copy->GBLT_DST0TOP - FBMEM_OFFSET;
     emma->system_dst0bot = 
	dfb_gfxcard_memory_virtual(emma_device,0) + mmio_copy->GBLT_DST0BOT - FBMEM_OFFSET;

     state->modified &= ~SMF_DESTINATION;
}


inline void emma_set_source(
       EMMADriverData *drv,
       EMMADeviceData *emma,
       CardState *state
	)
{
    int format;
    int shift;
     emma_waitfifo( mmio_copy, 3);
     
     switch (state->source->format) {
          case DSPF_LUT8:		
	       format = 3;
	       shift=0;
               break;          
          case DSPF_RGB15:		
	       format = 6;
	       shift=1;
               break;          
          case DSPF_RGB16:
	       format = 4;
	       shift=1;
               break;
          case DSPF_RGB24:
               break;
          case DSPF_RGB32:
	       format = 5;
	       shift=2;
	       break;
          case DSPF_ARGB:
	       format = 5;
	       shift=2;
               break;
          default:
               BUG( "unexpected pixelformat!" );
               break;
     }
     sformat = format;
     mmio_copy->GBLT_MODE &= ~(7<<4);
     mmio_copy->GBLT_MODE |= format<<4; 
     if(sformat >= 3 && dformat >= 3) {
	 mmio_copy->GBLT_MODE &= ~(0x0f<<24);
	 mmio_copy->GBLT_MODE |= format_conv[sformat-3][dformat-3] << 24;
     }
     
     mmio_copy->GBLT_SRC0TOP =   state->source->front_buffer->video.offset + FBMEM_OFFSET;
     emma->source_pitch = state->source->front_buffer->video.pitch;
     if(state->source->caps & DSCAPS_SEPARATED)
	 emma->source_fields = state->source->height/2;
     else 
	 emma->source_fields = 0; 
     if(emma->source_fields) {
	 mmio_copy->GBLT_SRC0BOT = mmio_copy->GBLT_SRC0TOP +
	     	emma->source_fields * state->source->front_buffer->video.pitch;
     } else {
	 mmio_copy->GBLT_SRC0BOT = mmio_copy->GBLT_SRC0TOP +
	     	state->source->front_buffer->video.pitch;
	 emma->source_pitch <<=1;
     }
     // fprintf(stderr,"emma->source_pitch=%d shift=%d\n",emma->source_pitch,shift);
     mmio_copy->GBLT_SRC_PITCH = emma->source_pitch>>shift;
     emma->system_src0top = 
	dfb_gfxcard_memory_virtual(emma_device,0) + mmio_copy->GBLT_SRC0TOP - FBMEM_OFFSET;
     emma->system_src0bot = 
	dfb_gfxcard_memory_virtual(emma_device,0) + mmio_copy->GBLT_SRC0BOT - FBMEM_OFFSET;

     state->modified &= ~SMF_SOURCE;
}

inline void emma_set_clip(
       EMMADriverData *drv,
       EMMADeviceData *emma,
       CardState *state
	)
{     
     emma_waitfifo( mmio_copy, 2 );

      //   state->clip.y1 << 16) | (ati->state->clip.x1*3) );

       //     (state->clip.y2 << 16) | ((ati->state->clip.x2*3) + 3));

     state->modified &= ~SMF_CLIP;
}

inline void emma_set_color(
       EMMADriverData *drv,
       EMMADeviceData *emma,
       CardState *state
	)
{
     __u32 fill_color = 0; 
     
     switch (state->destination->format) {
          case DSPF_RGB15:
               fill_color = PIXEL_RGB15( state->color.r,
                                         state->color.g,
                                         state->color.b );
               break;
          case DSPF_RGB16:
               fill_color = PIXEL_RGB16( state->color.r,
                                         state->color.g,
                                         state->color.b );
               break;
          case DSPF_RGB24:
          case DSPF_RGB32:                              
               fill_color = PIXEL_RGB32( state->color.r,
                                         state->color.g,
                                         state->color.b );
               break;
          case DSPF_ARGB:
               fill_color = PIXEL_ARGB( state->color.a,
                                        state->color.r,
                                        state->color.g,
                                        state->color.b );
               break;
          default:
               BUG( "unexpected pixelformat!" );
               break;
     }
     mmio_copy->GBLT_FONT0 = fill_color;
     
     state->modified &= ~SMF_COLOR;
}

inline void emma_set_blittingflags(
       EMMADriverData *drv,
       EMMADeviceData *emma,
       CardState *state
	)
{     
     state->modified &= ~SMF_BLITTING_FLAGS;
}

inline void emma_set_blending_function(
       EMMADriverData *drv,
       EMMADeviceData *emma,
       CardState *state
	)
{     
      //                emmaSourceBlend[ati->state->src_blend - 1] |
      //                emmaDestBlend  [ati->state->dst_blend - 1] |
     
     state->modified &= ~(SMF_DST_BLEND | SMF_SRC_BLEND);
}

