/*
   (c) Copyright 2000  convergence integrated media GmbH.
   All rights reserved.

   Written by Denis Oliver Kropp <dok@convergence.de> and
              Andreas Hundt <andi@convergence.de>.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/
// #define DEBUG
#define BLITLOG stderr
// #define BLITCHECK
#include <asm/types.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/mman.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <directfb.h>

#include <core/coredefs.h>
#include <core/coretypes.h>
#include <core/palette.h>

#include <core/state.h>
#include <core/gfxcard.h>
#include <core/surfaces.h>

#include <gfx/convert.h>

#include <core/graphics_driver.h>



#include <mmac/hardware/hwgblt.h>

#include "emma.h"

volatile EMMA_GBLT *mmio_base;
static volatile unsigned int *hw_palette;
EMMA_GBLT blit_regs_copy;
EMMA_GBLT *mmio_copy = &blit_regs_copy;
GraphicsDevice *emma_device;

static int blit_busy=0;
static FILE *blitlog;

#include "printregs.h"

/* driver capability flags */

/*
#define EMMA_SUPPORTED_DRAWINGFLAGS \
               (DSDRAW_NOFX)

#define EMMA_SUPPORTED_DRAWINGFUNCTIONS \
               (DFXL_FILLRECTANGLE | DFXL_DRAWRECTANGLE | DFXL_DRAWLINE)

#define EMMA_SUPPORTED_BLITTINGFLAGS \
               (DSBLIT_SRC_COLORKEY | DSBLIT_BLEND_ALPHACHANNEL)

#define EMMA_SUPPORTED_BLITTINGFUNCTIONS \
               (DFXL_BLIT | DFXL_STRETCHBLIT)
	       */
#define EMMA_SUPPORTED_DRAWINGFLAGS \
               (DSDRAW_NOFX)

#define EMMA_SUPPORTED_DRAWINGFUNCTIONS \
               (/*DFXL_FILLRECTANGLE |  DFXL_DRAWLINE */ 0 )

#define EMMA_SUPPORTED_BLITTINGFLAGS \
               (/*DSBLIT_SRC_COLORKEY |  DSBLIT_BLEND_ALPHACHANNEL */ 0)

#define EMMA_SUPPORTED_BLITTINGFUNCTIONS \
    		(DFXL_BLIT /* | DFXL_STRETCHBLIT*/)


DFB_GRAPHICS_DRIVER( emma )

/** CARD FUNCTIONS **/
static void emmaFillRectangle(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRectangle *rect );
static void emmaFillBlendRectangle(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRectangle *rect );
static void emmaDrawRectangle(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRectangle *rect );
static void emmaDrawBlendRectangle(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRectangle *rect );

static void set_hw_palette( EMMADeviceData *dev )
{
    CorePalette *pal;
    int i;
    switch (dev->state->source->format) {
	 case DSPF_LUT8:
	     pal = dev->state->source->palette;
	     break;
	 default:
	     return;
    }
    switch (dev->state->destination->format) {
	 case DSPF_RGB16:
	     for(i=0;i<pal->num_entries;i++)
		 hw_palette[i] = PIXEL_RGB16( pal->entries[i].r ,
			 		      pal->entries[i].g ,
					      pal->entries[i].b );
	     break;
	 case DSPF_RGB15:
	     for(i=0;i<pal->num_entries;i++)
		 hw_palette[i] = PIXEL_RGB15( pal->entries[i].r ,
			 		      pal->entries[i].g ,
					      pal->entries[i].b );
	     break;
	default:
	     fprintf(stderr,"set_hw_palette dest_format not supported\n");
	     break;
    }
}

inline void emma_blit(int cmd)
{
    mmio_base->GBLT_CMD = cmd;
    blit_busy = 1;
}
#define USLEEP_TIMEOUT 100000

inline void emma_waitfifo(volatile EMMA_GBLT *base,int fifo)
{
    int timeout=0;
    int i;
    if(!blit_busy)
	return;
    while((mmio_base->GBLT_INT_READ & 1) == 0) {
	volatile int j;
	for(j=0;j<10000;j++);
	timeout++;
	// if(timeout>USLEEP_TIMEOUT)
	  //  usleep(10000);
	if(timeout>USLEEP_TIMEOUT+100) {
	    fprintf(BLITLOG,"timed out\n"); // 1 second
	    break;
	}
    }
    

    if( timeout>USLEEP_TIMEOUT+100) {
	fprintf(BLITLOG,"timeout = %d\n",timeout); // DEBUG_JAN
	GBLT_print(BLITLOG,mmio_base);
	mmio_base->GBLT_IN_LOAD2 = 1;
	usleep(10000);
	mmio_base->GBLT_IN_LOAD2 = 0;
	sleep(1);
    }
    blit_busy = 0;
    base->GBLT_INT_READ = 0;


}

inline void emma_waitidle(volatile EMMA_GBLT *base)
{
    if(!blit_busy)
	return;
    emma_waitfifo(base,0);
}

/* required implementations */

static void emmaEngineSync()
{
     emma_waitidle( mmio_base );
}

static void emmaCheckState(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       CardState *state,
      	DFBAccelerationMask accel )
{
     switch (state->destination->format) {
          case DSPF_LUT8:
          case DSPF_RGB15:
          case DSPF_RGB16:
          case DSPF_RGB32:
          case DSPF_ARGB:
               break;
          default:
	       fprintf(BLITLOG,"bad dest format\n"); 
               return;
     }
     
     /* check for the special drawing function that does not support
        the usually supported drawingflags */
     if (accel == DFXL_DRAWLINE  &&  state->drawingflags != DSDRAW_NOFX) {
	 fprintf(BLITLOG,"bad draw line\n");
          return;
     }

     if(accel == DFXL_STRETCHBLIT &&
	(state->blittingflags ||
	 state->destination->format != state->source->format)) {
	 return;
     }

     
     /* if there are no other drawing flags than the supported */
     if (!(accel & ~EMMA_SUPPORTED_DRAWINGFUNCTIONS) &&
         !(state->drawingflags & ~EMMA_SUPPORTED_DRAWINGFLAGS))
          state->accel |= EMMA_SUPPORTED_DRAWINGFUNCTIONS;

     /* if there are no other blitting flags than the supported
        and the source has the minimum size */
     // fprintf(BLITLOG,"accel & ~emma = %x\n",accel & ~EMMA_SUPPORTED_BLITTINGFUNCTIONS);
     // fprintf(BLITLOG,"bflags & ~emma = %x\n",state->blittingflags & ~EMMA_SUPPORTED_BLITTINGFLAGS);

     if (!(accel & ~EMMA_SUPPORTED_BLITTINGFUNCTIONS) &&
         !(state->blittingflags & ~EMMA_SUPPORTED_BLITTINGFLAGS)  &&
         state->source  &&
         state->source->front_buffer != state->destination->back_buffer) 
     {
	 // fprintf(BLITLOG,"back_buffer = %x\n",state->destination->back_buffer);
          switch (state->source->format) {
               case DSPF_RGB15:
               case DSPF_RGB16:
               case DSPF_RGB32:
               case DSPF_ARGB:
               case DSPF_LUT8:
                    state->accel |= EMMA_SUPPORTED_BLITTINGFUNCTIONS;
		    break;
               default:
		    fprintf(BLITLOG,"bad source format %x argb=%x\n",
			    state->source->format,DSPF_ARGB);
                    ;
          }
     }
}


static void emmaSetState(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       GraphicsDeviceFuncs *funcs,
       	CardState *state, DFBAccelerationMask accel )
{
     if (state != dev->state) {
          state->modified |= SMF_ALL;
          state->set = 0;
          dev->state = state;
     }

     if ((state->modified & SMF_SOURCE)  &&  state->source)
          emma_set_source(drv,dev,state);


     if (state->modified & SMF_DESTINATION)
          emma_set_destination(drv,dev,state);
     
     if (state->modified & SMF_CLIP)
          emma_set_clip(drv,dev,state);

     if (state->modified & SMF_COLOR)
          emma_set_color(drv,dev,state);
     
     if (state->modified & SMF_SRC_COLORKEY) {
	  mmio_base->GBLT_CHRKEY_MIN = state->src_colorkey;
	  mmio_base->GBLT_CHRKEY_MAX = state->src_colorkey;
     }

     if (state->modified & SMF_BLITTING_FLAGS)
          emma_set_blittingflags(drv,dev,state);

     if (state->modified & (SMF_SRC_BLEND | SMF_DST_BLEND)) 
          emma_set_blending_function(drv,dev,state);

     if (state->modified & SMF_DRAWING_FLAGS) {
          if (state->drawingflags & DSDRAW_BLEND) {
               funcs->FillRectangle = emmaFillBlendRectangle;
               funcs->DrawRectangle = emmaDrawBlendRectangle;
          }
          else {
               funcs->FillRectangle = emmaFillRectangle;
               funcs->DrawRectangle = emmaDrawRectangle;
          }
     }
     /*
     switch (accel)
	{
	case DFXL_BLIT:
          emma_set_source(drv,dev,state);
          emma_set_destination(drv,dev,state);
	  state->set |= DFXL_BLIT;
	  break;
	default:
	  BUG( "unexpected drawing/blitting function!" );
	  break;
      }
      */
}

static void emmaFillRectangle(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRectangle *rect )
{
     static int count=0;
     static int xadd,yadd;
#ifdef DEBUG
     printf("waitfifo\n");
#endif
     emma_waitfifo( mmio_base, 4 );    
#ifdef DEBUG
     fprintf(BLITLOG,"Fill Rectangle %d %d %d %d\n",rect->x,rect->y,
	     rect->x+rect->w,rect->y+rect->h); 
     /* rect: x,y,w,h */
     fprintf(BLITLOG,"dst = 0x%x pitch=%d\n",mmio_copy->GBLT_DST0TOP,
	     		    mmio_copy->GBLT_DST_PITCH);
     fprintf(BLITLOG,"spos= %08x\n",mmio_copy->GBLT_MODE);
#endif
     mmio_base->GBLT_LDRW_RL = mmio_copy->GBLT_LDRW_RL | 3; 
     mmio_base->GBLT_FONT0 = mmio_copy->GBLT_FONT0;
     mmio_base->GBLT_MODE = 0;
     mmio_base->GBLT_DST0TOP = mmio_copy->GBLT_DST0TOP;
     mmio_base->GBLT_DST_PITCH = mmio_copy->GBLT_DST_PITCH;
     mmio_base->GBLT_SRC0TOP = mmio_copy->GBLT_DST0TOP;
     mmio_base->GBLT_SRC_PITCH = mmio_copy->GBLT_DST_PITCH;
     mmio_base->GBLT_LDRW_SPOS = (rect->x+1) | ((rect->y+1)<<16);
     if(rect->x+rect->w >= mmio_copy->GBLT_DST_PITCH) 
         mmio_base->GBLT_LDRW_EPOS = (mmio_copy->GBLT_DST_PITCH-16) | ((rect->y+rect->h-1)<<16);
     else
     	 mmio_base->GBLT_LDRW_EPOS = (rect->x+rect->w-1) | ((rect->y+rect->h-1)<<16);
     emma_blit(5);  // Line Draw

}

static void emmaFillRectangle2( DFBRectangle *rect )
{
    DFBRectangle r;
    int x,y;
    r.w=32;
    r.h=32;
    for(y=rect->y;y+32<rect->y+rect->h;y+=32) {
    	for(x=rect->x;x+32<rect->x+rect->w;x+=32) {
	    r.x=x;
	    r.y=y;
	    if(y>31)
	    	emmaFillRectangle2(&r);
	}
    }
}

static void emmaFillBlendRectangle(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRectangle *rect )
{
}

static void emmaDrawRectangle(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRectangle *rect )
{
     emma_waitfifo( mmio_base, 2 );
}

static void emmaDrawBlendRectangle(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRectangle *rect )
{
}


static void emmaDrawLine(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRegion *line )
{
#ifdef DEBUG
     fprintf(BLITLOG,"waitfifo\n");
#endif
     emma_waitfifo( mmio_base, 4 );    
#ifdef DEBUG
     fprintf(BLITLOG,"Draw Line %d %d %d %d\n",line->x1,line->y1,
	     line->x2,line->y2); 
     /* rect: x,y,w,h */
     fprintf(BLITLOG,"dst = 0x%x pitch=%d\n",mmio_copy->GBLT_DST0TOP,
	     		    mmio_copy->GBLT_DST_PITCH);
     fprintf(BLITLOG,"mode= %08x\n",mmio_copy->GBLT_MODE);
#endif
     mmio_base->GBLT_LDRW_RL = mmio_copy->GBLT_LDRW_RL; 
     mmio_base->GBLT_FONT0 = mmio_copy->GBLT_FONT0;
     mmio_base->GBLT_MODE = 0;
     mmio_base->GBLT_DST0TOP = mmio_copy->GBLT_DST0TOP;
     mmio_base->GBLT_DST_PITCH = mmio_copy->GBLT_DST_PITCH;
     mmio_base->GBLT_SRC0TOP = mmio_copy->GBLT_DST0TOP;
     mmio_base->GBLT_SRC_PITCH = mmio_copy->GBLT_DST_PITCH;
     mmio_base->GBLT_LDRW_SPOS = (line->x1) | ((line->y1)<<16);
     mmio_base->GBLT_LDRW_EPOS = (line->x2) | ((line->y2)<<16);
     emma_blit(5);  // Line Draw

}

static int get_scale_reg(int a,int b) // a/b
{
    int v;
    printf("a/b=%f\n",(float)a/(float)b);
    if(a==b) {
	return 1024;
    }
    if(a==2*b) { // 1/2
	return 513;
    }
    if(2*a==b) { // 2
	return 511;
    }
    if(a<b) {
	v = (a*1024)/b;
	if(v<=512) 
	    return 0;
	return v;
    } else {
	v = (b*1024)/a;
	if(v<=512) 
	    return 0;
	return 1024 | v;
    }
    return 0;
}


static void emmaStretchBlit(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       	DFBRectangle *sr, DFBRectangle *dr )
{
     int scalex,scaley;
     printf("stretchBlit\n");
     scalex = get_scale_reg(dr->w,sr->w);
     scaley = get_scale_reg(dr->h,sr->h);
     printf("scalex=%d scaley=%d\n",scalex,scaley);
     if(!scalex || !scaley) {
	 printf("invalid scaling factor\n");
         scalex = 513;
         scaley = 513;
     }
     
     emma_waitfifo( mmio_base, 9 );
     mmio_base->GBLT_IN_LOAD1 = 1;
     usleep(10000);
     mmio_base->GBLT_IN_LOAD1 = 0;
     usleep(10000);
     
     switch (dev->state->source->format) {
          case DSPF_RGB15:
	       break;
          

          case DSPF_RGB16:

               break;
          case DSPF_RGB32:          

               break;
          case DSPF_ARGB:
               break;
          default:
               BUG( "unexpected pixelformat!" );
               return;
     }
     emma_waitfifo( mmio_base, 9 );
#ifdef DEBUG
     fprintf(BLITLOG,"emmaBlit dx=%d dy=%x sx=%d sy=%d w=%d h=%d\n",dr->x,dr->y,
	     sr->x,sr->y,sr->w,sr->h);
#endif
     mmio_base->GBLT_MODE = mmio_copy->GBLT_MODE | (1<<12); // | src_1d | dst_1d;
     mmio_base->GBLT_DST0TOP = mmio_copy->GBLT_DST0TOP>>3;
     mmio_base->GBLT_SRC0TOP = mmio_copy->GBLT_SRC0TOP>>3;
     mmio_base->GBLT_DST0BOT = mmio_copy->GBLT_DST0TOP>>3;
     mmio_base->GBLT_SRC0BOT = mmio_copy->GBLT_SRC0TOP>>3;
     mmio_base->GBLT_DST_PITCH = mmio_copy->GBLT_DST_PITCH;
     mmio_base->GBLT_SRC_PITCH = mmio_copy->GBLT_SRC_PITCH;
     mmio_base->GBLT_SRC_SPOS = (sr->x) | ((sr->y)<<16);
     mmio_base->GBLT_DST_SPOS = (dr->x) | ((dr->y)<<16);
     mmio_base->GBLT_BLT_SIZE = (sr->w) | ((sr->h)<<16);

     mmio_base->GBLT_MODE = mmio_copy->GBLT_MODE | (1<<12); // | src_1d | dst_1d;
     emma_blit(8);
     emma_waitfifo(mmio_base,0);   
     mmio_base->GBLT_MODE = mmio_copy->GBLT_MODE; // | src_1d | dst_1d;
     mmio_base->GBLT_DST0TOP = mmio_copy->GBLT_DST0TOP;
     mmio_base->GBLT_SRC0TOP = mmio_copy->GBLT_SRC0TOP;
     mmio_base->GBLT_SCLR_BASEW = scalex | (scaley<<16) | (3<<30);
     emma_blit(3); // strech blit 
}


// #define DEBUG
static void emmaBlit(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       DFBRectangle *rect, int dx, int dy , int field )
{
    int src_1d=0,dst_1d=0;
    int sx,sy,w,h;
    int field_blit = 1;
    int i;
    int areas = 0;
     __u32 dir_cmd = 0;
    sx = rect->x;
    sy = rect->y;
    w = rect->w;
    h = rect->h;
     if(rect->h<2) {
	 fprintf(BLITLOG,"emmaBlit:height too small\n");
	 return;
     }
     if(rect->w<2) {
	 fprintf(BLITLOG,"emmaBlit:width too small\n");
	 return;
     }
     emma_waitfifo( mmio_base, 9 );
     // sleep(1);
#ifdef DEBUG
     fprintf(BLITLOG,"field=%d\n",field);
	     
     fprintf(BLITLOG,"emmaBlit dx=%d dy=%x sx=%d sy=%d w=%d h=%d sf=%d df=%d\n",dx,dy,
		 rect->x,rect->y,rect->w,rect->h,
		 dev->source_fields,dev->dest_fields);
     fprintf(stderr,"GBLT_MODE=%08x\n",mmio_copy->GBLT_MODE);
     fprintf(stderr,"SRC0TOP=%08x\n",mmio_copy->GBLT_SRC0TOP);
     fprintf(stderr,"SRC0BOT=%08x\n",mmio_copy->GBLT_SRC0BOT);
     fprintf(stderr,"DST0TOP=%08x\n",mmio_copy->GBLT_DST0TOP);
     fprintf(stderr,"DST0BOT=%08x\n",mmio_copy->GBLT_DST0BOT);
     fprintf(stderr,"SRC_PITCH=%d\n",mmio_copy->GBLT_SRC_PITCH);
     fprintf(stderr,"DST_PITCH=%d\n",mmio_copy->GBLT_DST_PITCH);

#endif
     if(!dev->dest_fields && !dev->source_fields) {
	 field_blit = 0;
	 if(rect->x == 0 && rect->w==mmio_copy->GBLT_SRC_PITCH/2) {
	     src_1d = 1<<14;
#ifdef DEBUG
	     fprintf(stderr,"src 1d\n");
#endif
	 }
	 if(dx == 0 && rect->w==mmio_copy->GBLT_DST_PITCH/2) {
	     dst_1d = 1<<15;
#ifdef DEBUG
	     fprintf(stderr,"dst 1d\n");
#endif
	 }
     } else {
	 if(rect->x == 0 && rect->w==mmio_copy->GBLT_SRC_PITCH) {
	     src_1d = 1<<14;
#ifdef DEBUG
	     fprintf(stderr,"src 1d\n");
#endif
	 }
	 if(dx == 0 && rect->w==mmio_copy->GBLT_DST_PITCH) {
	     dst_1d = 1<<15;
#ifdef DEBUG
	     fprintf(stderr,"dst 1d\n");
#endif
	 }
	 if(field == 3) 
	     areas = 1<<12;
	 else
	     areas = 0;
     }
     mmio_base->GBLT_MODE = mmio_copy->GBLT_MODE & 0xff000000;
     mmio_base->GBLT_MODE = mmio_copy->GBLT_MODE & 0xff000ff0;
     mmio_base->GBLT_MODE = (mmio_copy->GBLT_MODE & 0xff000ff0) | areas;
     mmio_base->GBLT_MODE = (mmio_copy->GBLT_MODE & 0xff000ff0) | areas | src_1d | dst_1d;

     if(!(dy & 1) || !field_blit
#ifdef BLITCHECK
	     || 1
#endif
	     ) {
	 if(field == 1)
	     mmio_base->GBLT_DST0TOP = (mmio_copy->GBLT_DST0BOT)>>3;
	 else {
	     mmio_base->GBLT_DST0BOT = (mmio_copy->GBLT_DST0BOT)>>3;
	     mmio_base->GBLT_DST1TOP = (mmio_copy->GBLT_DST0BOT)>>3;
	     mmio_base->GBLT_DST0TOP = (mmio_copy->GBLT_DST0TOP)>>3;
	 }
     } else {
	 mmio_base->GBLT_DST0TOP = (mmio_copy->GBLT_DST0BOT)>>3;
	 mmio_base->GBLT_DST0BOT = (mmio_copy->GBLT_DST0TOP+dev->dest_pitch)
	     				>>3;
     }
     if(!(rect->y & 1) || !field_blit
#ifdef BLITCHECK
	     || 1
#endif
	     ) {
	 if(field == 1)
	     mmio_base->GBLT_SRC0TOP = (mmio_copy->GBLT_SRC0BOT)>>3;
	 else {
	     mmio_base->GBLT_SRC1TOP = (mmio_copy->GBLT_SRC0BOT)>>3;
	     mmio_base->GBLT_SRC0BOT = (mmio_copy->GBLT_SRC0BOT)>>3;
	     mmio_base->GBLT_SRC0TOP = (mmio_copy->GBLT_SRC0TOP)>>3;
	 }
     } else {
	 mmio_base->GBLT_SRC0TOP = (mmio_copy->GBLT_SRC0BOT)>>3;
	 mmio_base->GBLT_SRC0BOT = (mmio_copy->GBLT_SRC0TOP
		 		+dev->source_pitch)>>3;
     }
     if(field_blit) {
	 sy>>=1;
	 h>>=1;
	 dy>>=1;
	 mmio_base->GBLT_DST_PITCH = mmio_copy->GBLT_DST_PITCH;
	 mmio_base->GBLT_SRC_PITCH = mmio_copy->GBLT_SRC_PITCH;
     } else {
	 mmio_base->GBLT_DST_PITCH = mmio_copy->GBLT_DST_PITCH >> 1;
	 mmio_base->GBLT_SRC_PITCH = mmio_copy->GBLT_SRC_PITCH >> 1;
     }
     mmio_base->GBLT_SRC_SPOS = (sx) | (sy<<16);
     mmio_base->GBLT_DST_SPOS = (dx) | (dy<<16);
     mmio_base->GBLT_BLT_SIZE = (w) | (h<<16);
     if(rect->h & 1) {
	 fprintf(BLITLOG,"emmaBlit: odd height\n");
     }
     dev->estimated_usecs = (w * h)/20;
     // memset(emma->framebuffer.base,0,emma->framebuffer.length);
     // memset(emma->framebuffer.base+mmio_copy->GBLT_SRC0TOP-0x1000,0x47,640*2);
     if (dev->state->source->format != dev->state->destination->format)  {
	 mmio_base->GBLT_ALPHA = 0xff;
	 if(field != 1)
	     set_hw_palette(dev);
	 // sleep(1);
	 emma_blit(1); // data formatter
     } else {
        if(dev->state->blittingflags == DSBLIT_SRC_COLORKEY)
	    emma_blit(7); // chroma keying
	else {
      	    emma_blit(8); // copy 
	    // soft_blit();
	    // gBlit(rect,dx,dy);
	}

#ifdef BLITCHECK
	if(mmio_copy->GBLT_SRC0TOP != mmio_copy->GBLT_DST0TOP) {
	{ int i,j;
	    static int xff = 0xff;
	  unsigned short *stop,*sbot,*dtop,*dbot;
	  stop = dev->system_src0top + sy*(dev->source_pitch>>1);
	  sbot = dev->system_src0bot + sy*(dev->source_pitch>>1);
	  dtop = dev->system_dst0top + dy*(dev->dest_pitch>>1);
	  dbot = dev->system_dst0bot + dy*(dev->dest_pitch>>1);

	  for(i=h-1;i<h;i++) {
	      for(j=0;j<w;j++) {
		  if(dtop[j+dx] != stop[j+sx]
	            || dbot[j+dx] != sbot[j+sx]) {
		      fprintf(BLITLOG,"blit failed i=%d j=%d\n",
			      i,j);
		      return;
		  }
	      }
	      stop+= dev->source_pitch>>1;
	      sbot+= dev->source_pitch>>1;
	      dtop+= dev->dest_pitch>>1;
	      dbot+= dev->dest_pitch>>1;
	  }
	}
	}
#endif
#ifdef DEBUG
	// sleep(10);
#endif
     }

          
}
static void fieldsFillRectangle(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       DFBRectangle *rect)
{
#ifdef DEBUG
    fprintf(BLITLOG,"fieldsFillRectangle %d %d %d %d \n",
	    rect->x,rect->y,rect->w,rect->h);
    fprintf(BLITLOG,"source_fields=%d dest_fields=%d\n",
	    dev->source_fields,dev->dest_fields);
#endif
    if(!dev->dest_fields) {
	emmaFillRectangle(drv,dev,rect);
	return;
    }
    {
	DFBRectangle t_rect,b_rect;
	if(rect->y & 1) {
	    t_rect = *rect;
	    t_rect.y = rect->y/2 + dev->dest_fields;
	    t_rect.h /= rect->y/2;
	    emmaFillRectangle(drv,dev,&t_rect);
	    b_rect = *rect;
	    b_rect.y = rect->y/2+1;
	    b_rect.h = rect->h / 2;
	    emmaFillRectangle(drv,dev,&b_rect);
	} else {
	    t_rect = *rect;
	    t_rect.y = rect->y/2+1 + dev->dest_fields;
	    t_rect.h = rect->h / 2;
	    emmaFillRectangle(drv,dev,&t_rect);
	    b_rect = *rect;
	    b_rect.y = rect->y/2;
	    b_rect.h = rect->h / 2;
	    emmaFillRectangle(drv,dev,&b_rect);
	}
    }
}

static bool overlapBlit(
       EMMADriverData *drv,
       EMMADeviceData *dev,
       DFBRectangle *rect, int dx, int dy )
{
    int xdir = 1;
    int ydir = 1;
    int xdiff;
    int ydiff;
    DFBRectangle sr;
#ifdef DEBUG
    fprintf(BLITLOG,"overlapBlit %d %d %d %d -> %d %d\n",
	    rect->x,rect->y,rect->w,rect->h,dx,dy);
    fprintf(BLITLOG,"source_fields=%d dest_fields=%d\n",
	    dev->source_fields,dev->dest_fields);
#endif
     xdiff = dx - rect->x;
     ydiff = dy - rect->y;
#ifdef DEBUG
     if(mmio_copy->GBLT_DST0TOP == mmio_copy->GBLT_SRC0TOP) {
	 fprintf(BLITLOG,"emmaBlit: same buffer xdiff=%d ydiff=%d w=%d h=%d\n",
		     xdiff,ydiff,rect->w,rect->h);
     }
#endif
     if(mmio_copy->GBLT_DST0TOP == mmio_copy->GBLT_SRC0TOP
	&& (xdiff > -rect->w && xdiff<rect->w && 
	  ydiff >= 0 && ydiff<rect->h)  
	     ) {
	 if(ydiff > 0  ) {
	     fprintf(BLITLOG,"ycase\n"); 
	     ydiff &= 0xfffe;
	     dy = dy + rect->h - ydiff;
	     sr = *rect;
	     sr.y = sr.y + sr.h -ydiff;
	     sr.h = ydiff;
	     while(sr.y >= rect->y) {
		 fprintf(BLITLOG,"subblit %d %d %d %d -> %d %d\n",
			 sr.x,sr.y,sr.w,sr.h,dx,dy);
		 emmaBlit(drv,dev,&sr,dx,dy,3);
		 sr.y -= ydiff;
		 dy -= ydiff;
	     }
	     sr.y += ydiff;
	     dy += ydiff;
	     if(sr.y>rect->y) {
		 dy += sr.y - rect->y;
	         sr.h = sr.y - rect->y+1;	 
		 sr.y = rect->y;
		 if(sr.h>0) {
		     fprintf(BLITLOG,"lsubblit %d %d %d %d -> %d %d\n",
			 sr.x,sr.y,sr.w,sr.h,dx,dy);
		     emmaBlit(drv,dev,&sr,dx,dy,3);
		 }
	     }
	 } else if(xdiff > 0) {
	     fprintf(BLITLOG,"xcase\n"); 
	     ydiff &= 0xfffe;
	     dx = dx + rect->w - xdiff;
	     sr = *rect;
	     sr.x = sr.x + sr.w -xdiff;
	     sr.w = xdiff;
	     while(sr.x >= rect->x) {
		 fprintf(BLITLOG,"subblit %d %d %d %d -> %d %d\n",
			 sr.x,sr.y,sr.w,sr.h,dx,dy);
		 emmaBlit(drv,dev,&sr,dx,dy,3);
		 sr.x -= xdiff;
		 dx -= xdiff;
	     }
	     sr.x += xdiff;
	     dx += xdiff;
	     if(sr.x>rect->x) {
		 dx += sr.x - rect->x;
	         sr.w = sr.x - rect->x;	 
		 sr.x = rect->x;
		 if(sr.w>0) {
		     fprintf(BLITLOG,"lsubblit %d %d %d %d -> %d %d\n",
			 sr.x,sr.y,sr.w,sr.h,dx,dy);
		     emmaBlit(drv,dev,&sr,dx,dy,3);
		 }
	     }
	 } else if(xdiff<0) {
		 int i;
		 sr = *rect;
		 sr.w = -xdiff;
		 while(sr.x-xdiff<= rect->x + rect->w) {
		     emmaBlit(drv,dev,&sr,dx,dy,3);
		     dx -= xdiff;
		     sr.x -= xdiff;
		 }
		 if(sr.x<rect->x + rect->w) {
		     sr.w = rect->w - sr.x;
		     emmaBlit(drv,dev,&sr,dx,dy,3);
		 }
	  } else {
		emmaBlit(drv,dev,rect,dx,dy,3);
	  }
     } else {
	emmaBlit(drv,dev,rect,dx,dy,3);
	usleep(10000);
     }
     return true;
}
/* exported symbols */

static int
driver_get_abi_version()
{
     return DFB_GRAPHICS_DRIVER_ABI_VERSION;
}

static int
driver_probe( GraphicsDevice *device )
{
    return 1; // DEBUG_JAN
}

static void
driver_get_info( GraphicsDevice     *device,
                 GraphicsDriverInfo *info )
{
     /* fill driver info structure */
     snprintf( info->name,
               DFB_GRAPHICS_DRIVER_INFO_NAME_LENGTH,
               "EMMA 2 Driver" );

     snprintf( info->vendor,
               DFB_GRAPHICS_DRIVER_INFO_VENDOR_LENGTH,
               "DiscVision" );

     info->version.major = 0;
     info->version.minor = 1;

     info->driver_data_size = sizeof (EMMADriverData);
     info->device_data_size = sizeof (EMMADeviceData);
}

static DFBResult
driver_init_driver( GraphicsDevice      *device,
                    GraphicsDeviceFuncs *funcs,
                    void                *driver_data )
{
     EMMADriverData *adrv = (EMMADriverData*) driver_data;
     emma_device = device;

     adrv->mmio_base = mmio_base = (volatile EMMA_GBLT*) dfb_gfxcard_map_mmio( device, 0, -1 );
     if (!adrv->mmio_base)
          return DFB_IO;
     hw_palette = (unsigned int *)(((unsigned int) mmio_base) + 0x400);

     funcs->CheckState    = emmaCheckState;
     funcs->SetState      = emmaSetState;
     funcs->EngineSync    = emmaEngineSync;

     funcs->FillRectangle = fieldsFillRectangle;
     funcs->DrawRectangle = emmaDrawRectangle;
     funcs->DrawLine      = emmaDrawLine;
     funcs->Blit          = overlapBlit;
     funcs->StretchBlit   = emmaStretchBlit;
     fprintf(BLITLOG,"emma accel initialised\n");

     return DFB_OK;
}


static DFBResult
driver_init_device( GraphicsDevice     *device,
                    GraphicsDeviceInfo *device_info,
                    void               *driver_data,
                    void               *device_data )
{
     EMMADriverData *adrv = (EMMADriverData*) driver_data;
     EMMADeviceData *adev = (EMMADeviceData*) device_data;
     volatile EMMA_GBLT    *mmio_base = adrv->mmio_base;

     // device_info->info.driver_version.major = 0;
     // device_info->info.driver_version.minor = 0;

     device_info->caps.flags    = 0;
     device_info->caps.accel    = EMMA_SUPPORTED_DRAWINGFUNCTIONS |
                           EMMA_SUPPORTED_BLITTINGFUNCTIONS;
     device_info->caps.drawing  = EMMA_SUPPORTED_DRAWINGFLAGS;
     device_info->caps.blitting = EMMA_SUPPORTED_BLITTINGFLAGS;

     /* fill device info */
     snprintf( device_info->name,
               DFB_GRAPHICS_DEVICE_INFO_NAME_LENGTH, "EMMA" );

     snprintf( device_info->vendor,
               DFB_GRAPHICS_DEVICE_INFO_VENDOR_LENGTH, "NEC" );



     device_info->limits.surface_byteoffset_alignment = 8;
     device_info->limits.surface_pixelpitch_alignment = 8;
     emma_waitfifo( mmio_base, 6 );

 
     // reset hardware
     mmio_base->GBLT_IN_LOAD1 = 1;
     usleep(10000);
     mmio_base->GBLT_IN_LOAD1 = 0;
     usleep(10000);
     mmio_base->GBLT_MODE = 0;
     mmio_base->GBLT_INT_MASK = 1;
     memset(mmio_copy,0,sizeof(*mmio_copy));
     // BLITLOG = fopen("/var/log/blit.log","w");
     BLITLOG = stderr;
     if(!BLITLOG) {
	 perror("open");
	 fprintf(stderr,"can not open blit.log\n");
	 return -1;
     }
     return DFB_OK;
}

static DFBResult
driver_init_layers( void *driver_data,
                    void *device_data )
{
    fprintf(BLITLOG,"driver_init_layers\n"); // DEBUG_JAN
    return DFB_OK;
}

static void
driver_close_device( GraphicsDevice *device,
                     void           *device_data,
       		     void *driver_data	)
{
}

static void
driver_close_driver( GraphicsDevice *device,
                     void           *driver_data
       			)
{
}

