/* common.c  -  Common functions */

/* Written 1993 by Werner Almesberger */

/* FAT32, VFAT, Atari format support, and various fixes additions May 1998
 * by Roman Hodek <Roman.Hodek@informatik.uni-erlangen.de> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/syscall.h>

_syscall6 (__ptr_t, mmap2, __ptr_t, addr, size_t, len, int, prot,
	   int, flags, int, fd, __off_t, pgoffset);

#include "common.h"
#include "fat32_swap.h"

extern char global_device_path[4096];
int global_used_clusters = -1;

typedef struct _link {
    void *data;
    struct _link *next;
} LINK;


void die(char *msg,...)
{
    va_list args;

    va_start(args,msg);
    vfprintf(stderr,msg,args);
    va_end(args);
    fprintf(stderr,"\n");
    exit(1);
}


void pdie(char *msg,...)
{
    va_list args;

    va_start(args,msg);
    vfprintf(stderr,msg,args);
    va_end(args);
    fprintf(stderr,":%s\n",strerror(errno));
    exit(1);
}
unsigned char *mmap_adr = NULL;
int mmap_size = 0;

void sync_mem()
{
	/*
	if(mmap_adr) {
		fprintf(stderr,"msync\n");
		if(msync(mmap_adr,mmap_size,MS_SYNC)<0)
			perror("msync");
		fprintf(stderr,"msync done\n");
	}
	*/
}

void *alloc(int size)
{
    void *this;
    // fprintf(stderr,"alloc %d\n",size); // DEBUG_JAN
    if(size > 0x1000000) {
	unsigned char *adr,*adr2;
	int used;
	long long map_offset = -1;
	int map_flags = MAP_SHARED;
	int pagesize = 4096;
	char *device = global_device_path;
    	fprintf(stderr,"alloc %d\n",size); // DEBUG_JAN
	if(mmap_adr && size<=mmap_size) {
		fprintf(stderr,"use mmap_adr\n");
		return mmap_adr;
	}
	mmap_size = size;
	fprintf(stderr,"global_used_clusters=%d\n",
			global_used_clusters);
	if(/*global_used_clusters*16 > 35000000 && */  size > 35000000) {
		map_offset = fat32_get_free_region(device,size,&used);
		if(used>=0 && used*16<35000000)
			map_offset = -1;
	}
	fprintf(stderr,"map_offset=%lld\n",map_offset);
	if(map_offset<0) {
		fprintf(stderr,"mapping /dev/zero\n");
		device = "/dev/zero";
		map_flags = MAP_PRIVATE | MAP_NORESERVE;
		map_offset = 0;
	}
	adr2 = NULL;
	adr = mmap2(NULL,size,PROT_READ | PROT_WRITE ,map_flags ,
		open(device,O_RDWR),map_offset/pagesize);
	if(adr == MAP_FAILED)  {
		fprintf(stderr,"MAP_FAILED 2\n");
		exit(1);
	}
	if(strcmp(device,"/dev/zero")!=0)
		memset(adr,0,size);
	fprintf(stderr,"adr=%x adr2=%x\n",adr,adr2);
	mmap_adr = adr;
	return adr;
    }
    if ((this = malloc(size))) {
	    memset(this,0,size);
	    return this;
    }
    pdie("malloc");
    return NULL; /* for GCC */
}


void *qalloc(void **root,int size)
{
    LINK *link;

    link = alloc(sizeof(LINK));
    link->next = *root;
    *root = link;
    return link->data = alloc(size);
}


void qfree(void **root)
{
    LINK *this;

    while (*root) {
	this = (LINK *) *root;
	*root = this->next;
	// fprintf(stderr,"qfree this->data=%x\n",this->data);
	if(this->data != mmap_adr)
		free(this->data);
	free(this);
    }
}


int min(int a,int b)
{
    return a < b ? a : b;
}


char get_key(char *valid,char *prompt)
{
    int ch,okay;

    while (1) {
	if (prompt) printf("%s ",prompt);
	fflush(stdout);
	while (ch = getchar(), ch == ' ' || ch == '\t');
	if (ch == EOF) exit(1);
	if (!strchr(valid,okay = ch)) okay = 0;
	while (ch = getchar(), ch != '\n' && ch != EOF);
	if (ch == EOF) exit(1);
	if (okay) return okay;
	printf("Invalid input.\n");
    }
}

/* Local Variables: */
/* tab-width: 8     */
/* End:             */
