/*
  File: __attr_get_fd.c

  Copyright (C) 2000
  Andreas Gruenbacher, <a.gruenbacher@computer.org>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "linux.h"


int
__attr_get_fd(
	int fd,
	const char *name,
	char **value,
	size_t size_hint)
{
	char *new_value;
	int retval = 0;

	*value = NULL;
	if (size_hint != 0) {
		*value = (char *)malloc(size_hint);
		if (*value == 0)
			return -1;
		retval = fgetxattr(fd, name, *value, size_hint);
	}
	if (size_hint == 0 || (retval < 0 && errno == ERANGE)) {
		retval = fgetxattr(fd, name, NULL, 0);
		if (retval >= 0) {
			size_hint = retval;
			new_value = (char *)realloc(*value, size_hint);
			if (new_value == NULL) {
				free(*value);
				*value = NULL;
				return -1;
			} else
				*value = new_value;
			retval = fgetxattr(fd, name, *value, size_hint);
		}
	}
	if (retval < 0) {
		if (*value) {
			free(*value);
			*value = NULL;
		}
		return -1;
	} else {
		size_hint = retval;
		new_value = (char *)realloc(*value, size_hint);
		if (new_value)
			*value = new_value;
		return retval;
	}
}

