/*
  File: acl_get_file.c

  Copyright (C) 1999, 2000
  Andreas Gruenbacher, <a.gruenbacher@computer.org>

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "linux.h"


/* 23.4.16 */
acl_t
acl_get_file(
	const char *path_p,
	acl_type_t type)
{
	char *ext_acl_p;
	const char *name;
	int retval;
	acl_t acl;

	switch(type) {
		case ACL_TYPE_ACCESS:
			name = ACL_EA_ACCESS;
			break;
		case ACL_TYPE_DEFAULT:
			name = ACL_EA_DEFAULT;
			break;
		default:
			errno = EINVAL;
			return NULL;
	}
	retval = __attr_get_file(path_p, name, &ext_acl_p,
	                         acl_ea_size(64));  /* <= 64 ACL entries */
	if (retval > 0) {
		acl = __acl_copy_int(ext_acl_p, retval);
		free(ext_acl_p);
		return acl;
	} else if (errno == ENOATTR) {
		if (type == ACL_TYPE_ACCESS)
			return acl_get_file_mode(path_p);
		else
			return acl_init(0);
	} else
		return NULL;
}

