/*
  File: syscalls.c

  Copyright (C) 1999, 2000
  Andreas Gruenbacher, <a.gruenbacher@computer.org>
 	
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <unistd.h>
#if __i386__
#  define HAVE_EA_SYSCALLS 1
#  ifndef __NR_setxattr
#define __NR_setxattr           226
#define __NR_lsetxattr          227
#define __NR_fsetxattr          228
#define __NR_getxattr           229
#define __NR_lgetxattr          230
#define __NR_fgetxattr          231
#define __NR_listxattr          232
#define __NR_llistxattr         233
#define __NR_flistxattr         234
#define __NR_removexattr        235
#define __NR_lremovexattr       236
#define __NR_fremovexattr       237
#  endif
#elif __ia64__
#  define HAVE_EA_SYSCALLS 1
#  ifndef __NR_setxattr
#define __NR_setxattr           1217
#define __NR_lsetxattr          1218
#define __NR_fsetxattr          1219
#define __NR_getxattr           1220
#define __NR_lgetxattr          1221
#define __NR_fgetxattr          1222
#define __NR_listxattr          1223
#define __NR_llistxattr         1224
#define __NR_flistxattr         1225
#define __NR_listxattr          1226
#define __NR_llistxattr         1227
#define __NR_flistxattr         1228
#  endif
#else
#  define HAVE_EA_SYSCALLS 0
#endif


#if HAVE_EA_SYSCALLS

int  setxattr(const char *path, const char *name, void *value, size_t sz, int f)
{
	return syscall(__NR_setxattr, path, name, value, sz, f);
}
int lsetxattr(const char *path, const char *name, void *value, size_t sz, int f)
{
	return syscall(__NR_lsetxattr, path, name, value, sz, f);
}
int fsetxattr(int fd, const char *name, void *value, size_t sz, int f)
{
	return syscall(__NR_fsetxattr, fd, name, value, sz, f);
}

ssize_t getxattr(const char *path, const char *name, void *value, size_t sz)
{
	return syscall(__NR_getxattr, path, name, value, sz);
}
ssize_t lgetxattr(const char *path, const char *name, void *value, size_t sz)
{
	return syscall(__NR_lsetxattr, path, name, value, sz);
}
ssize_t fgetxattr(int fd, const char *name, void *value, size_t sz)
{
	return syscall(__NR_fsetxattr, fd, name, value, sz);
}

ssize_t listxattr(const char *path, char *list, size_t sz)
{
	return syscall(__NR_listxattr, path, list, sz);
}
ssize_t llistxattr(const char *path, char *list, size_t sz)
{
	return syscall(__NR_llistxattr, path, list, sz);
}
ssize_t flistxattr(int fd, char *list, size_t sz)
{
	return syscall(__NR_flistxattr, fd, list, sz);
}

int  removexattr(const char *path, const char *name)
{
	return syscall(__NR_removexattr, path, name);
}
int lremovexattr(const char *path, const char *name)
{
	return syscall(__NR_lremovexattr, path, name);
}
int fremovexattr(int fd, const char *name)
{
	return syscall(__NR_fremovexattr, fd, name);
}

#else

int setxattr(const char *path, const char *name, void *value, size_t sz, int f)
{
	errno = ENOSYS;
	return -1;
}
int lsetxattr(const char *path, const char *name, void *value, size_t sz, int f)
{
	errno = ENOSYS;
	return -1;
}
int fsetxattr(int fd, const char *name, void *value, size_t sz, int f)
{
	errno = ENOSYS;
	return -1;
}

ssize_t getxattr(const char *path, const char *name, void *value, size_t sz)
{
	errno = ENOSYS;
	return -1;
}
ssize_t lgetxattr(const char *path, const char *name, void *value, size_t sz)
{
	errno = ENOSYS;
	return -1;
}
ssize_t fgetxattr(int fd, const char *name, void *value, size_t sz)
{
	errno = ENOSYS;
	return -1;
}

ssize_t listxattr(const char *path, char *list, size_t sz)
{
	errno = ENOSYS;
	return -1;
}
ssize_t llistxattr(const char *path, char *list, size_t sz)
{
	errno = ENOSYS;
	return -1;
}
ssize_t flistxattr(int fd, char *list, size_t sz)
{
	errno = ENOSYS;
	return -1;
}

int removexattr(const char *path, const char *name)
{
	errno = ENOSYS;
	return -1;
}
int lremovexattr(const char *path, const char *name)
{
	errno = ENOSYS;
	return -1;
}
int fremovexattr(int fd, const char *name)
{
	errno = ENOSYS;
	return -1;
}

#endif
