/*
 * Copyright (c) 2001 Silicon Graphics, Inc.  All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2.1 of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * Further, this software is distributed without any warranty that it is
 * free of the rightful claim of any third person regarding infringement
 * or the like.  Any license provided herein, whether implied or
 * otherwise, applies only to this software file.  Patent licenses, if
 * any, provided herein do not apply to combinations of this program with
 * other software, or any other product whatsoever.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307,
 * USA.
 * 
 * Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 * Mountain View, CA  94043, or:
 * 
 * http://www.sgi.com 
 * 
 * For further information regarding this notice, see: 
 * 
 * http://oss.sgi.com/projects/GenInfo/SGIGPLNoticeExplan/
 */

#include <errno.h>
#ifndef ENOATTR
# define ENOATTR ENODATA
#endif

#define XATTR_CREATE  0x1       /* set value, fail if attr already exists */
#define XATTR_REPLACE 0x2       /* set value, fail if attr does not exist */

#include <unistd.h>
#if __i386__
#  define HAVE_EA_SYSCALLS 1
#  ifndef __NR_setxattr
#define __NR_setxattr           226
#define __NR_lsetxattr          227
#define __NR_fsetxattr          228
#define __NR_getxattr           229
#define __NR_lgetxattr          230
#define __NR_fgetxattr          231
#define __NR_listxattr          232
#define __NR_llistxattr         233
#define __NR_flistxattr         234
#define __NR_removexattr        235
#define __NR_lremovexattr       236
#define __NR_fremovexattr       237
#  endif
#elif __ia64__
#  define HAVE_EA_SYSCALLS 1
#  ifndef __NR_setxattr
#define __NR_setxattr           1217
#define __NR_lsetxattr          1218
#define __NR_fsetxattr          1219
#define __NR_getxattr           1220
#define __NR_lgetxattr          1221
#define __NR_fgetxattr          1222
#define __NR_listxattr          1223
#define __NR_llistxattr         1224
#define __NR_flistxattr         1225
#define __NR_removexattr        1223
#define __NR_lremovexattr       1224
#define __NR_fremovexattr       1225
#  endif
#else
#  define HAVE_EA_SYSCALLS 0
#endif

#if HAVE_EA_SYSCALLS
# define     setxattr(path, name, value, size, flags)	\
			syscall(__NR_setxattr,  path, name, value, size, flags)
# define    lsetxattr(path, name, value, size, flags)	\
			syscall(__NR_lsetxattr, path, name, value, size, flags)
# define    fsetxattr(fd,   name, value, size, flags)	\
			syscall(__NR_fsetxattr, fd,   name, value, size, flags)
# define     getxattr(path, name, value, size)		\
			syscall(__NR_getxattr,  path, name, value, size)
# define    lgetxattr(path, name, value, size)		\
			syscall(__NR_lgetxattr, path, name, value, size)
# define    fgetxattr(fd,   name, value, size)		\
			syscall(__NR_fgetxattr, fd,   name, value, size)
# define    listxattr(path, list, size)			\
			syscall(__NR_listxattr,  path, list, size)
# define   llistxattr(path, list, size)			\
			syscall(__NR_llistxattr, path, list, size)
# define   flistxattr(fd,   list, size)			\
			syscall(__NR_flistxattr, fd,   list, size)
# define  removexattr(path, name)			\
			syscall(__NR_removexattr,  path, name)
# define lremovexattr(path, name)			\
			syscall(__NR_lremovexattr, path, name)
# define fremovexattr(fd,   name)			\
			syscall(__NR_fremovexattr, fd,   name)
#else

# define     setxattr(path, name, value, size, flags)	{ errno = ENOSYS, -1 }
# define    lsetxattr(path, name, value, size, flags)	{ errno = ENOSYS, -1 }
# define    fsetxattr(fd,   name, value, size, flags)	{ errno = ENOSYS, -1 }
# define     getxattr(path, name, value, size)		{ errno = ENOSYS, -1 }
# define    lgetxattr(path, name, value, size)		{ errno = ENOSYS, -1 }
# define    fgetxattr(fd,   name, value, size)		{ errno = ENOSYS, -1 }
# define    listxattr(path, list, size)			{ errno = ENOSYS, -1 }
# define   llistxattr(path, list, size)			{ errno = ENOSYS, -1 }
# define   flistxattr(fd,   list, size)			{ errno = ENOSYS, -1 }
# define  removexattr(path, name)			{ errno = ENOSYS, -1 }
# define lremovexattr(path, name)			{ errno = ENOSYS, -1 }
# define fremovexattr(fd,   name)			{ errno = ENOSYS, -1 }

#endif
