#include "mmac/types.h"
#include "mmac/mmac_config.h"
#include "mmac/hardware/hwpfur.h"
#include "mmac/hardware/hwpurt.h"
#include "mmac/hardware/hwbhif.h"
#include "mmac/hardware/hwvenc.h"

#include "mmac/support/dsi.h"

#include "mmac/hardware/hwvenc.h"


/***********************************************************************/
/* This file provides a poll serial interface for general debugging 	*/
/* of system problems 													*/
/************************************************************************/
#define MMAC_DSI_PMON_BUILD 	(MMAC_NO)

#define NUM_DSI_PORTS 4

/* Buffer for DSI_printf */
static char outBuff[1024];

/*-------------------------------------------------------------------------
** Purpose:    Initialise debug serial port.
**
** Returns:    Nothing.
**
** Notes  :
*/
static int dsi_init_done=0;
void DSI_init( DSI_DEVICE dev )
{
	/* assume dev 0 */
	if( dev == DSI_PORT_0 )
	{
		DSI_set_baudrate(dev, MMAC_DSI_UART0_BAUD_RATE );

		emmaPFUR0.PFUR_IIR_FCR		= 0x00;		
		emmaPFUR0.PFUR_IIR_FCR		= 0x01;
		emmaBHIF.BHIF_STRAP_WRITE_2 |= 0x00000002;
		emmaPFUR0.PFUR_IIR_FCR |= 0x00000006;
		emmaPFUR0.PFUR_IIR_FCR &= ~0x00000006;
		emmaPFUR0.PFUR_IIR_FCR		= 0x01;		

	}
	else if( dev == DSI_PORT_1 )
	{
        	DSI_set_baudrate(dev, MMAC_DSI_UART1_BAUD_RATE);
		emmaPFUR1.PFUR_IIR_FCR		= 0x01;
		emmaBHIF.BHIF_STRAP_WRITE_2 |= 0x00000004;
	}
	else if( dev == DSI_PORT_2 )
	{
		emmaPURT0.PURT_ASIM = 0xC8;
        	DSI_set_baudrate(dev, MMAC_DSI_UART2_BAUD_RATE);
		emmaBHIF.BHIF_STRAP_WRITE_2 |= 0x00000008;
	}
	else if( dev == DSI_PORT_3 )
	{
		emmaPURT1.PURT_ASIM = 0xC8;
        	DSI_set_baudrate(dev, MMAC_DSI_UART2_BAUD_RATE);
		emmaBHIF.BHIF_STRAP_WRITE_2 |= 0x00000010;

	}
	else
	{
       		 DSI_set_baudrate(dev, MMAC_DSI_UART0_BAUD_RATE);
		emmaPFUR0.PFUR_IIR_FCR		= 0x01;
		emmaBHIF.BHIF_STRAP_WRITE_2 |= 0x00000002;
	}
}

/*---------------------------------------------------------------------------
** Purpose:    Set baudrate.
**
** Returns:    Nothing.
**
** Notes  :
*/
void DSI_set_baudrate( DSI_DEVICE dev, DSI_BAUDRATE	baudrate )
{
	UI32	divisor = 0;
	UI32	divisor2 = 0;

	if(( dev == DSI_PORT_0 ) || ( dev == DSI_PORT_1))
	{
		/*
		** Set baud rate divisors - clock rate is 18.432 MHz
		*/
		switch (baudrate)
		{
			case DSI_BR_300:
				divisor = 3840;
			break;

			case DSI_BR_1200:
				divisor = 960;
			break;

			case DSI_BR_2400:
				divisor = 480;
			break;

			case DSI_BR_4800:
				divisor = 240;
			break;

			case DSI_BR_9600:
				divisor = 120;
			break;


			case DSI_BR_19200:
				divisor = 60;
			break;

			case DSI_BR_38400:
				divisor = 30;
			break;

			case DSI_BR_57600:
				divisor = 20;
			break;

			case DSI_BR_115200:
				divisor = 10;
			break;

			default:	/* 19200 */
				divisor = 60;
			break;
		}

		if( dev == DSI_PORT_0)
		{
		emmaPFUR0.PFUR_LCR = 0x80;
		emmaPFUR0.PFUR_RBR_THR_DLL = divisor & 0xFF;
		emmaPFUR0.PFUR_IER_DLM = (divisor >> 8 ) & 0xFF;
		emmaPFUR0.PFUR_LCR = 0x03;
		}
		else
		{
		emmaPFUR1.PFUR_LCR = 0x80;
		emmaPFUR1.PFUR_RBR_THR_DLL = divisor & 0xFF;
		emmaPFUR1.PFUR_IER_DLM = (divisor >> 8 ) & 0xFF;
		emmaPFUR1.PFUR_LCR = 0x03;

		}
		
	}
	else
	{


		/*
		** Set baud rate divisors. divisor is BRGC setting,
		** divisor2 is BRGMC setting. Upper 5 bits of BRGMC
		** are test bits and should always be set as
		** 00100XXXb. Clock rate is 18.432 MHz
		*/
		switch (baudrate)
		{
			case DSI_BR_9600:
				divisor = 0x0F;
				divisor2 = 0x25;
			break;

			case DSI_BR_19200:
				divisor =  0x0F;
				divisor2 = 0x24;
			break;

			case DSI_BR_38400:
				divisor = 0x0F;
				divisor2 = 0x23;
			break;

			case DSI_BR_57600:
				divisor = 0x05;
				divisor2 = 0x24;
			break;

			case DSI_BR_115200:
				divisor = 0x0A;
				divisor2 = 0x22;
			break;

			default:
				divisor =  0x0F;
				divisor2 = 0x24;
			break;
		}

		if( dev == DSI_PORT_2 )
		{
		    emmaPURT0.PURT_BRGC  = divisor;
		    emmaPURT0.PURT_BRGMC = divisor2;
		}
		else
		{
		    emmaPURT1.PURT_BRGC  = divisor;
		    emmaPURT1.PURT_BRGMC = divisor2;
		}

	}
}


MMAC_BOOL DSI_tx_byte( DSI_DEVICE dev, UI8 tx_byte )
{
	UI32 count;

	emmaBHIF.BHIF_STRAP_WRITE_2 |= 0x00000002;
	
	/*
	** Wait for Tx buffer to empty.
	*/
	if( dev == DSI_PORT_0 )
	{
		emmaPFUR0.PFUR_LSR &= ~0x40;
		emmaPFUR0.PFUR_RBR_THR_DLL = tx_byte;
#if !MMAC_DSI_PMON_BUILD
		while(( emmaPFUR0.PFUR_LSR & 0x40 ) == 0x00 )
		{
			count++;
			if( count == 0x10000 )
			{
				break;
			}
		}
#endif		
	}
	else if( dev == DSI_PORT_1 )
	{
		emmaPFUR1.PFUR_LSR &= ~0x40;
		emmaPFUR1.PFUR_RBR_THR_DLL = tx_byte;		
#if !MMAC_DSI_PMON_BUILD	
		while(( emmaPFUR1.PFUR_LSR & 0x40 ) == 0x00 );
#endif
	}
	else
	{
		if( dev == DSI_PORT_2 )
		{
			emmaPURT0.PURT_INT &= ~0x02;
			emmaPURT0.PURT_RXBTXS = tx_byte;
#if !MMAC_DSI_PMON_BUILD
			while( 1 )
			{
				if( emmaPURT0.PURT_INT & 0x02 )
				{
					break;
				}			
			}
#endif		
		}
		else
		{	
			emmaPURT1.PURT_INT &= ~0x02;
			emmaPURT1.PURT_RXBTXS = tx_byte;
#if !MMAC_DSI_PMON_BUILD
			while( 1 )
			{
				if( emmaPURT1.PURT_INT & 0x02 )
				{
					break;
				}			
			}
#endif
		}



	}
	return MMAC_TRUE;
}

MMAC_BOOL DSI_rx_byte( DSI_DEVICE dev, UI8 *pbyte )
{
	if( dev == DSI_PORT_0 )
	{
#if !MMAC_DSI_PMON_BUILD
		if(( emmaPFUR0.PFUR_LSR & 0x01 ))
#endif		
		{
			*pbyte = (UI8)emmaPFUR0.PFUR_RBR_THR_DLL;
			emmaPFUR0.PFUR_LSR &= ~0x01;
			return MMAC_TRUE;
		}
	}
	else if( dev == DSI_PORT_1 )
	{
#if !MMAC_DSI_PMON_BUILD	
		if(( emmaPFUR1.PFUR_LSR & 0x01 ))
#endif		
		{
			*pbyte = (UI8)emmaPFUR1.PFUR_RBR_THR_DLL;
			emmaPFUR1.PFUR_LSR &= ~0x01;			
			return MMAC_TRUE;
		}
	}
	else
	{
		if( dev == DSI_PORT_2 )
		{
#if !MMAC_DSI_PMON_BUILD
			if( emmaPURT0.PURT_INT & 0x04 )
#endif		
			{
				*pbyte = emmaPURT0.PURT_RXBTXS;
				emmaPURT0.PURT_INT &= ~0x04;
				return MMAC_TRUE;
			}
		}
		else
		{
#if !MMAC_DSI_PMON_BUILD
			if( emmaPURT1.PURT_INT & 0x04 )
#endif		
			{
				*pbyte = emmaPURT1.PURT_RXBTXS;
				emmaPURT1.PURT_INT &= ~0x04;
				return MMAC_TRUE;
			}
		}

	}
	return MMAC_FALSE;
}


MMAC_BOOL DSI_IsBreakDetected( DSI_DEVICE dev )
{
	UI8 lsr;

	if( dev == DSI_PORT_0 )
	{
		lsr = emmaPFUR0.PFUR_LSR;

		if( (lsr & 0x10 ) || (lsr & 0x08 ))
			return MMAC_TRUE;

		return MMAC_FALSE;
	}
	if( dev == DSI_PORT_1 )
	{
		lsr = emmaPFUR1.PFUR_LSR;

		if( (lsr & 0x10 ) || (lsr & 0x08 ))
			return MMAC_TRUE;

		return MMAC_FALSE;
	}
	else
	{
		/* No break detect for DSI 1 and 2 */
	}
	return MMAC_FALSE;
}

void DSI_PrintNum( DSI_DEVICE dev, unsigned long int num )
{
	char buff[16];
	int i = 0;
	int j;

   	DSI_tx_byte( dev,  '<' );
	DSI_tx_byte( dev, '0');
	DSI_tx_byte( dev,  'x' );

	if (num==0 )
	{
		DSI_tx_byte( dev, '0');
		DSI_tx_byte( dev, '0');
    	DSI_tx_byte( dev,  '>' );
		return;
	}

	while(num)
	{
		buff[i++] = num % 16;
		num = num / 16;
	}

	i--;

	for(j=i;j>=0;j--)
	{
		if( buff[j] <= 9 )
		{
			DSI_tx_byte( dev,  buff[j] + '0' );
		}
		else
		{
			DSI_tx_byte( dev, (buff[j]-10) + 'A');
		}
	}
	DSI_tx_byte( dev, '>' );
}

void DSI_write( DSI_DEVICE dev, char *str )
{
	while(*str)
	{
	    	if(*str == '\n')
		    DSI_tx_byte( dev, '\r' );
		DSI_tx_byte( dev, *str++ );
	}
}

void DSI_printf (DSI_DEVICE dev, const char *format, ...)
{
#if !(MMAC_DSI_PMON_BUILD)
	va_list args;

	va_start( args, format );

	vsprintf( outBuff, format, args );
	outBuff[1023] = 0;

	DSI_write( dev, outBuff );
#endif
	/* don't worry about va_end doesn't do much */
}

int printk(const char *format, ...)
{
	va_list args;
	if(!dsi_init_done) {
	    DSI_init(0);
	    dsi_init_done = 1;
	}

	va_start( args, format );

	vsprintf( outBuff, format, args );

	DSI_write( 0, outBuff );
	return strlen(outBuff);
}

MMAC_BOOL DSI_is_tx_ready( DSI_DEVICE dev  )
{
	/*
	** Wait for Tx buffer to empty.
	*/
	if( dev == DSI_PORT_0 )
	{
		if(( emmaPFUR0.PFUR_LSR & 0x40 ) == 0x00 )
		{
			return MMAC_FALSE;
		}
	}
	else if( dev == DSI_PORT_1 )
	{
		if(( emmaPFUR1.PFUR_LSR & 0x40 ) == 0x00 )
		{
			return MMAC_FALSE;
		}
	}
	else if( dev == DSI_PORT_2 )
	{
		if(( emmaPURT0.PURT_INT & 0x02 ) == 0 )
		{
			return MMAC_FALSE;
		}

	}
	else if( dev == DSI_PORT_3 )
	{
		if(( emmaPURT1.PURT_INT & 0x02 ) == 0 )
		{
			return MMAC_FALSE;
		}

	}
	return MMAC_TRUE;
}


MMAC_BOOL DSI_is_rx_ready( DSI_DEVICE dev  )
{
	if( dev == DSI_PORT_0 )
	{
		if(( emmaPFUR0.PFUR_LSR & 0x01 ) == 0)
		{
			return MMAC_FALSE;
		}
	}
	else if( dev == DSI_PORT_1 )
	{
		if(( emmaPFUR1.PFUR_LSR & 0x01 )==0)
		{
			return MMAC_FALSE;
		}
	}
	else if( dev == DSI_PORT_2 )
	{
		if(( emmaPURT0.PURT_INT & 0x04 ) == 0 )
		{
			return MMAC_FALSE;
		}
	}
	else if( dev == DSI_PORT_3 )
	{
		if(( emmaPURT1.PURT_INT & 0x04 ) == 0 )
		{
			return MMAC_FALSE;
		}
	}
	return MMAC_TRUE;
}

UI32 DSI_measure_freq( void )
{
	/* Determine clock frequency */
	return 111;
}
