#include <linux/config.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/bootmem.h>
#include <asm/irq.h>
#include <asm/cpu.h>
#include <asm/io.h>
#include <asm/pci_channel.h>
#include <asm/vr41xx/vr41xx.h>
#include <asm/vr41xx/emma2.h>

#define USE_EXTERNAL_ARBITER
/* PCI config space address macro */
#define PCI_CONFIG_ADRS(dev, reg) \
			( (dev->bus->number<<16) | (dev->devfn << 8) | (reg))

#define PCI_BASE_ADR(dev) 0xb3000000
#define PCI_INIT_REG(dev) ( (uint32_t *)&pci1Regs->init0 )

#define	PCI_INIT_TYPE_IO					(0x00000001 << 1)
#define	PCI_INIT_TYPE_MEM					(0x00000003 << 1)
#define	PCI_INIT_TYPE_CONFIG				(0x00000005 << 1)
#define	PCI_INIT_TYPE_MRM					(0x00000006 << 1)
#define	PCI_INIT_TYPE_MRL					(0x00000007 << 1)
#define	PCI_INIT_TYPE_DEFAULT				(PCI_INIT_TYPE_MEM)

/*
 * PCI Register Bit Definitions
 */

/* PCI Arbiter Enable Register */
#define	PCI_ARBEN_GNTOUT_ENABLE				(0x00000001)
#define	PCI_ARBEN_REQOUT_ENABLE				(0x00000002)

/* PCI TBUS Endian Control Register */
#define PCI_TENDIAN_BIG_ENDIAN				(0x00000000)
#define PCI_TENDIAN_LITTLE_ENDIAN			(0x00000001)

/* PCI Interrupt Register */
#define	PCI_INT_PCI_SERR					(0x00000001)
#define	PCI_INT_PCI_IERR					(0x00000002)
#define	PCI_INT_PCI_DMA_ADR					(0x00000004)
#define	PCI_INT_PCI_DMA_BUSERR				(0x00000008)
#define	PCI_INT_PCI_CBI_BUSERR				(0x00000010)

/* PCI Interrupt Enable Register */
#define	PCI_INTEN_ERR_INT_MASK				(0x0000001b)

/* PCI Control Register-H */
#define PCI_CTRL_H_ADDR_MASK				(0x0080000f)
#define PCI_CTRL_H_SERR_MASK				(0x01002f00)
#define PCI_CTRL_H_IERR_MASK				(0x022f0000)
#define PCI_CTRL_H_TACH						(0x00000001)
#define PCI_CTRL_H_MACH						(0x00000002)
#define PCI_CTRL_H_RTYCH					(0x00000004)
#define PCI_CTRL_H_PERCH					(0x00000008)
#define PCI_CTRL_H_TASE						(0x00000100)
#define PCI_CTRL_H_MASE						(0x00000200)
#define PCI_CTRL_H_RTYSE					(0x00000400)
#define PCI_CTRL_H_PERSE					(0x00000800)
#define PCI_CTRL_H_AERSE					(0x00002000)
#define PCI_CTRL_H_TAIN						(0x00010000)
#define PCI_CTRL_H_MAIN						(0x00020000)
#define PCI_CTRL_H_RTYIN					(0x00040000)
#define PCI_CTRL_H_PERIN					(0x00080000)
#define PCI_CTRL_H_AERIN					(0x00200000)
#define	PCI_CTRL_H_IBERCH					(0x00800000)
#define	PCI_CTRL_H_IBERSE					(0x01000000)
#define	PCI_CTRL_H_IBERIN					(0x02000000)
#define	PCI_CTRL_H_PCIWRST					(0x40000000)
#define	PCI_CTRL_H_PCICRST					(0x80000000)

/* PCI Error Register */
#define	PCI_ERR_IS_CPU						(0x00000001)

/* PCI Reset Register */
#define	PCI_RST_ALL							(0x00000001)
#define	PCI_RST_PCI							(0x00000002)
#define	PCI_RST_BRG							(0x00000004)
#define	PCI_RST_CBI							(0x00000008)
#define	PCI_RST_DMA							(0x00000010)
#define	PCI_RST_ARB							(0x00000020)
#define	PCI_RST_CFG							(0x00000040)

#define SIEMENS_WB2 // !!!!!!for evaluation board WB2!!!!!
#undef SIEMENS_WB2
#define SIEMENS_S //Siemens one tuner sat version

/****************************************************************/
/*			LOCAL TYPES				*/
/****************************************************************/

/* Control block for each PCI initiator window */

typedef enum
{ 
	MMAC_PCI_ERR_NO_ERROR 	= 	0, 
	MMAC_PCI_ERR_TARGET_ABORT,
	MMAC_PCI_ERR_MASTER_ABORT,
	MMAC_PCI_ERR_RETRY_LIMIT,
	MMAC_PCI_ERR_READ_PARITY,
	MMAC_PCI_ERR_WRITE_PARITY,
	MMAC_PCI_ERR_DISCARD_TIMEOUT,
	MMAC_PCI_ERR_TBUS_INTERNAL
} MMAC_PCI_ERROR_TYPE; 

/* pointers to device registers */
static volatile pci0_regs *pci0Regs = (pci0_regs *) 0xb0013000;
static volatile pci1_regs *pci1Regs = (pci1_regs *) 0xb0200000;

static int vr41xx_pci_read_config_byte(struct pci_dev *dev, int where, u8 *val)
{
	volatile u32 *adr;
    	unsigned int *initReg,oldInit;
	initReg = PCI_INIT_REG(dev);
	oldInit = *initReg;

	*initReg = PCI_INIT_TYPE_CONFIG;
	adr = (u32 *) (PCI_BASE_ADR(dev) + PCI_CONFIG_ADRS(dev, where & 0xfc));
	if (((long int)adr & 0x400) != 0)
	{
		*val = 0xff;
		*initReg = oldInit;
		printk("\n\nvr41xx_pci_read_config_byte read wrong address %lx\n\n",adr);
		return PCIBIOS_SUCCESSFUL;
	}
	printk("read config byte at adr = %x\n",adr);
	*adr;
	*val = *adr >> ((where & 3)*8);

	*initReg = oldInit;
	return PCIBIOS_SUCCESSFUL;
}

static int vr41xx_pci_read_config_word(struct pci_dev *dev, int where, u16 *val)
{
	volatile u16 *adr;
    	unsigned int *initReg,oldInit;
	initReg = PCI_INIT_REG(dev);
	oldInit = *initReg;

	*initReg = PCI_INIT_TYPE_CONFIG;
	adr = (u16 *) (PCI_BASE_ADR(dev) + PCI_CONFIG_ADRS(dev, where));
	if (((long int)adr & 0x00000400) != 0)
	{
		*val = 0xffff;
		*initReg = oldInit;
		printk("\n\n vr41xx_pci_read_config_word read wrong address %lx\n\n",adr);
		return PCIBIOS_SUCCESSFUL;
	}
	printk("read config word at adr = %x\n",adr);
	*adr;
	*val = *adr;

	*initReg = oldInit;
	return PCIBIOS_SUCCESSFUL;
}

static int vr41xx_pci_read_config_dword(struct pci_dev *dev, int where, u32 *val)
{
	volatile u32 *adr;
    	unsigned int *initReg,oldInit;
	initReg = PCI_INIT_REG(dev);
	oldInit = *initReg;

	*initReg = PCI_INIT_TYPE_CONFIG;
	adr = (u32 *) (PCI_BASE_ADR(dev) + PCI_CONFIG_ADRS(dev, where));
	if (((long int)adr & 0x00000400) != 0)
	{
		*val = 0xffffffff;
		*initReg = oldInit;
		printk("\n\n vr41xx_pci_read_config_dword read wrong address %lx\n\n",adr);
		return PCIBIOS_SUCCESSFUL;
	}
	printk("read config dword at adr = %x\n",adr);
	*adr;
	*val = *adr;

	*initReg = oldInit;
	return PCIBIOS_SUCCESSFUL;
}

static int vr41xx_pci_write_config_byte(struct pci_dev *dev, int where, u8 val)
{
	volatile u32 *adr;
    	unsigned int *initReg,oldInit;
	initReg = PCI_INIT_REG(dev);
	oldInit = *initReg;

	*initReg = PCI_INIT_TYPE_CONFIG;
	adr = (u32 *) (PCI_BASE_ADR(dev) + PCI_CONFIG_ADRS(dev, where & 0xfc));
	if (((long int)adr & 0x00000400) != 0)
	{
		printk("\n\n vr41xx_pci_write_config_byte write to wrong address %lx\n\n",adr);
		*initReg = oldInit;
		return PCIBIOS_SUCCESSFUL;
	}
	printk("write config byte at adr = %x\n",adr);
	*adr;
	*adr = *adr |  (val<<((where & 3))*8);

	*initReg = oldInit;
	return PCIBIOS_SUCCESSFUL;
}

static int vr41xx_pci_write_config_word(struct pci_dev *dev, int where, u16 val)
{
	volatile u16 *adr;
    	unsigned int *initReg,oldInit;
	initReg = PCI_INIT_REG(dev);
	oldInit = *initReg;

	*initReg = PCI_INIT_TYPE_CONFIG;
	adr = (u16 *) (PCI_BASE_ADR(dev) + PCI_CONFIG_ADRS(dev, where));
	if (((long int)adr & 0x00000400) != 0)
	{
		printk("\n\n vr41xx_pci_write_config_word write to wrong address %lx\n\n",adr);
		*initReg = oldInit;
		return PCIBIOS_SUCCESSFUL;
	}
	printk("write config word at adr = %x\n",adr);
	*adr = val;

	*initReg = oldInit;
	return PCIBIOS_SUCCESSFUL;
}

static int vr41xx_pci_write_config_dword(struct pci_dev *dev, int where, u32 val)
{
	volatile u32 *adr;
    	unsigned int *initReg,oldInit;
	initReg = PCI_INIT_REG(dev);
	oldInit = *initReg;

	*initReg = PCI_INIT_TYPE_CONFIG;
	adr = (u32 *) (PCI_BASE_ADR(dev) + PCI_CONFIG_ADRS(dev, where));
	if (((long int)adr & 0x00000400) != 0)
	{
		printk("\n\n vr41xx_pci_write_config_dword write to wrong address %lx\n\n",adr);
		*initReg = oldInit;
		return PCIBIOS_SUCCESSFUL;
	}
	printk("write config dword at adr = %x\n",adr);
	*adr = val;

	*initReg = oldInit;
	return PCIBIOS_SUCCESSFUL;
}

struct pci_ops vr41xx_pci_ops = {
	vr41xx_pci_read_config_byte,
	vr41xx_pci_read_config_word,
	vr41xx_pci_read_config_dword,
	vr41xx_pci_write_config_byte,
	vr41xx_pci_write_config_word,
	vr41xx_pci_write_config_dword
};

/*
           mode  cond_a conb_b
disabled      0     0      0
rising edge   0     1      0
falling edge  0     0      1
both edges    0     1      1
low level     1     0      x
high level    1     1      x
*/

static void set_pin_irq(int pin)
{
    if(pin < 32) {
	printk("pin must be >= 32\n");
	return;
    }
    pin -= 32;
    printk("pin %d = %d\n",pin+32, (zpio_regs_base->port_data_1 & (1<<pin)) >> pin );
    zpio_regs_base->port_direction_1 &= ~(1 << pin);
    zpio_regs_base->int_status_1 &= ~(1 << pin);
    zpio_regs_base->int_mode_1 |= (1 << pin);  /* level trigger */
    zpio_regs_base->int_condition_a1 &= ~(1 << pin); /* low */
    zpio_regs_base->int_condition_b1 |= (1 << pin);
}

static void set_pin_irq_lower(int pin)
{
    if(pin >= 32) {
	printk("pin must be < 32\n");
	return;
    }
    printk("pin %d = %d\n",pin, (zpio_regs_base->port_data_0 & (1<<pin)) >> pin );
    zpio_regs_base->port_direction_0 &= ~(1 << pin);
    zpio_regs_base->int_status_0 &= ~(1 << pin);
    zpio_regs_base->int_mode_0 |= (1 << pin);  /* level trigger */
    zpio_regs_base->int_condition_a0 &= ~(1 << pin); /* low */
    zpio_regs_base->int_condition_b0 |= (1 << pin);
}

void clear_zpio0_interrupt()
{
    zpio_regs_base->int_status_0 = 0;
}

void clear_zpio1_interrupt()
{
    zpio_regs_base->int_status_1 = 0;
}

#ifndef SIEMENS_WB2
#ifdef SIEMENS_S
static void pci_pin_isr()
{
    zpio_regs_base->int_status_0 = 0;
}
#else
static void pci_pin_isr()
{
    zpio_regs_base->int_status_1 = 0;
}
#endif
#else
static void pci_pin_isr()
{
    zpio_regs_base->int_status_0 = 0;
}
#endif

static void PCI_ISR(void)
{
	uint32_t intStatus;
	MMAC_PCI_ERROR_TYPE errorType;

	intStatus = pci0Regs->pci_int;
	pci0Regs->pci_int = 0;

	errorType = (MMAC_PCI_ERROR_TYPE)
					((pci1Regs->ctrl_h >> 5) & 0x00000007);

	if(errorType != MMAC_PCI_ERR_NO_ERROR) {
	    	unsigned int adr;
		printk("#pci errortype = %d\n",errorType);
		if(pci1Regs->err & PCI_ERR_IS_CPU)
		   	printk("#cpu as master\n"); 
		else
		   	printk("#cpu as slave\n"); 
		adr = pci1Regs->err & 0xfffffffc;
		printk("#adr = %08x\n",adr);
		if(intStatus & PCI_INT_PCI_CBI_BUSERR) {
			adr = pci0Regs->cbi_eadr & 0xffffffe0;
			printk("#cbus_err tbus adr = %08x\n",adr);
			pci0Regs->cbi_eadr = 0;
		} else if(intStatus & PCI_INT_PCI_DMA_BUSERR) {
			adr = pci0Regs->dma_eadr & 0xffffffe0;
		    	printk("#dma_buserr tbus_adr=%08x\n",adr);
			pci0Regs->dma_eadr = 0;
		}

	}
	pci1Regs->ctrl_h &= 0xffffff1f;
	pci1Regs->err = 0;
	if(intStatus & PCI_INT_PCI_SERR)
	{
	    	printk("#SERR bit set\n");
	}
	if(intStatus & PCI_INT_PCI_IERR)
	{
	    	printk("#IERR bit set\n");
	}
}

#define IRQ_ZPIO0 48
#define IRQ_ZPIO1 49
#define IRQ_PCI   19

static void emma2_init_irqs()
{
    int retval;
    unsigned int value;
    int i;

#ifndef SIEMENS_WB2
#ifdef SIEMENS_S
		set_pin_irq_lower(28);
		set_pin_irq_lower(29);
		set_pin_irq_lower(30);
		set_pin_irq_lower(31);
		printk("Wavebox S!\n");
#else
    // retval = request_irq(IRQ_ZPIO1, pci_pin_isr, SA_SHIRQ, "pci pins" , 0x12345678);
    for(i=33;i<=40;i++)
	set_pin_irq(i);
#endif
#else
    // retval = request_irq(IRQ_ZPIO0, pci_pin_isr, SA_SHIRQ, "pci pins" , 0x12345678);
    for(i=24;i<=31;i++)
		set_pin_irq_lower(i);
		printk(" Version for WB2\n");
#endif
    pci0Regs->inten = 0;
    pci0Regs->pci_int = 0;

		/* enable the PCI interrupt and install a handler */
    retval = request_irq(IRQ_PCI, PCI_ISR, SA_SHIRQ, "pci controller" , 0x12345678);
    pci0Regs->inten = 0x1f; /* all enabled */
    value = pci1Regs->ctrl_h & 0xff7fffe0;
    value |= PCI_CTRL_H_TACH | PCI_CTRL_H_MACH | PCI_CTRL_H_PERCH | PCI_CTRL_H_IBERCH;
    pci1Regs->ctrl_h = value;
    value = pci1Regs->ctrl_h & ~PCI_CTRL_H_IERR_MASK;
    value |= PCI_CTRL_H_TAIN;
    value |= PCI_CTRL_H_MAIN;
    value |= PCI_CTRL_H_RTYIN;
    value |= PCI_CTRL_H_PERIN;
    value |= PCI_CTRL_H_AERIN;
    value |= PCI_CTRL_H_IBERIN;
    pci1Regs->ctrl_h = value;
}

void __init vr41xx_pciu_init(struct vr41xx_pci_address_map *map)
{
	struct vr41xx_pci_address_space *s;
	u32 config;
	int n;
/*
	printk("1 DV: PCI bus not used!");
	return;
*/
	if (!map)
		return;
	pci0Regs->tendian = PCI_TENDIAN_LITTLE_ENDIAN;
	pci0Regs->inten = 0;
	pci0Regs->pci_int = 0;
	pci1Regs->cmd_sts |= 0x2; 
	pci0Regs->pdar_brg      = 0x00000027;
    	pci1Regs->barc = 0x00000000;
    	pci1Regs->bar0 = 0x20000000;
#ifdef USE_EXTERNAL_ARBITER
	pci0Regs->arben = PCI_ARBEN_REQOUT_ENABLE;
#else
	pci0Regs->arben = PCI_ARBEN_GNTOUT_ENABLE;
#endif
	pci1Regs->init0 = ((unsigned int) map->mem1->pci_base)   | PCI_INIT_TYPE_MEM;
	pci1Regs->init1 = ((unsigned int) map->io->pci_base) | PCI_INIT_TYPE_IO;
    	pci1Regs->swp = 8;
}

void __init pcibios_fixup_resources(struct pci_dev *dev)
{
}

void __init pcibios_fixup(void)
{
}
#ifndef SIEMENS_WB2
#ifdef SIEMENS_S
void __init pcibios_fixup_irqs(void)
{
	struct pci_dev *dev;

	pci_for_each_dev(dev) {
		dev->irq = IRQ_ZPIO0; /* ZPIO0 */
	}
	emma2_init_irqs();
	printk("pcibios_fixup_irqs Siemens S\n");
}
#else
void __init pcibios_fixup_irqs(void)
{
	struct pci_dev *dev;

	pci_for_each_dev(dev) {
		dev->irq = IRQ_ZPIO1; /* ZPIO1 */
	}
	emma2_init_irqs();
}
#endif
#else
void __init pcibios_fixup_irqs(void)
{
	struct pci_dev *dev;

/*
	printk("2 DV: PCI bus not used!");
	return;
*/

	pci_for_each_dev(dev) {
		dev->irq = IRQ_ZPIO0; /* ZPIO0 */
	}
	emma2_init_irqs();
}
#endif

unsigned int pcibios_assign_all_busses(void)
{
	return 0;
}

struct pci_fixup pcibios_fixups[] = {
	{ PCI_FIXUP_HEADER, PCI_ANY_ID, PCI_ANY_ID, pcibios_fixup_resources },
	{ 0 }
};

/*
 * PCI autoconfiguration library
 *
 * Author: Matt Porter <mporter@mvista.com>
 *
 * Copyright 2000, 2001 MontaVista Software Inc.
 * Copyright 2001 Bradley D. LaRonde <brad@ltc.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

/*
 * Modified for MIPS by Jun Sun, jsun@mvista.com
 *
 * . Simplify the interface between pci_auto and the rest: a single function.
 * . Assign resources from low address to upper address.
 * . change most int to u32.
 *
 * Further modified to include it as mips generic code, ppopov@mvista.com.
 *
 * 2001-10-26  Bradley D. LaRonde <brad@ltc.com>
 * - Add a top_bus argument to the "early config" functions so that
 *   they can set a fake parent bus pointer to convince the underlying
 *   pci ops to use type 1 configuration for sub busses.
 * - Set bridge base and limit registers correctly.
 * - Align io and memory base properly before and after bridge setup.
 * - Don't fall through to pci_setup_bars for bridge.
 * - Reformat the debug output to look more like lspci's output.
 */


#define	DEBUG
#ifdef 	DEBUG
#define	DBG(x...)	printk(x)
#else
#define	DBG(x...)
#endif

/*
 * These functions are used early on before PCI scanning is done
 * and all of the pci_dev and pci_bus structures have been created.
 */
static struct pci_dev *fake_pci_dev(struct pci_channel *hose,
	int top_bus, int busnr, int devfn)
{
	static struct pci_dev dev;
	static struct pci_bus bus;

	dev.bus = &bus;
	dev.sysdata = hose;
	dev.devfn = devfn;
	bus.number = busnr;
	bus.ops = hose->pci_ops;

	if(busnr != top_bus)
		/* Fake a parent bus structure. */
		bus.parent = &bus;
	else
		bus.parent = NULL;

	return &dev;
}

#define EARLY_PCI_OP(rw, size, type)					\
int early_##rw##_config_##size(struct pci_channel *hose,		\
	int top_bus, int bus, int devfn, int offset, type value)	\
{									\
	return pci_##rw##_config_##size(				\
		fake_pci_dev(hose, top_bus, bus, devfn),		\
		offset, value);						\
}

EARLY_PCI_OP(read, byte, u8 *)
EARLY_PCI_OP(read, word, u16 *)
EARLY_PCI_OP(read, dword, u32 *)
EARLY_PCI_OP(write, byte, u8)
EARLY_PCI_OP(write, word, u16)
EARLY_PCI_OP(write, dword, u32)

static struct resource *io_resource_inuse;
static struct resource *mem_resource_inuse;

static u32 pciauto_lower_iospc;
static u32 pciauto_upper_iospc;

static u32 pciauto_lower_memspc;
static u32 pciauto_upper_memspc;

void __init
pciauto_setup_bars(struct pci_channel *hose,
		   int top_bus,
		   int current_bus,
		   int pci_devfn)
{
	u32 bar_response, bar_size, bar_value;
	u32 bar, addr_mask, bar_nr = 0;
	u32 * upper_limit;
	u32 * lower_limit;
	int found_mem64 = 0;

/*
	printk("3 DV: PCI bus not used!");
	return;
*/

	for (bar = PCI_BASE_ADDRESS_0; bar <= PCI_BASE_ADDRESS_5; bar+=4) {
		/* Tickle the BAR and get the response */
		early_write_config_dword(hose, top_bus,
					 current_bus,
					 pci_devfn,
					 bar,
					 0xffffffff);
		early_read_config_dword(hose, top_bus,
					current_bus,
					pci_devfn,
					bar,
					&bar_response);

		/* If BAR is not implemented go to the next BAR */
		if (!bar_response)
			continue;

		/*
		 * Workaround for a BAR that doesn't use its upper word,
		 * like the ALi 1535D+ PCI DC-97 Controller Modem (M5457).
		 * bdl <brad@ltc.com>
		 */
		if (!(bar_response & 0xffff0000))
			bar_response |= 0xffff0000;

retry:
		/* Check the BAR type and set our address mask */
		if (bar_response & PCI_BASE_ADDRESS_SPACE) {
			addr_mask = PCI_BASE_ADDRESS_IO_MASK;
			upper_limit = &pciauto_upper_iospc;
			lower_limit = &pciauto_lower_iospc;
			DBG("        I/O");
		} else {
			if ((bar_response & PCI_BASE_ADDRESS_MEM_TYPE_MASK) ==
			    PCI_BASE_ADDRESS_MEM_TYPE_64)
				found_mem64 = 1;

			addr_mask = PCI_BASE_ADDRESS_MEM_MASK;
			upper_limit = &pciauto_upper_memspc;
			lower_limit = &pciauto_lower_memspc;
			DBG("        Mem");
		}


		/* Calculate requested size */
		bar_size = ~(bar_response & addr_mask) + 1;

		/* Allocate a base address */
		bar_value = ((*lower_limit - 1) & ~(bar_size - 1)) + bar_size;

		if ((bar_value + bar_size) > *upper_limit) {
			if (bar_response & PCI_BASE_ADDRESS_SPACE) {
				if (io_resource_inuse->child) {
					io_resource_inuse =
						io_resource_inuse->child;
					pciauto_lower_iospc =
						io_resource_inuse->start;
					pciauto_upper_iospc =
						io_resource_inuse->end + 1;
					goto retry;
				}

			} else {
				if (mem_resource_inuse->child) {
					mem_resource_inuse =
						mem_resource_inuse->child;
					pciauto_lower_memspc =
						mem_resource_inuse->start;
					pciauto_upper_memspc =
						mem_resource_inuse->end + 1;
					goto retry;
				}
			}
			DBG(" unavailable -- skipping\n");
			continue;
		}

		/* Write it out and update our limit */
		early_write_config_dword(hose, top_bus, current_bus, pci_devfn,
					 bar, bar_value);

		*lower_limit = bar_value + bar_size;

		/*
		 * If we are a 64-bit decoder then increment to the
		 * upper 32 bits of the bar and force it to locate
		 * in the lower 4GB of memory.
		 */
		if (found_mem64) {
			bar += 4;
			early_write_config_dword(hose, top_bus,
						 current_bus,
						 pci_devfn,
						 bar,
						 0x00000000);
		}

		DBG(" at 0x%.8x [size=0x%x]\n", bar_value, bar_size);

		bar_nr++;
	}

}

void __init
pciauto_prescan_setup_bridge(struct pci_channel *hose,
			     int top_bus,
			     int current_bus,
			     int pci_devfn,
			     int sub_bus)
{
/*
	printk("4 DV: PCI bus not used!");
	return;
*/

	/* Configure bus number registers */
	early_write_config_byte(hose, top_bus, current_bus, pci_devfn,
	                        PCI_PRIMARY_BUS, current_bus);
	early_write_config_byte(hose, top_bus, current_bus, pci_devfn,
				PCI_SECONDARY_BUS, sub_bus + 1);
	early_write_config_byte(hose, top_bus, current_bus, pci_devfn,
				PCI_SUBORDINATE_BUS, 0xff);

	/* Align memory and I/O to 1MB and 4KB boundaries. */
	pciauto_lower_memspc = (pciauto_lower_memspc + (0x100000 - 1))
		& ~(0x100000 - 1);
	pciauto_lower_iospc = (pciauto_lower_iospc + (0x1000 - 1))
		& ~(0x1000 - 1);

	/* Set base (lower limit) of address range behind bridge. */
	early_write_config_word(hose, top_bus, current_bus, pci_devfn,
		PCI_MEMORY_BASE, pciauto_lower_memspc >> 16);
	early_write_config_byte(hose, top_bus, current_bus, pci_devfn,
		PCI_IO_BASE, (pciauto_lower_iospc & 0x0000f000) >> 8);
	early_write_config_word(hose, top_bus, current_bus, pci_devfn,
		PCI_IO_BASE_UPPER16, pciauto_lower_iospc >> 16);

	/* We don't support prefetchable memory for now, so disable */
	early_write_config_word(hose, top_bus, current_bus, pci_devfn,
				PCI_PREF_MEMORY_BASE, 0);
	early_write_config_word(hose, top_bus, current_bus, pci_devfn,
				PCI_PREF_MEMORY_LIMIT, 0);
}

void __init
pciauto_postscan_setup_bridge(struct pci_channel *hose,
			      int top_bus,
			      int current_bus,
			      int pci_devfn,
			      int sub_bus)
{
	u32 temp;

/*
	printk("5 DV: PCI bus not used!");
	return;
*/
	/* Configure bus number registers */
	early_write_config_byte(hose, top_bus, current_bus, pci_devfn,
				PCI_SUBORDINATE_BUS, sub_bus);

	/* Set upper limit of address range behind bridge. */
	early_write_config_word(hose, top_bus, current_bus, pci_devfn,
		PCI_MEMORY_LIMIT, pciauto_lower_memspc >> 16);
	early_write_config_byte(hose, top_bus, current_bus, pci_devfn,
		PCI_IO_LIMIT, (pciauto_lower_iospc & 0x0000f000) >> 8);
	early_write_config_word(hose, top_bus, current_bus, pci_devfn,
		PCI_IO_LIMIT_UPPER16, pciauto_lower_iospc >> 16);

	/* Align memory and I/O to 1MB and 4KB boundaries. */
	pciauto_lower_memspc = (pciauto_lower_memspc + (0x100000 - 1))
		& ~(0x100000 - 1);
	pciauto_lower_iospc = (pciauto_lower_iospc + (0x1000 - 1))
		& ~(0x1000 - 1);

	/* Enable memory and I/O accesses, enable bus master */
	early_read_config_dword(hose, top_bus, current_bus, pci_devfn,
		PCI_COMMAND, &temp);
	early_write_config_dword(hose, top_bus, current_bus, pci_devfn,
		PCI_COMMAND, temp | PCI_COMMAND_IO | PCI_COMMAND_MEMORY
		| PCI_COMMAND_MASTER);
}

#define      PCIAUTO_IDE_MODE_MASK           0x05

int __init
pciauto_bus_scan(struct pci_channel *hose, int top_bus, int current_bus)
{
	int sub_bus;
	u32 pci_devfn, pci_class, cmdstat, found_multi=0;
	unsigned short vid, did;
	unsigned char header_type;
	int devfn_start = 0;
	int devfn_stop = 0xff;

/*
	printk("6 DV: PCI bus not used!");
	return;
*/

	sub_bus = current_bus;

	if (hose->first_devfn)
		devfn_start = hose->first_devfn;
	if (hose->last_devfn)
		devfn_stop = hose->last_devfn;

	for (pci_devfn=devfn_start; pci_devfn<devfn_stop; pci_devfn++) {

		printk("pci_devfn = %x\n",pci_devfn);
		if (PCI_FUNC(pci_devfn) && !found_multi)
			continue;

		early_read_config_word(hose, top_bus, current_bus, pci_devfn,
				       PCI_VENDOR_ID, &vid);

		if (vid == 0xffff) continue;

		early_read_config_byte(hose, top_bus, current_bus, pci_devfn,
				       PCI_HEADER_TYPE, &header_type);

		if (!PCI_FUNC(pci_devfn))
			found_multi = header_type & 0x80;

		early_read_config_word(hose, top_bus, current_bus, pci_devfn,
				       PCI_DEVICE_ID, &did);

		early_read_config_dword(hose, top_bus, current_bus, pci_devfn,
					PCI_CLASS_REVISION, &pci_class);

		DBG("%.2x:%.2x.%x Class %.4x: %.4x:%.4x",
			current_bus, PCI_SLOT(pci_devfn), PCI_FUNC(pci_devfn),
			pci_class >> 16, vid, did);
		if (pci_class & 0xff)
			DBG(" (rev %.2x)", pci_class & 0xff);
		DBG("\n");

		if ((pci_class >> 16) == PCI_CLASS_BRIDGE_PCI) {
			DBG("        Bridge: primary=%.2x, secondary=%.2x\n",
				current_bus, sub_bus + 1);
			pciauto_prescan_setup_bridge(hose, top_bus, current_bus,
						     pci_devfn, sub_bus);
			DBG("Scanning sub bus %.2x, I/O 0x%.8x, Mem 0x%.8x\n",
				sub_bus + 1,
				pciauto_lower_iospc, pciauto_lower_memspc);
			sub_bus = pciauto_bus_scan(hose, top_bus, sub_bus+1);
			DBG("Back to bus %.2x\n", current_bus);
			pciauto_postscan_setup_bridge(hose, top_bus, current_bus,
						      pci_devfn, sub_bus);
			continue;
		} else if ((pci_class >> 16) == PCI_CLASS_STORAGE_IDE) {

			unsigned char prg_iface;

			early_read_config_byte(hose, top_bus, current_bus,
				pci_devfn, PCI_CLASS_PROG, &prg_iface);
			if (!(prg_iface & PCIAUTO_IDE_MODE_MASK)) {
				DBG("Skipping legacy mode IDE controller\n");
				continue;
			}
		}

 		/*
		 * Found a peripheral, enable some standard
		 * settings
		 */
		early_read_config_dword(hose, top_bus, current_bus, pci_devfn,
					PCI_COMMAND, &cmdstat);
		early_write_config_dword(hose, top_bus, current_bus, pci_devfn,
					 PCI_COMMAND, cmdstat | PCI_COMMAND_IO |
					 PCI_COMMAND_MEMORY |
					 PCI_COMMAND_MASTER);
		early_write_config_byte(hose, top_bus, current_bus, pci_devfn,
					PCI_LATENCY_TIMER, 0x80);

		/* Allocate PCI I/O and/or memory space */
		pciauto_setup_bars(hose, top_bus, current_bus, pci_devfn);
	}
	return sub_bus;
}

int __init
pciauto_assign_resources(int busno, struct pci_channel *hose)
{
/*
	printk("7 DV: PCI bus not used!");
	return;
*/

	DBG("Autoconfig PCI channel 0x%p\n", hose);
	DBG("Scanning bus %.2x, I/O 0x%.8x:0x%.8x, Mem 0x%.8x:0x%.8x\n",
		busno, pciauto_lower_iospc, pciauto_upper_iospc,
		pciauto_lower_memspc, pciauto_upper_memspc);

	return pciauto_bus_scan(hose, busno, busno);
}

extern int pciauto_assign_resources(int busno, struct pci_channel * hose);

void __init pcibios_init(void)
{
	struct pci_channel *p;
	struct pci_bus *bus;
	int busno;

/*
	printk("8 DV: PCI bus not used!\n");
	return;
*/

	/* assign resources */
	busno=2;
	/* setup resource limits */
	io_resource_inuse = mips_pci_channels->io_resource;
	mem_resource_inuse = mips_pci_channels->mem_resource;

	pciauto_lower_iospc = io_resource_inuse->start + 0x1000;
	pciauto_upper_iospc = io_resource_inuse->end + 1;
	pciauto_lower_memspc = mem_resource_inuse->start;
	pciauto_upper_memspc = mem_resource_inuse->end + 1;
	for (p= mips_pci_channels; p->pci_ops != NULL; p++) {
		pciauto_assign_resources(busno, p);
		busno *= 2;
	}
	printk("scan the buses\n"); // DEBUG_JAN

	/* scan the buses */
	busno = 2;
	for (p= mips_pci_channels; p->pci_ops != NULL; p++) {
	    struct pci_dev dev0;
	    bus = pci_alloc_primary_bus(busno);
	    if (bus) {
		    bus->sysdata = p;
		    bus->ops = p->pci_ops;
		    bus->subordinate = busno;
		    /* Create a device template */
		    memset(&dev0, 0, sizeof(dev0));
		    dev0.bus = bus;
		    dev0.sysdata = bus->sysdata;

		    dev0.devfn = 0;
		    printk("scan_slot bus=%x\n",bus); // DEBUG_JAN
		    pci_scan_slot(&dev0);
		    printk("fixup_bus\n"); // DEBUG_JAN
		    pcibios_fixup_bus(bus);
		    printk("fixup_bus done\n"); // DEBUG_JAN
	    }
	    busno *= 2;
	}
	printk("pcibios_fixup\n"); // DEBUG_JAN

	/* machine dependent fixups */
	pcibios_fixup();
	printk("pcibios_fixup_irqs\n"); // DEBUG_JAN
	/* fixup irqs (board specific routines) */
	pcibios_fixup_irqs();
	printk("pcibios_fixup_irqs done\n"); // DEBUG_JAN
}

int pcibios_enable_device(struct pci_dev *dev,int mask)
{
	/* pciauto_assign_resources() will enable all devices found */
	return 0;
}

unsigned long __init pci_bridge_check_io(struct pci_dev *bridge)
{
	u16 io;

/*
	printk("9 DV: PCI bus not used!");
	return;
*/

	pci_read_config_word(bridge, PCI_IO_BASE, &io);
	if (!io) {
		pci_write_config_word(bridge, PCI_IO_BASE, 0xf0f0);
		pci_read_config_word(bridge, PCI_IO_BASE, &io);
		pci_write_config_word(bridge, PCI_IO_BASE, 0x0);
	}
	if (io)
		return IORESOURCE_IO;
	printk(KERN_WARNING "PCI: bridge %s does not support I/O forwarding!\n",
				bridge->name);
	return 0;
}

void __init pcibios_fixup_bus(struct pci_bus *bus)
{
	/* Propogate hose info into the subordinate devices.  */

	struct pci_channel *hose = bus->sysdata;
	struct pci_dev *dev = bus->self;

/*
	printk("10 DV: PCI bus not used!");
	return;
*/

	if (!dev) {
		/* Root bus */
		bus->resource[0] = hose->io_resource;
		bus->resource[1] = hose->mem_resource;
	} else {
		/* This is a bridge. Do not care how it's initialized,
		   just link its resources to the bus ones */
		int i;

		for(i=0; i<3; i++) {
			bus->resource[i] =
				&dev->resource[PCI_BRIDGE_RESOURCES+i];
			bus->resource[i]->name = bus->name;
		}
		bus->resource[0]->flags |= pci_bridge_check_io(dev);
		bus->resource[1]->flags |= IORESOURCE_MEM;
		/* For now, propogate hose limits to the bus;
		   we'll adjust them later. */
		bus->resource[0]->end = hose->io_resource->end;
		bus->resource[1]->end = hose->mem_resource->end;
		/* Turn off downstream PF memory address range by default */
		bus->resource[2]->start = 1024*1024;
		bus->resource[2]->end = bus->resource[2]->start - 1;
	}
}

char *pcibios_setup(char *str)
{
	return str;
}

void
pcibios_align_resource(void *data, struct resource *res, unsigned long size,
		       unsigned long align)
{
	/* this should not be called */
}

void
pcibios_update_resource(struct pci_dev *dev, struct resource *root,
			struct resource *res, int resource)
{
	/* this should not be called */
}
