/* $Id: power.c,v 1.1.1.1 2001/05/02 13:47:08 mfloeer Exp $
 *
 * VR41xx reset and power management interface
 * PM User interface loosely based on Andrew Henroid's i386 ACPI driver
 *
 * Copyright (C) 2000 Michael Klar
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/sysctl.h>
#include <linux/pm.h>
#include <linux/acpi.h>
#include <linux/irq.h>	/* for disable_irq, enable_irq, to be removed */
#include <asm/vr41xx.h>
#include <asm/cacheops.h>
#include <asm/mipsregs.h>
#include <asm/pgalloc.h>
#include <asm/power.h>

/*
 * vr41xx_dma_sem is needed because we cannot suspend while a DMA transfer
 * is in progress, otherwise the CPU will be left in an undefined state.
 */
extern struct semaphore vr41xx_dma_sem;

void vr41xx_restart(char* c)
{
}

void vr41xx_hibernate(void)
{
#ifdef CONFIG_PM
	pm_send_all(PM_SUSPEND, (void *)1);
#endif
	while(1);
}

#ifdef CONFIG_PM

unsigned int powerevent_queued;
unsigned int hibernation_state = LOAD_MAGIC;
extern void do_pm_irq_request(pm_request_t rqst);

asmlinkage void do_hibernate(void *sp)
{
	int i;

	cli();
	powerevent_queued = 0;

	if (pm_send_all(PM_SUSPEND, (void *)1)) {
		sti();
		up(&vr41xx_dma_sem);
		return;
	}
	do_pm_irq_request(PM_SUSPEND);
	hibernation_state = HIB_MAGIC;
	flush_cache_all();
	vr41xx_hibernate();
	// never reached...
}

asmlinkage int do_wakeup(void)
{
	return 0;
}

/*
 * Enter system sleep state (hibernate, really soft-off)
 *
 * Actually, all this does is to set a flag, so the tail end of the syscall
 * handler jumps to do_hibernate above, rather than executing
 * o32_ret_from_sys_call.  That just makes for a convenient place to restore
 * context from, since context is already saved on the stack at that point,
 * and bottom-half processing already done.  The wakeup part is done in
 * kernel_entry, since waking up from hibernate will restart at reset vector.
 */
static int pm_do_sleep(ctl_table *ctl,
			 int write,
			 struct file *file,
			 void *buffer,
			 size_t *len)
{
	if (!write) {
		if (file->f_pos) {
			*len = 0;
			return 0;
		}
	} else {
		if (file->f_flags & O_NONBLOCK) {
                	if (down_trylock(&vr41xx_dma_sem))
                	        return -EAGAIN;
	        } else {
	                if (down_interruptible(&vr41xx_dma_sem))
        	                return -ERESTARTSYS;
	        }

		powerevent_queued = 1;
	}
	file->f_pos += *len;
	return 0;
}

/*
 * NOTE1: We're hijacking the ACPI binary ctl_name's for now.  Don't expect
 *        this to necessarily behave as ACPI does, some are blatently wrong.
 * NOTE2: Don't get used to the text names, either, they are subject to change
 *        if we ever come up with a real naming convention....
 * NOTE3: For that matter, don't get used to this being done via sysctl,
 *        that's subject to change, too.
 */
static struct ctl_table pm_table[] =
{
	{ACPI_SLEEP, "sleep", NULL, 0, 0600, NULL, &pm_do_sleep},
	{0}
};

static struct ctl_table pm_dir_table[] =
{
	{CTL_ACPI, "pm", NULL, 0, 0555, pm_table},
	{0}
};

/*
 * Initialize the power management user interface
 */
static int __init pm_init(void)
{
	register_sysctl_table(pm_dir_table, 1);
	return 0;
}

__initcall(pm_init);

#endif // CONFIG_PM
