/*
	linux/arch/mips/vr41xx/vrc4173-audio.c

	VRC4173 audio driver derived from linux/drivers/sound/vr41xx.c

	VR41XX DMA based 10bit DAC sound driver derived from dmasound.c (of AMIGA)

	Copyright (C) 2000 Michael R. McDonald
	Copyright (C) 2000 by Hiroshi Kawashima <kawashima@iname.com>

	This is distributed under the GNU GENERAL PUBLIC LICENSE (GPL)
	Version 2 (June 1991). See the "COPYING" file distributed with this
	software for more info.
*/

/*
	**** WARNING ****
	This code is Alpha code. It may hang up your kernel.
	Please use this with care.

	**** Supported feature ****
	* Linear signed/unsigned 8/16bit big/little endian PCM.
	* 8bit A-Law/Mu-Law.
	* Stereo -> Mono conversion (just compute average of R/L-ch).

	**** Known Limitation ****
	* Will work only on :
		* MobileGear-II/Pro
		* IBM WorkPad Z50
	  plarform. (Because speaker On/Off code (via GPIO) depends on
	  each platform).
	  Tested on MobileGear-II/R300 (Japanese model) && Z50.
	* If your machine has more than 32M byte, it may not work.
	  (Limitation of VR41xx's sound DMA architecture)
	* Volume setting is ignored. (always full volume)
	
	Enjoy! -- kawashima
*/

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/major.h>
#include <linux/config.h>
#include <linux/fcntl.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/sound.h>
#include <linux/init.h>
#include <linux/pci.h>

#include <linux/delay.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <asm/uaccess.h>

#include <asm/vrc4173.h>
#include <asm/vr41xx-platdep.h>

#include <linux/soundcard.h>

extern struct pci_dev *vrc4173_dev;

void vrc4173_turn_on_device(struct pci_dev *pdev, int device);
void vrc4173_turn_off_device(struct pci_dev *pdev, int device);
void vrc4173_and_or_reg(struct pci_dev *pdev, int reg, unsigned short and, unsigned short or);

static struct pci_device_id vrc4173_ac97_tbl[] __devinitdata = {
        { PCI_VENDOR_ID_NEC, PCI_DEVICE_ID_NEC_VRC4173_AC97, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
        { 0, }
};
MODULE_DEVICE_TABLE(pci, vrc4173_ac97_tbl);

static struct pci_dev *ac97_dev;

static void ac97_speaker_on(struct pci_dev *);
static void ac97_speaker_off(struct pci_dev *);

// #define DEBUG	1

#define	HAS_10BIT_TABLES

#ifdef MODULE
static int sq_unit = -1;
static int mixer_unit = -1;
static int state_unit = -1;
static int irq_installed = 0;
#endif /* MODULE */
static char **sound_buffers = NULL;
static char **vrc4173_dma_bufs = NULL;

/*** Some declarations *******************************************************/
/*==== Copied from old dmasound.h ====*/
/*
 * Minor numbers for the sound driver.
 *
 * Unfortunately Creative called the codec chip of SB as a DSP. For this
 * reason the /dev/dsp is reserved for digitized audio use. There is a
 * device for true DSP processors but it will be called something else.
 * In v3.0 it's /dev/sndproc but this could be a temporary solution.
 */

#define SND_NDEVS	256		/* Number of supported devices */
#define SND_DEV_CTL	0		/* Control port /dev/mixer */
#define SND_DEV_SEQ	1		/* Sequencer output /dev/sequencer (FM
							   synthesizer and MIDI output) */
#define SND_DEV_MIDIN	2	/* Raw midi access */
#define SND_DEV_DSP		3	/* Digitized voice /dev/dsp */
#define SND_DEV_AUDIO	4	/* Sparc compatible /dev/audio */
#define SND_DEV_DSP16	5	/* Like /dev/dsp but 16 bits/sample */
#define SND_DEV_STATUS	6	/* /dev/sndstat */
/* #7 not in use now. Was in 2.4. Free for use after v3.0. */
#define SND_DEV_SEQ2	8	/* /dev/sequencer, level 2 interface */
#define SND_DEV_SNDPROC 9	/* /dev/sndproc for programmable devices */
#define SND_DEV_PSS	SND_DEV_SNDPROC

#define DSP_DEFAULT_SPEED	8000

#define ON		1
#define OFF		0

#define MAX_AUDIO_DEV	5
#define MAX_MIXER_DEV	2
#define MAX_SYNTH_DEV	3
#define MAX_MIDI_DEV	6
#define MAX_TIMER_DEV	3
/*====================================*/

#define DMASND_TT				1
#define DMASND_FALCON			2
#define DMASND_AMIGA			3
#define DMASND_AWACS			4
#ifdef CONFIG_SOUND_VRC4173
#  define	DMASND_VRC4173		5
#endif

#define MAX_CATCH_RADIUS		10
#define MIN_BUFFERS				4
#ifdef CONFIG_SOUND_VRC4173
#  define MIN_BUFSIZE 			2	/* Always just only 2K byte allowed */
#  define MAX_BUFSIZE			2	/* Always just only 2K byte allowed */
#else
#  define MIN_BUFSIZE 			4
#  define MAX_BUFSIZE			128	/* Limit for Amiga */
#endif

static int catchRadius = 0;
/* Real buffer size is bufSize * 1024 byte */
static int numBufs = 64, bufSize = 2;

MODULE_PARM(catchRadius, "i");
MODULE_PARM(numBufs, "i");
MODULE_PARM(bufSize, "i");
MODULE_PARM(numreadBufs, "i");
MODULE_PARM(readbufSize, "i");

#define arraysize(x)	(sizeof(x)/sizeof(*(x)))
#define min(x, y)	((x) < (y) ? (x) : (y))
#define le2be16(x)	(((x)<<8 & 0xff00) | ((x)>>8 & 0x00ff))
#define le2be16dbl(x)	(((x)<<8 & 0xff00ff00) | ((x)>>8 & 0x00ff00ff))

#define IOCTL_IN(arg, ret) \
	do { int error = get_user(ret, (int *)(arg)); \
		if (error) return error; \
	} while (0)
#define IOCTL_OUT(arg, ret)	ioctl_return((int *)(arg), ret)


/*** Some low level helpers **************************************************/

/* 10-bit mu-law */
static unsigned short ulaw2dma10[] = {
      10,   26,   42,   58,   74,   90,  106,  122,
     138,  154,  170,  186,  202,  218,  234,  250,
     262,  270,  278,  286,  294,  302,  310,  318,
     326,  334,  342,  350,  358,  366,  374,  382,
     388,  392,  396,  400,  404,  408,  412,  416,
     420,  424,  428,  432,  436,  440,  444,  448,
     451,  453,  455,  457,  459,  461,  463,  465,
     467,  469,  471,  473,  475,  477,  479,  481,
     482,  483,  484,  485,  486,  487,  488,  489,
     490,  491,  492,  493,  494,  495,  496,  497,
     498,  498,  499,  499,  500,  500,  501,  501,
     502,  502,  503,  503,  504,  504,  505,  505,
     506,  506,  506,  506,  507,  507,  507,  507,
     508,  508,  508,  508,  509,  509,  509,  509,
     510,  510,  510,  510,  510,  510,  510,  510,
     511,  511,  511,  511,  511,  511,  511,  511,
    1013,  997,  981,  965,  949,  933,  917,  901,
     885,  869,  853,  837,  821,  805,  789,  773,
     761,  753,  745,  737,  729,  721,  713,  705,
     697,  689,  681,  673,  665,  657,  649,  641,
     635,  631,  627,  623,  619,  615,  611,  607,
     603,  599,  595,  591,  587,  583,  579,  575,
     572,  570,  568,  566,  564,  562,  560,  558,
     556,  554,  552,  550,  548,  546,  544,  542,
     541,  540,  539,  538,  537,  536,  535,  534,
     533,  532,  531,  530,  529,  528,  527,  526,
     525,  525,  524,  524,  523,  523,  522,  522,
     521,  521,  520,  520,  519,  519,  518,  518,
     517,  517,  517,  517,  516,  516,  516,  516,
     515,  515,  515,  515,  514,  514,  514,  514,
     513,  513,  513,  513,  513,  513,  513,  512,
     512,  512,  512,  512,  512,  512,  512,  511
};

/* 10-bit A-law */
static unsigned short alaw2dma10[] = {
     425,  429,  417,  421,  441,  445,  433,  437,
     393,  397,  385,  389,  409,  413,  401,  405,
     468,  470,  464,  466,  476,  478,  472,  474,
     452,  454,  448,  450,  460,  462,  456,  458,
     167,  183,  135,  151,  231,  247,  199,  215,
      39,   55,    7,   23,  103,  119,   71,   87,
     339,  347,  323,  331,  371,  379,  355,  363,
     275,  283,  259,  267,  307,  315,  291,  299,
     506,  506,  506,  506,  507,  507,  507,  507,
     504,  504,  504,  504,  505,  505,  505,  505,
     510,  510,  510,  510,  511,  511,  511,  511,
     508,  508,  508,  508,  509,  509,  509,  509,
     490,  491,  488,  489,  494,  495,  492,  493,
     482,  483,  480,  481,  486,  487,  484,  485,
     501,  501,  500,  500,  503,  503,  502,  502,
     497,  497,  496,  496,  499,  499,  498,  498,
     597,  593,  605,  601,  581,  577,  589,  585,
     629,  625,  637,  633,  613,  609,  621,  617,
     554,  552,  558,  556,  546,  544,  550,  548,
     570,  568,  574,  572,  562,  560,  566,  564,
     855,  839,  887,  871,  791,  775,  823,  807,
     983,  967, 1015,  999,  919,  903,  951,  935,
     683,  675,  699,  691,  651,  643,  667,  659,
     747,  739,  763,  755,  715,  707,  731,  723,
     517,  517,  517,  517,  516,  516,  516,  516,
     519,  519,  519,  519,  518,  518,  518,  518,
     513,  513,  513,  513,  512,  512,  512,  512,
     515,  515,  515,  515,  514,  514,  514,  514,
     533,  532,  535,  534,  529,  528,  531,  530,
     541,  540,  543,  542,  537,  536,  539,  538,
     522,  522,  523,  523,  520,  520,  521,  521,
     526,  526,  527,  527,  524,  524,  525,  525
};

/*** Translations ************************************************************/
static ssize_t vrc4173_ct_law(const u_char *userPtr, size_t userCount,
			  u_char frame[], ssize_t *frameUsed,
			  ssize_t frameLeft);
static ssize_t vrc4173_ct_s8(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed,
			 ssize_t frameLeft);
static ssize_t vrc4173_ct_u8(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed,
			 ssize_t frameLeft);

/*** Machine definitions *****************************************************/


typedef struct {
	int type;
	void *(*dma_alloc)(unsigned int, int);
	void (*dma_free)(void *, unsigned int);
	int (*irqinit)(void);
#ifdef MODULE
	void (*irqcleanup)(void);
#endif /* MODULE */
	void (*init)(void);
	void (*silence)(void);
	int (*setFormat)(int);
	int (*setVolume)(int);
	int (*setBass)(int);
	int (*setTreble)(int);
	int (*setGain)(int);
	void (*play)(void);
} MACHINE;


/*** Low level stuff *********************************************************/


typedef struct {
	int format;		/* AFMT_* */
	int stereo;		/* 0 = mono, 1 = stereo */
	int size;		/* 8/16 bit*/
	int speed;		/* speed */
} SETTINGS;

typedef struct {
	ssize_t (*ct_ulaw)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_alaw)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_s8)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_u8)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_s16be)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_u16be)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_s16le)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
	ssize_t (*ct_u16le)(const u_char *, size_t, u_char *, ssize_t *, ssize_t);
} TRANS;

struct sound_settings {
	MACHINE mach;		/* machine dependent things */
	SETTINGS hard;		/* hardware settings */
	SETTINGS soft;		/* software settings */
	SETTINGS dsp;		/* /dev/dsp default settings */
	TRANS *trans;		/* supported translations */
	int volume_left;	/* volume (range is machine dependent) */
	int volume_right;
	int bass;		/* tone (range is machine dependent) */
	int treble;
	int gain;
	int minDev;		/* minor device number currently open */
};

static struct sound_settings sound;

static int VRC4173IrqInit(void);
#ifdef MODULE
static void VRC4173IrqCleanUp(void);
#endif /* MODULE */
static void VRC4173Silence(void);
static void VRC4173Init(void);
static int VRC4173SetFormat(int format);
/* static int VRC4173SetVolume(int volume); */
/* static int VRC4173SetTreble(int treble); */
static void vrc4173_sq_play_next_frame(int index);
static void VRC4173Play(void);
static void vrc4173_sq_interrupt(int irq, void *dummy, struct pt_regs *fp);


/*** Mid level stuff *********************************************************/


static void sound_silence(void);
static void sound_init(void);
static int sound_set_format(int format);
static int sound_set_speed(int speed);
static int sound_set_stereo(int stereo);
// static int sound_set_volume(int volume);
static ssize_t sound_copy_translate(const u_char *userPtr,
				    size_t userCount,
				    u_char frame[], ssize_t *frameUsed,
				    ssize_t frameLeft);
/*
 * /dev/mixer abstraction
 */

struct sound_mixer {
    int busy;
    int modify_counter;
};

static struct sound_mixer mixer;

/*
 * Sound queue stuff, the heart of the driver
 */

struct sound_queue {
	int max_count, block_size;
	char **buffers;
	int max_active;

	/* it shouldn't be necessary to declare any of these volatile */
	int front, rear, count;
	int rear_size;
	/*
	 *	The use of the playing field depends on the hardware
	 *
	 *	Atari, PMac: The number of frames that are loaded/playing
	 *
	 *	Amiga: Bit 0 is set: a frame is loaded
	 *	       Bit 1 is set: a frame is playing
	 */
	int active;
	wait_queue_head_t action_queue, open_queue, sync_queue;
	int open_mode;
	int busy, syncing;
	u_short wake_up_threshold;
};

static struct sound_queue sq;

#define sq_block_address(i)	(sq.buffers[i])
#define SIGNAL_RECEIVED	(signal_pending(current))
#define NON_BLOCKING(open_mode)	(open_mode & O_NONBLOCK)
#define ONE_SECOND	HZ	/* in jiffies (100ths of a second) */
#define NO_TIME_LIMIT	0xffffffff
#define SLEEP(queue, time_limit) \
	interruptible_sleep_on_timeout(&queue, (time_limit));
#define WAKE_UP(queue)	(wake_up_interruptible(&queue))

/*
 * /dev/sndstat
 */

struct sound_state {
	int busy;
	char buf[512];
	int len, ptr;
};

static struct sound_state state;

/*** Common stuff ********************************************************/

static loff_t sound_lseek(struct file *file, long long offset, int orig);
static inline int ioctl_return(int *addr, int value)
{
	if (value < 0)
		return(value);

	return put_user(value, addr);
}


/*** Config & Setup **********************************************************/


void dmasound_init(void);
void dmasound_setup(char *str, int *ints);


/*** Translations ************************************************************/


/* ++TeSche: radically changed for new expanding purposes...
 *
 * These two routines now deal with copying/expanding/translating the samples
 * from user space into our buffer at the right frequency. They take care about
 * how much data there's actually to read, how much buffer space there is and
 * to convert samples into the right frequency/encoding. They will only work on
 * complete samples so it may happen they leave some bytes in the input stream
 * if the user didn't write a multiple of the current sample size. They both
 * return the number of bytes they've used from both streams so you may detect
 * such a situation. Luckily all programs should be able to cope with that.
 *
 * I think I've optimized anything as far as one can do in plain C, all
 * variables should fit in registers and the loops are really short. There's
 * one loop for every possible situation. Writing a more generalized and thus
 * parameterized loop would only produce slower code. Feel free to optimize
 * this in assembler if you like. :)
 *
 * I think these routines belong here because they're not yet really hardware
 * independent, especially the fact that the Falcon can play 16bit samples
 * only in stereo is hardcoded in both of them!
 *
 * ++geert: split in even more functions (one per format)
 */

static inline void vrc4173_fill(u_short d, u_short *p, int n)
{
	while (n--)
		*p++ = d;
}

static ssize_t vrc4173_ct_law(const u_char *userPtr, size_t userCount,
			  u_char frame[], ssize_t *frameUsed,
			  ssize_t frameLeft)
{
	u_short *table =
		(sound.soft.format == AFMT_MU_LAW) ? ulaw2dma10 : alaw2dma10;
	ssize_t count, uUsed, fUsed;
	u_short *p = (u_short *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 1, frameLeft >> 1);
		uUsed = count * 2;
		fUsed = count * 2;
	} else {
		count = min(userCount, frameLeft >> 1);
		uUsed = count;
		fUsed = count * 2;
	}
	while (count > 0) {
		u_char data;
		u_short d;
		if (get_user(data, userPtr++))
			return -EFAULT;
		d = data;
		if (sound.hard.stereo) {
			// Translate stereo -> mono
			if (get_user(data, userPtr++))
				return -EFAULT;
			d += data;
			d >>= 1;	// take average of R/L ch
		}
		*p++ = ((u_short)table[d]);
		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	vrc4173_fill(0x0200, p, (frameLeft - fUsed) >> 1);

	return(uUsed);
}


static ssize_t vrc4173_ct_s8(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;

	u_short *p = (u_short *)(&frame[*frameUsed]);
	if (sound.hard.stereo) {
		count = min(userCount >> 1, frameLeft >> 1);
		uUsed = count * 2;
		fUsed = count * 2;
	} else {
		count = min(userCount, frameLeft >> 1);
		uUsed = count;
		fUsed = count * 2;
	}
	while (count > 0) {
		u_char data;
		u_short d;
		if (get_user(data, userPtr++))
			return -EFAULT;
		d = data;
		if (sound.hard.stereo) {
			// Translate stereo -> mono
			if (get_user(data, userPtr++))
				return -EFAULT;
			d += data;
			d >>= 1;	// take average of R/L ch
		}
		*p++ = (d << 2) ^ 0x0200;
		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	vrc4173_fill(0x0200, p, (frameLeft - fUsed) >> 1);

	return (uUsed);
}


static ssize_t vrc4173_ct_u8(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;

	u_short  *p = (u_short *)(&frame[*frameUsed]);
	if (sound.hard.stereo) {
		count = min(userCount >> 1, frameLeft >> 1);
		uUsed = count * 2;
		fUsed = count * 2;
	} else {
		count = min(userCount, frameLeft >> 1);
		uUsed = count;
		fUsed = count * 2;
	}
	while (count > 0) {
		u_char data;
		u_short d;
		if (get_user(data, userPtr++))
			return -EFAULT;
		d = data;
		if (sound.hard.stereo) {
			// Translate stereo -> mono
			if (get_user(data, userPtr++))
				return -EFAULT;
			d += data;
			d >>= 1;	// take average of R/L ch
		}
		*p++ = d << 2;	// make it 10-bit
		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	vrc4173_fill(0x0200, p, (frameLeft - fUsed) >> 1);

	return (uUsed);
}

static ssize_t vrc4173_ct_s16le(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	// uUsed & fUsed is byte count!!
	ssize_t count, uUsed, fUsed;
	u_short  *p = (u_short *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 2, frameLeft >> 1);
		uUsed = count * 4;
		fUsed = count * 2;
	} else {
		count = min(userCount >> 1, frameLeft >> 1);
		uUsed = count * 2;
		fUsed = count * 2;
	}
	while (count > 0) {
		u_short data;
		u_int d;
		if (get_user(data, ((u_short *)userPtr)++))
			return -EFAULT;
		d = data ^ 0x8000;
		if (sound.hard.stereo) {
			if (get_user(data, ((u_short *)userPtr)++))
				return -EFAULT;
			d += data ^ 0x8000;
			d >>= 1;	// take average of R/L ch
		}
		*p++ = d >> 6;	// 16bit -> 10 bit
		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	vrc4173_fill(0x0200, p, (frameLeft - fUsed) >> 1);

	return (uUsed);
}

static ssize_t vrc4173_ct_u16le(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;
	u_short  *p = (u_short *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 2, frameLeft >> 1);
		uUsed = count * 4;
		fUsed = count * 2;
	} else {
		count = min(userCount >> 1, frameLeft >> 1);
		uUsed = count * 2;
		fUsed = count * 2;
	}
	while (count > 0) {
		u_short data;
		u_int d;
		if (get_user(data, ((u_short *)userPtr)++))
			return -EFAULT;
		d = data;
		if (sound.hard.stereo) {
			if (get_user(data, ((u_short *)userPtr)++))
				return -EFAULT;
			d += data;
			d >>= 1;	// take average of R/L ch
		}
		*p++ = d >> 6;	// 16bit -> 10 bit
		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	vrc4173_fill(0x0200, p, (frameLeft - fUsed) >> 1);

	return(uUsed);
}

/* Convert Big Endian short -> Little Endian short */
static inline u_short be2le(u_short d)
{
	return (((d << 8) & 0xff00) + ((d >>8) & 0x00ff));
}

static ssize_t vrc4173_ct_s16be(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;
	u_short  *p = (u_short *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 2, frameLeft >> 1);
		uUsed = count * 4;
		fUsed = count * 2;
	} else {
		count = min(userCount >> 1, frameLeft >> 1);
		uUsed = count * 2;
		fUsed = count * 2;
	}
	while (count > 0) {
		u_short data;
		u_int d;
		if (get_user(data, ((u_short *)userPtr)++))
			return -EFAULT;
		d = be2le(data) ^ 0x8000;
		if (sound.hard.stereo) {
			if (get_user(data, ((u_short *)userPtr)++))
				return -EFAULT;
			d += be2le(data) ^ 0x8000;
			d >>= 1;	// take average of R/L ch
		}
		*p++ = d >> 6;	// 16bit -> 10 bit
		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	vrc4173_fill(0x0200, p, (frameLeft - fUsed) >> 1);

	return(uUsed);
}

static ssize_t vrc4173_ct_u16be(const u_char *userPtr, size_t userCount,
			 u_char frame[], ssize_t *frameUsed, ssize_t frameLeft)
{
	ssize_t count, uUsed, fUsed;
	u_short  *p = (u_short *)(&frame[*frameUsed]);

	if (sound.hard.stereo) {
		count = min(userCount >> 2, frameLeft >> 1);
		uUsed = count * 4;
		fUsed = count * 2;
	} else {
		count = min(userCount >> 1, frameLeft >> 1);
		uUsed = count * 2;
		fUsed = count * 2;
	}
	while (count > 0) {
		u_short data;
		u_int d;
		if (get_user(data, ((u_short *)userPtr)++))
			return -EFAULT;
		d = be2le(data);
		if (sound.hard.stereo) {
			if (get_user(data, ((u_short *)userPtr)++))
				return -EFAULT;
			d += be2le(data);
			d >>= 1;	// take average
		}
		*p++ = d >> 6;	// 16bit -> 10 bit
		count--;
	}
	*frameUsed += fUsed;
	// Clear rest of frame
	vrc4173_fill(0x0200, p, (frameLeft - fUsed) >> 1);

	return (uUsed);
}

static TRANS transVRC4173 = {
	vrc4173_ct_law, vrc4173_ct_law, vrc4173_ct_s8, vrc4173_ct_u8,
	vrc4173_ct_s16be, vrc4173_ct_u16be, vrc4173_ct_s16le, vrc4173_ct_u16le
};

/*** Low level stuff *********************************************************/

static void vrc4173_set_rate(void)
{
    unsigned short rate;

    switch (sound.soft.speed) {
    case 8000:
	rate = 0x4;
	break;
    case 44100:
	rate = 0x2;
	break;
    case 22050:
	rate = 0x1;
	break;
    case 11025:
	rate = 0x0;
	break;
    default:
	rate = 0x4;	// default to 8K
	printk("DMA sound: Illegal sound rate %d\n", sound.soft.speed);
	break;
    }
    outw(rate, pci_resource_start (vrc4173_dev, 0) + VRC4173_SCNVRREG);
}

static void vrc4173_enable_irq(void)
{
	outw(0x8, pci_resource_start (vrc4173_dev, 0) + VRC4173_MAIUINTREG);	// Enable Audio Output IRQs (DMA page 2)
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_MSYSINT1REG, 0xffff, 1<<6);
}

static void vrc4173_disable_irq(void)
{
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_MSYSINT1REG, ~(1<<6), 0);
	outw(0x0, pci_resource_start (vrc4173_dev, 0) + VRC4173_MAIUINTREG);	// Disable Audio Output IRQs
}

static void vrc4173_clear_irq(void)
{
	// Clear All Audio Output IRQs
	outw(0xe, pci_resource_start (vrc4173_dev, 0) + VRC4173_INTREG);
	barrier();
}

static void vrc4173_enable_dma(void)
{
	// Enable DMA in DCU
        while (!inw(pci_resource_start (vrc4173_dev, 0) + VRC4173_DMAIDLEREG))
	    ;
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_DMARSTREG, 0xffff, 0x1);
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_DMAMSKREG, 0xffff, 0x4);
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_DMASENREG, 0xffff, 0x1);
	// But DMA not started yet, need 'AUISEN := 1'
}

static void vrc4173_disable_dma(void)
{
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_DMAMSKREG, 0xfffb, 0);
}

static void vrc4173_resume_dma(void)
{
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_DMAMSKREG, 0xffff, 0x4);
}

static void vrc4173_speaker_on(void)
{
    ac97_speaker_on(ac97_dev);
}

static void vrc4173_disable_sound(void)
{
    vrc4173_disable_dma();
    vrc4173_clear_irq();
    // vrc4173_disable_irq();

    /* Speaker operation disable */
    outw(0x0, pci_resource_start (vrc4173_dev, 0) + VRC4173_SEQREG);

    /* Speaker power OFF */
    ac97_speaker_off(ac97_dev);

    /* DAC vref Off */
    vrc4173_and_or_reg(vrc4173_dev, VRC4173_SCNTREG, 0x7fff, 0);

    /* Disable AIU clock in CMU */
//    vrc4173_turn_off_device(vrc4173_dev, AIU_DEV);
}

static void vrc4173_kick_dma(void)
{
    /* conversion rate */
    vrc4173_set_rate();

    /* Enable DMA */
    vrc4173_enable_dma();

    /* DAC Vref on */
    vrc4173_and_or_reg(vrc4173_dev, VRC4173_SCNTREG, 0xffff, 0x8000);

    /* Wait for Vref stablization */
    // udelay(10);
    mdelay(50);	// 50 msec is enough ?

    // Speaker Power On
    vrc4173_speaker_on();

    // Speaker operation enable
    outw(0x1, pci_resource_start (vrc4173_dev, 0) + VRC4173_SEQREG);
}


/*
 * VRC4173
 */
static int __init VRC4173IrqInit(void)
{
	/* turn off DMA for audio channels */
	vrc4173_disable_dma();
	vrc4173_clear_irq();
	vrc4173_disable_sound();
	return(1);
}

void vrc4173_handle_audio(struct pci_dev *pdev, int index)
{
    vrc4173_sq_interrupt(VR41XX_IRQ_4173, pdev, NULL);
}

#ifdef MODULE
static void VRC4173IrqCleanUp(void)
{
	/* turn off DMA for audio channels */
	vrc4173_disable_dma();
	vrc4173_disable_irq();
	vrc4173_clear_irq();
	vrc4173_disable_sound();
}
#endif /* MODULE */

static void VRC4173Silence(void)
{
	/* Disable sound & DMA */
	vrc4173_disable_sound();
	vrc4173_disable_irq();
}

static char vrc4173_first_dma_kick = 1;

static void VRC4173Init(void)
{
	sound.hard = sound.soft;
	sound.trans = &transVRC4173;

	vrc4173_first_dma_kick = 1;

	/* Enable AIU clock in CMU */
	vrc4173_turn_on_device(vrc4173_dev, AIU_DEV);
	vrc4173_enable_irq();

}

static int VRC4173SetFormat(int format)
{
	int size;

	/* VRC4173 sound DMA supports only 8bit (expanded to 10bit mode */
	switch (format) {
	case AFMT_QUERY:
		return(sound.soft.format);
	case AFMT_MU_LAW:
		size = 8;
		break;
	case AFMT_A_LAW:
		size = 8;
		break;
	case AFMT_U8:
		size = 8;
		break;
	case AFMT_S8:
		size = 8;
		break;
	case AFMT_S16_BE:
		size = 16;
		break;
	case AFMT_U16_BE:
		size = 16;
		break;
	case AFMT_S16_LE:
		size = 16;
		break;
	case AFMT_U16_LE:
		size = 16;
		break;
	default: /* :-) */
#ifdef DEBUG
		printk("default: AFMT_S8\n");
#endif
		size = 8;
		format = AFMT_S8;
	}

	sound.soft.format = format;
	sound.soft.size = size;
	if (sound.minDev == SND_DEV_DSP) {
		sound.dsp.format = format;
		sound.dsp.size = sound.soft.size;
	}

	return(format);
}

#define VRC4173_PLAY_LOADED		1
#define VRC4173_PLAY_PLAYING		2
#define VRC4173_PLAY_MASK		3


static void vrc4173_sq_play_next_frame(int index)
{
	u_char *start;
	u_long size;
	u_int ladr, hadr;


	/* used by VRC4173Play() if all doubts whether there really is something
	 * to be played are already wiped out.
	 */
	start = (u_char *)sq_block_address(sq.front);
	size = (sq.count == index ? sq.rear_size : sq.block_size)>>1;

	// Setup DMA address
	ladr = (u_int)start & 0xffff;
	// Mask upper 7bit (only lower 32M byte can set to DMA adr reg.)
	hadr = ((u_int)start >> 16) & 0x01ff;
	outw(ladr, pci_resource_start (vrc4173_dev, 0) + VRC4173_AIUOBALREG);
	outw(hadr, pci_resource_start (vrc4173_dev, 0) + VRC4173_AIUOBAHREG);
	barrier();
	if (vrc4173_first_dma_kick) {
		// On first DMA, DMA address reg & DMA base reg should be set
	    outw(ladr, pci_resource_start (vrc4173_dev, 0) + VRC4173_AIUOALREG);
	    outw(hadr, pci_resource_start (vrc4173_dev, 0) + VRC4173_AIUOAHREG);

	}
	vrc4173_clear_irq();
	sq.front = (sq.front+1) % sq.max_count;
	sq.active |= VRC4173_PLAY_LOADED;

	vrc4173_resume_dma();

	if (vrc4173_first_dma_kick) {
		vrc4173_kick_dma();
		vrc4173_first_dma_kick = 0;
	}
}

static void VRC4173Play(void)
{
	// fixme
	int minframes = 1;
	vrc4173_disable_irq();

	if (sq.active & VRC4173_PLAY_LOADED) {
		/* There's already a frame loaded */
		vrc4173_enable_irq();
		return;
	}

	if (sq.active & VRC4173_PLAY_PLAYING)
		/* Increase threshold: frame 1 is already being played */
		minframes = 2;

	if (sq.count < minframes) {
		/* Nothing to do */
		vrc4173_enable_irq();
		return;
	}
	if (sq.count <= minframes && sq.rear_size < sq.block_size && !sq.syncing) {
		/* hmmm, the only existing frame is not
		 * yet filled and we're not syncing?
		 */
		vrc4173_enable_irq();
		return;
	}

	vrc4173_sq_play_next_frame(minframes);

	vrc4173_enable_irq();
}

static void vrc4173_sq_interrupt(int irq, void *dummy, struct pt_regs *fp)
{
	int minframes = 1;

	// vrc4173_clear_irq();
	if (!sq.active) {
		/* Playing was interrupted and sq_reset() has already cleared
		 * the sq variables, so better don't do anything here.
		 */
		WAKE_UP(sq.sync_queue);
		vrc4173_clear_irq();
		return;
	}

	if (sq.active & VRC4173_PLAY_PLAYING) {
		/* We've just finished a frame */
		sq.count--;
		if (sq.count < sq.wake_up_threshold)
			WAKE_UP(sq.action_queue);
	}

	if (sq.active & VRC4173_PLAY_LOADED)
		/* Increase threshold: frame 1 is already being played */
		minframes = 2;

	/* Shift the flags */
	sq.active = (sq.active << 1) & VRC4173_PLAY_MASK;

	if (!sq.active) {
		/* No frame is playing, disable audio DMA */
		vrc4173_clear_irq();
/* 000315 kei -- to avoid kernel hang up when heavy load...
		vrc4173_suspend_dma();
*/
	}

	if (sq.count >= minframes) {
		/* Try to play the next frame */
		VRC4173Play();
	}

	if (!sq.active) {
		/* Nothing to play anymore.
		   Wake up a process waiting for audio output to drain. */
		WAKE_UP(sq.sync_queue);
	}
}


/*** Machine definitions *****************************************************/

static MACHINE machVRC4173 = {
	DMASND_VRC4173,		// int type
	NULL,				// void *dma_alloc(uint, int)
	NULL,				// void  dma_free(void *, uint)
	VRC4173IrqInit,		// void  irqinit(void)
#ifdef MODULE
	VRC4173IrqCleanUp,	// void  irqcleanup(void)
#endif /* MODULE */
	VRC4173Init,			// void  init(void)
	VRC4173Silence,		// void  silence(void)
	VRC4173SetFormat,	// int   setFormat(int)
	NULL,				// int   setVolume(int)
	NULL,				// int   setBass(int)
	NULL,				// int   setTreble(int)
	NULL,				// int   setGain(int)
	VRC4173Play			// void  play(void)
};

/*** Mid level stuff *********************************************************/


static void sound_silence(void)
{
	/* update hardware settings one more */
	(*sound.mach.init)();

	(*sound.mach.silence)();
}


static void sound_init(void)
{
	(*sound.mach.init)();
}


static int sound_set_format(int format)
{
	return(*sound.mach.setFormat)(format);
}


static int sound_set_speed(int speed)
{
	if (speed < 0)
		return(sound.soft.speed);

	sound.soft.speed = speed;
	(*sound.mach.init)();
	if (sound.minDev == SND_DEV_DSP)
		sound.dsp.speed = sound.soft.speed;

	return(sound.soft.speed);
}


static int sound_set_stereo(int stereo)
{
	if (stereo < 0)
		return(sound.soft.stereo);
	stereo = !!stereo;    /* should be 0 or 1 now */

	sound.soft.stereo = stereo;
	if (sound.minDev == SND_DEV_DSP)
		sound.dsp.stereo = stereo;
	(*sound.mach.init)();

	return(stereo);
}

static ssize_t sound_copy_translate(const u_char *userPtr,
				    size_t userCount,
				    u_char frame[], ssize_t *frameUsed,
				    ssize_t frameLeft)
{
	ssize_t
		(*ct_func)(const u_char *, size_t, u_char *, ssize_t *, ssize_t) = NULL;

	switch (sound.soft.format) {
	case AFMT_MU_LAW:
		ct_func = sound.trans->ct_ulaw;
		break;
	case AFMT_A_LAW:
		ct_func = sound.trans->ct_alaw;
		break;
	case AFMT_S8:
		ct_func = sound.trans->ct_s8;
		break;
	case AFMT_U8:
		ct_func = sound.trans->ct_u8;
		break;
	case AFMT_S16_BE:
		ct_func = sound.trans->ct_s16be;
		break;
	case AFMT_U16_BE:
		ct_func = sound.trans->ct_u16be;
		break;
	case AFMT_S16_LE:
		ct_func = sound.trans->ct_s16le;
		break;
	case AFMT_U16_LE:
		ct_func = sound.trans->ct_u16le;
		break;
	}
	if (ct_func)
		return ct_func(userPtr, userCount, frame, frameUsed, frameLeft);
	else
		return 0;
}


/*
 * /dev/mixer abstraction
 */


#define RECLEVEL_VOXWARE_TO_GAIN(v) \
	((v) < 0 ? 0 : (v) > 100 ? 15 : (v) * 3 / 20)
#define RECLEVEL_GAIN_TO_VOXWARE(v) (((v) * 20 + 2) / 3)


static int mixer_open(struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	mixer.busy = 1;
	return 0;
}


static int mixer_release(struct inode *inode, struct file *file)
{
	mixer.busy = 0;
	MOD_DEC_USE_COUNT;
	vrc4173_disable_sound();
	return 0;
}


static int mixer_ioctl(struct inode *inode, struct file *file, u_int cmd,
		       u_long arg)
{
	// int data;
	if (_SIOC_DIR(cmd) & _SIOC_WRITE)
	    mixer.modify_counter++;
	if (cmd == OSS_GETVERSION)
	    return IOCTL_OUT(arg, SOUND_VERSION);
	switch (sound.mach.type) {
	case DMASND_VRC4173:
		switch (cmd) {
		case SOUND_MIXER_INFO: {
		    mixer_info info;
		    strncpy(info.id, "VRC4173", sizeof(info.id));
		    strncpy(info.name, "VRC4173", sizeof(info.name));
		    info.name[sizeof(info.name)-1] = 0;
		    info.modify_counter = mixer.modify_counter;
		    if (copy_to_user((int *)arg, &info, sizeof(info)))
			return -EFAULT;
		    return 0;
		}
		case SOUND_MIXER_READ_DEVMASK:
			return IOCTL_OUT(arg, 0);	// fixme
		case SOUND_MIXER_READ_RECMASK:
			return IOCTL_OUT(arg, 0);
		case SOUND_MIXER_READ_STEREODEVS:
			return IOCTL_OUT(arg, 0);
		case SOUND_MIXER_READ_VOLUME:
			return IOCTL_OUT(arg, 0);	// fixme
		case SOUND_MIXER_WRITE_VOLUME:
			return IOCTL_OUT(arg, 0);	// fixme
		case SOUND_MIXER_READ_TREBLE:
			return IOCTL_OUT(arg, 0);	// fixme
		case SOUND_MIXER_WRITE_TREBLE:
			return IOCTL_OUT(arg, 0);	// fixme
		}
		break;
	}

	return -EINVAL;
}


static struct file_operations mixer_fops =
{
	llseek:  sound_lseek,
	ioctl:   mixer_ioctl,
	open:    mixer_open,
	release: mixer_release
};

static void __init mixer_init(void)
{
#ifndef MODULE
	int mixer_unit;
#endif
	mixer_unit = register_sound_mixer(&mixer_fops, -1);
	if (mixer_unit < 0)
		return;

	mixer.busy = 0;
	sound.treble = 0;
	sound.bass = 0;
	switch (sound.mach.type) {
    case DMASND_VRC4173:
		sound.volume_left = 64;
		sound.volume_right = 64;
		// nothing to do ?? fixme
		break;
	}
}


/*
 * Sound queue stuff, the heart of the driver
 */


static int sq_allocate_buffers(void)
{
	int i;

	if (sound_buffers)
		return 0;
	sound_buffers = kmalloc (numBufs * sizeof(char *), GFP_DMA);
	if (!sound_buffers)
		return -ENOMEM;
	vrc4173_dma_bufs = kmalloc(numBufs * sizeof(char *), GFP_DMA);
	if (!vrc4173_dma_bufs) {
		kfree(sound_buffers);
		sound_buffers = 0;
		return -ENOMEM;
	}
	for (i = 0; i < numBufs; i++) {
		/* need to keep 2K byte boundary */
		/* First, try allocating 2K byte */
		vrc4173_dma_bufs[i] = kmalloc(2 * 1024, GFP_DMA);
		if (vrc4173_dma_bufs[i] == NULL) {
			// Fail to allocate
			while (i--)
				kfree(vrc4173_dma_bufs[i]);
			kfree(sound_buffers);
			kfree(vrc4173_dma_bufs);
			sound_buffers = vrc4173_dma_bufs = 0;
			return -ENOMEM;
		} else {
			if (((u_long)vrc4173_dma_bufs[i] & 0x7ff) == 0) {
				// Good, allocated on 2K byte boundary
				// Map to KSEG1
				sound_buffers[i] = (char *)
					((u_long)vrc4173_dma_bufs[i] | KSEG1);
			} else {
				// not allocated on 2K byte boundary
				// So, try allocating 4K byte
				kfree(vrc4173_dma_bufs[i]);
				vrc4173_dma_bufs[i] = kmalloc(4 * 1024, GFP_DMA);
				/*
					1) round to 2K byte boundary
					2) Map to KSEG1 (not TLB mapped, not cached)
					   (Because DMA access these area)
				*/
				sound_buffers[i] = (char *)
				((((u_long)vrc4173_dma_bufs[i] & 0xfffff800) | KSEG1) + 0x0800);
			}
		}
	}
	return 0;
}


static void sq_release_buffers(void)
{
	int i;

	if (vrc4173_dma_bufs) {
		for (i = 0; i < numBufs; i++)
			kfree(vrc4173_dma_bufs[i]);
		kfree (sound_buffers);
		kfree (vrc4173_dma_bufs);
		sound_buffers = vrc4173_dma_bufs = 0;
	}
}

static void sq_setup(int numBufs, int bufSize, char **write_buffers)
{
	sq.max_count = numBufs;
	sq.max_active = numBufs;
	sq.block_size = bufSize;
	sq.buffers = write_buffers;

	sq.front = sq.count = 0;
	sq.rear = -1;
	sq.syncing = 0;
	sq.active = 0;
	sq.wake_up_threshold = numBufs / 4;
}



static void sq_play(void)
{
	(*sound.mach.play)();
}


/* ++TeSche: radically changed this one too */

static ssize_t sq_write(struct file *file, const char *src, size_t uLeft,
			loff_t *ppos)
{
	ssize_t uWritten = 0;
	u_char *dest;
	ssize_t uUsed, bUsed, bLeft;

	/* ++TeSche: Is something like this necessary?
	 * Hey, that's an honest question! Or does any other part of the
	 * filesystem already checks this situation? I really don't know.
	 */
	if (uLeft == 0)
		return 0;

	/* The interrupt doesn't start to play the last, incomplete frame.
	 * Thus we can append to it without disabling the interrupts! (Note
	 * also that sq.rear isn't affected by the interrupt.)
	 */
	if (sq.count > 0 && (bLeft = sq.block_size-sq.rear_size) > 0) {
		dest = sq_block_address(sq.rear);
		bUsed = sq.rear_size;
		uUsed = sound_copy_translate(src, uLeft, dest, &bUsed, bLeft);
		if (uUsed <= 0)
			return uUsed;
		src += uUsed;
		uWritten += uUsed;
		uLeft -= uUsed;
		sq.rear_size = bUsed;
	}

	do {
		while (sq.count == sq.max_active) {
			/* All frames are in use.  -- kei */
			sq_play();
			if (NON_BLOCKING(sq.open_mode))
				return uWritten > 0 ? uWritten : -EAGAIN;
			SLEEP(sq.action_queue, ONE_SECOND);
			if (SIGNAL_RECEIVED)
				return uWritten > 0 ? uWritten : -EINTR;
		}

		/* Here, we can avoid disabling the interrupt by first
		 * copying and translating the data, and then updating
		 * the sq variables. Until this is done, the interrupt
		 * won't see the new frame and we can work on it
		 * undisturbed.
		 */

		dest = sq_block_address((sq.rear+1) % sq.max_count);
		bUsed = 0;
		bLeft = sq.block_size;
		uUsed = sound_copy_translate(src, uLeft, dest, &bUsed, bLeft);
		if (uUsed <= 0)
			break;
		src += uUsed;
		uWritten += uUsed;
		uLeft -= uUsed;
		if (bUsed) {
			// Need to be atomic -- kei
			vrc4173_disable_irq();
			sq.rear = (sq.rear+1) % sq.max_count;
			sq.rear_size = bUsed;
			sq.count++;
			vrc4173_enable_irq();
		}
	} while (bUsed);   /* uUsed may have been 0 */
	sq_play();
	return uUsed < 0? uUsed: uWritten;
}

/***********/


static int sq_open(struct inode *inode, struct file *file)
{
	int rc = 0;

	MOD_INC_USE_COUNT;
	if (file->f_mode & FMODE_WRITE) {
		if (sq.busy) {
			rc = -EBUSY;
			if (NON_BLOCKING(file->f_flags))
				goto err_out;
			rc = -EINTR;
			while (sq.busy) {
				SLEEP(sq.open_queue, ONE_SECOND);
				if (SIGNAL_RECEIVED)
					goto err_out;
			}
		}
		sq.busy = 1; /* Let's play spot-the-race-condition */

		if (sq_allocate_buffers()) goto err_out_nobusy;

		sq_setup(numBufs, bufSize<<10,sound_buffers);
		sq.open_mode = file->f_mode;
	}
	sound.minDev = MINOR(inode->i_rdev) & 0x0f;
	sound.soft = sound.dsp;
	sound.hard = sound.dsp;
	sound_init();
	if ((MINOR(inode->i_rdev) & 0x0f) == SND_DEV_AUDIO) {
		sound_set_speed(44100);
		sound_set_stereo(0);
		sound_set_format(AFMT_S8);
	}

	return 0;

err_out_nobusy:
	if (file->f_mode & FMODE_WRITE) {
		sq.busy = 0;
		WAKE_UP(sq.open_queue);
	}
err_out:
	MOD_DEC_USE_COUNT;
	return rc;
}


static void sq_reset(void)
{
	sound_silence();
	sq.active = 0;
	sq.count = 0;
	sq.front = (sq.rear+1) % sq.max_count;
}


static int sq_fsync(struct file *filp, struct dentry *dentry)
{
	int rc = 0;

	sq.syncing = 1;
	sq_play();	/* there may be an incomplete frame waiting */

	while (sq.active) {
		SLEEP(sq.sync_queue, ONE_SECOND);
		if (SIGNAL_RECEIVED) {
			/* While waiting for audio output to drain, an
			 * interrupt occurred.  Stop audio output immediately
			 * and clear the queue. */
			sq_reset();
			rc = -EINTR;
			break;
		}
	}

	sq.syncing = 0;
	return rc;
}

static int sq_release(struct inode *inode, struct file *file)
{
	int rc = 0;

	if (sq.busy)
		rc = sq_fsync(file, file->f_dentry);
	sound.soft = sound.dsp;
	sound.hard = sound.dsp;
	sound_silence();

	sq_release_buffers();
	MOD_DEC_USE_COUNT;

	/* There is probably a DOS atack here. They change the mode flag. */
	/* XXX add check here */
	if (file->f_mode & FMODE_WRITE) {
		sq.busy = 0;
		WAKE_UP(sq.open_queue);
	}

	/* Wake up a process waiting for the queue being released.
	 * Note: There may be several processes waiting for a call
	 * to open() returning. */

	return rc;
}


static int sq_ioctl(struct inode *inode, struct file *file, u_int cmd,
		    u_long arg)
{
	u_long fmt;
	int data;
	int size, nbufs;

	switch (cmd) {
	case SNDCTL_DSP_RESET:
		sq_reset();
		return 0;
	case SNDCTL_DSP_POST:
	case SNDCTL_DSP_SYNC:
		return sq_fsync(file, file->f_dentry);

		/* ++TeSche: before changing any of these it's
		 * probably wise to wait until sound playing has
		 * settled down. */
	case SNDCTL_DSP_SPEED:
		sq_fsync(file, file->f_dentry);
		IOCTL_IN(arg, data);
		return IOCTL_OUT(arg, sound_set_speed(data));
	case SNDCTL_DSP_STEREO:
		sq_fsync(file, file->f_dentry);
		IOCTL_IN(arg, data);
		return IOCTL_OUT(arg, sound_set_stereo(data));
	case SOUND_PCM_WRITE_CHANNELS:
		sq_fsync(file, file->f_dentry);
		IOCTL_IN(arg, data);
		return IOCTL_OUT(arg, sound_set_stereo(data-1)+1);
	case SNDCTL_DSP_SETFMT:
		sq_fsync(file, file->f_dentry);
		IOCTL_IN(arg, data);
		return IOCTL_OUT(arg, sound_set_format(data));
	case SNDCTL_DSP_GETFMTS:
		fmt = 0;
		if (sound.trans) {
			if (sound.trans->ct_ulaw)
				fmt |= AFMT_MU_LAW;
			if (sound.trans->ct_alaw)
				fmt |= AFMT_A_LAW;
			if (sound.trans->ct_s8)
				fmt |= AFMT_S8;
			if (sound.trans->ct_u8)
				fmt |= AFMT_U8;
			if (sound.trans->ct_s16be)
				fmt |= AFMT_S16_BE;
			if (sound.trans->ct_u16be)
				fmt |= AFMT_U16_BE;
			if (sound.trans->ct_s16le)
				fmt |= AFMT_S16_LE;
			if (sound.trans->ct_u16le)
				fmt |= AFMT_U16_LE;
		}
		return IOCTL_OUT(arg, fmt);
	case SNDCTL_DSP_GETBLKSIZE:
		size = sq.block_size
			* sound.soft.size * (sound.soft.stereo + 1)
			/ (sound.hard.size * (sound.hard.stereo + 1));
		return IOCTL_OUT(arg, size);
	case SNDCTL_DSP_SUBDIVIDE:
		break;
	case SNDCTL_DSP_SETFRAGMENT:
		if (sq.count || sq.active || sq.syncing)
			return -EINVAL;
		// Can't set fragment on VRC4173 Sound driver
		return -EINVAL;
		IOCTL_IN(arg, size);
		nbufs = size >> 16;
		if (nbufs < 2 || nbufs > numBufs)
			nbufs = numBufs;
		size &= 0xffff;
		if (size >= 8 && size <= 29) {
			size = 1 << size;
			size *= sound.hard.size * (sound.hard.stereo + 1);
			size /= sound.soft.size * (sound.soft.stereo + 1);
			if (size > (bufSize << 10))
				size = bufSize << 10;
		} else
			size = bufSize << 10;
		sq_setup(numBufs, size, sound_buffers);
		sq.max_active = nbufs;
		return 0;

	default:
		return mixer_ioctl(inode, file, cmd, arg);
	}
	return -EINVAL;
}



static struct file_operations sq_fops =
{
	llseek:  sound_lseek,
	write:   sq_write,
	ioctl:   sq_ioctl,
	open:    sq_open,
	release: sq_release
};


static void __init sq_init(void)
{
#ifndef MODULE
	int sq_unit;
#endif
	sq_unit = register_sound_dsp(&sq_fops, -1);
	if (sq_unit < 0)
		return;

	init_waitqueue_head(&sq.action_queue);
	init_waitqueue_head(&sq.open_queue);
	init_waitqueue_head(&sq.sync_queue);

	sq.busy = 0;

	/* whatever you like as startup mode for /dev/dsp,
	 * (/dev/audio hasn't got a startup mode). note that
	 * once changed a new open() will *not* restore these!
	 */
	sound.dsp.format = AFMT_U8;
	sound.dsp.stereo = 0;
	sound.dsp.size = 8;

	/* set minimum rate possible without expanding */
	switch (sound.mach.type) {
	case DMASND_VRC4173:
		sound.dsp.speed = 8000;
		break;
	}

	/* before the first open to /dev/dsp this wouldn't be set */
	sound.soft = sound.dsp;
	sound.hard = sound.dsp;

	sound_silence();
}

/*
 * /dev/sndstat
 */


/* state.buf should not overflow! */

static int state_open(struct inode *inode, struct file *file)
{
	char *buffer = state.buf, *mach = "";
	int len = 0;

	if (state.busy)
		return -EBUSY;

	MOD_INC_USE_COUNT;
	state.ptr = 0;
	state.busy = 1;

	switch (sound.mach.type) {
	case DMASND_VRC4173:
		mach = "VRC4173 ";
		break;
	}
	len += sprintf(buffer+len, "%sDMA sound driver:\n", mach);

	len += sprintf(buffer+len, "\tsound.format = 0x%x", sound.soft.format);
	switch (sound.soft.format) {
	case AFMT_MU_LAW:
		len += sprintf(buffer+len, " (mu-law)");
		break;
	case AFMT_A_LAW:
		len += sprintf(buffer+len, " (A-law)");
		break;
	case AFMT_U8:
		len += sprintf(buffer+len, " (unsigned 8 bit)");
		break;
	case AFMT_S8:
		len += sprintf(buffer+len, " (signed 8 bit)");
		break;
	case AFMT_S16_BE:
		len += sprintf(buffer+len, " (signed 16 bit big)");
		break;
	case AFMT_U16_BE:
		len += sprintf(buffer+len, " (unsigned 16 bit big)");
		break;
	case AFMT_S16_LE:
		len += sprintf(buffer+len, " (signed 16 bit little)");
		break;
	case AFMT_U16_LE:
		len += sprintf(buffer+len, " (unsigned 16 bit little)");
		break;
	}
	len += sprintf(buffer+len, "\n");
	len += sprintf(buffer+len, "\tsound.speed = %dHz (phys. %dHz)\n",
		       sound.soft.speed, sound.hard.speed);
	len += sprintf(buffer+len, "\tsound.stereo = 0x%x (%s)\n",
		       sound.soft.stereo, sound.soft.stereo ? "stereo" : "mono");
	switch (sound.mach.type) {
	}
	len += sprintf(buffer+len, "\tsq.block_size = %d sq.max_count = %d"
		       " sq.max_active = %d\n",
		       sq.block_size, sq.max_count, sq.max_active);
	len += sprintf(buffer+len, "\tsq.count = %d sq.rear_size = %d\n", sq.count,
		       sq.rear_size);
	len += sprintf(buffer+len, "\tsq.active = %d sq.syncing = %d\n",
		       sq.active, sq.syncing);
	state.len = len;
	return 0;
}


static int state_release(struct inode *inode, struct file *file)
{
	state.busy = 0;
	MOD_DEC_USE_COUNT;
	return 0;
}


static ssize_t state_read(struct file *file, char *buf, size_t count,
			  loff_t *ppos)
{
	int n = state.len - state.ptr;
	if (n > count)
		n = count;
	if (n <= 0)
		return 0;
	if (copy_to_user(buf, &state.buf[state.ptr], n))
		return -EFAULT;
	state.ptr += n;
	return n;
}


static struct file_operations state_fops =
{
	llseek:  sound_lseek,
	read:    state_read,
	open:    state_open,
	release: state_release
};


static void __init state_init(void)
{
#ifndef MODULE
	int state_unit;
#endif
	state_unit = register_sound_special(&state_fops, SND_DEV_STATUS);
	if (state_unit < 0)
		return;
	state.busy = 0;
}


/*** Common stuff ********************************************************/

static loff_t sound_lseek(struct file *file, long long offset, int orig)
{
	return -ESPIPE;
}


/*** Config & Setup **********************************************************/


void __init dmasound_init(void)
{
	int has_sound = 0;
	sound.mach = machVRC4173;
	has_sound = 1;

	if (!has_sound)
		return;

	/* Set up sound queue, /dev/audio and /dev/dsp. */

	/* Set default settings. */
	sq_init();

	/* Set up /dev/sndstat. */
	state_init();

	/* Set up /dev/mixer. */
	mixer_init();
	if (!sound.mach.irqinit()) {
		printk(KERN_ERR "VRC4173 AIU sound driver: Interrupt initialization failed\n");
		return;
	}
#ifdef MODULE
	irq_installed = 1;
#endif

	printk(KERN_INFO "VRC4173 AIU sound driver installed, using %d buffers of %dk.\n",
	       numBufs, bufSize);

	return;
}


#define MAXARGS		8	/* Should be sufficient for now */

void __init dmasound_setup(char *str, int *ints)
{
	/* check the bootstrap parameter for "dmasound=" */

	switch (ints[0]) {
	case 3:
		if ((ints[3] < 0) || (ints[3] > MAX_CATCH_RADIUS))
			printk("dmasound_setup: illegal catch radius, using default = %d\n", catchRadius);
		else
			catchRadius = ints[3];
		/* fall through */
	case 2:
		if (ints[1] < MIN_BUFFERS)
			printk("dmasound_setup: illegal number of buffers, using default = %d\n", numBufs);
		else
			numBufs = ints[1];
		if (ints[2] < MIN_BUFSIZE || ints[2] > MAX_BUFSIZE)
			printk("dmasound_setup: illegal buffer size, using default = %d\n", bufSize);
		else
			bufSize = ints[2];
		break;
	case 0:
		break;
	default:
		printk("dmasound_setup: illegal number of arguments\n");
	}
}


#ifdef MODULE

int init_module(void)
{
	dmasound_init();
	return 0;
}


void cleanup_module(void)
{
	if (irq_installed) {
		sound_silence();
		sound.mach.irqcleanup();
	}

	sq_release_buffers();

	if (mixer_unit >= 0)
		unregister_sound_mixer(mixer_unit);
	if (state_unit >= 0)
		unregister_sound_special(state_unit);
	if (sq_unit >= 0)
		unregister_sound_dsp(sq_unit);
}

#endif /* MODULE */

#undef DEBUG

/* vim:set ts=4:set sw=4: */

int vrc4173_audio_init(struct pci_dev *pdev, int channel)
{
	dmasound_init();
	return 0;
}

#include <linux/ac97_codec.h>

// Initial volome value
#define AC97_INITIALVOLUME	0x0

static void ac97_write_codec_reg(struct pci_dev *pdev, unsigned short reg, unsigned short value)
{
    long ioaddr;
    int timeout = 1000;

    ioaddr = pci_resource_start (pdev, 0);

    outl(inl(ioaddr + VRC4173_AC97_CODEC_WR) & ~VRC4173_AC97_CODEC_WR_RWC,
	 ioaddr + VRC4173_AC97_CODEC_WR);
    while (timeout > 0) {
	if ((inl(ioaddr + VRC4173_AC97_CODEC_WR) & VRC4173_AC97_CODEC_WR_WRDY) == 0)
	    break;
	timeout--;
    }
    if (timeout <= 0)
	printk("ac97_write_codec_reg timed out!\n");
    outl((reg << 16) | value, ioaddr + VRC4173_AC97_CODEC_WR);
}

static unsigned short ac97_read_codec_reg(struct pci_dev *pdev, unsigned short reg)
{
    long ioaddr;
    int timeout = 1000;
    unsigned short value;

    ioaddr = pci_resource_start (pdev, 0);

    outl(VRC4173_AC97_CODEC_WR_RWC,
	 ioaddr + VRC4173_AC97_CODEC_WR);
    while (timeout > 0) {
	if ((inl(ioaddr + VRC4173_AC97_CODEC_WR) & VRC4173_AC97_CODEC_WR_WRDY) == 0)
	    break;
	timeout--;
    }
    outl(((reg << 16) | VRC4173_AC97_CODEC_WR_RWC) & 0x00ff0000,
	 ioaddr + VRC4173_AC97_CODEC_WR);
    mdelay(10);
    timeout = 10000;
    while (timeout > 0) {
	if ((value = inl(ioaddr + VRC4173_AC97_CODEC_RD)) & VRC4173_AC97_CODEC_RD_RDRDY)
	    break;
	timeout--;
    }
    if (timeout <= 0)
	printk("ac97_read_codec_reg timed out (0x%08x)!\n", value);
    return (value % VRC4173_AC97_CODEC_RD_DMASK);
}

static void ac97_speaker_on(struct pci_dev *pdev)
{
    if (ac97_dev != NULL) {
	vrc4173_turn_on_device(vrc4173_dev, AC97_DEV);
	ac97_write_codec_reg(ac97_dev, AC97_MASTER_VOL_STEREO, AC97_INITIALVOLUME);
	ac97_write_codec_reg(ac97_dev, AC97_PCBEEP_VOL, AC97_INITIALVOLUME);
	ac97_write_codec_reg(ac97_dev, AC97_HEADPHONE_VOL, AC97_INITIALVOLUME);
	ac97_write_codec_reg(ac97_dev, AC97_PCMOUT_VOL, AC97_MUTE);
	printk("speaker on\n");
    } else
	printk("no VRC4173 AC97 device found to turn speaker on with!\n");
}

static void ac97_speaker_off(struct pci_dev *pdev)
{
    if (ac97_dev != NULL) {
	ac97_write_codec_reg(ac97_dev, AC97_PCBEEP_VOL, AC97_MUTE);
	ac97_write_codec_reg(ac97_dev, AC97_PCMOUT_VOL, AC97_MUTE);
	printk("speaker off\n");
    }
}

static void vrc4173_ac97_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
}

static int __devinit vrc4173_ac97_init_one (struct pci_dev *pdev,
					  const struct pci_device_id *ent)
{
    int i, irq;
    long ioaddr;

    ioaddr = pci_resource_start (pdev, 0);
    irq = pdev->irq;
    if (!ioaddr || ((pci_resource_flags(pdev, 0) & IORESOURCE_IO) == 0)) {
	printk (KERN_ERR "vrc4173 ac97: no I/O resource at PCI BAR #0\n");
	return -ENODEV;
    }
        
    i = pci_enable_device(pdev);
    if (i)
	return i;
        
    if (request_region(ioaddr, VRC4173_IO_EXTENT, "vrc4173") == NULL) {
	printk (KERN_ERR "vrc4173 ac97: I/O resource 0x%x @ 0x%lx busy\n",
		VRC4173_IO_EXTENT, ioaddr);
	return -EBUSY;
    }
    
    if (request_irq(irq, &vrc4173_ac97_interrupt, SA_SHIRQ, "vrc4173 ac97", (void *)pdev)) {
	return -EBUSY;
    }

    if (ac97_dev == NULL)
	ac97_dev = pdev;

    vrc4173_turn_off_device(vrc4173_dev, AC97_DEV);
    printk("VRC4173 AC97 initialized\n");
    return 0;
}

static void __devexit vrc4173_ac97_remove_one (struct pci_dev *pdev)
{
}

static struct pci_driver vrc4173_driver = {
        name:           "vrc4173_ac97",
        probe:          vrc4173_ac97_init_one,
        remove:         vrc4173_ac97_remove_one,
        id_table:       vrc4173_ac97_tbl,
};

static int __init vrc4173_ac97_init(void)
{
    int rc;

    rc = pci_module_init(&vrc4173_driver);
    return rc;
}

static void __exit vrc4173_ac97_cleanup(void)
{
    printk("vrc4173_ac97_cleanup called\n");
    pci_unregister_driver (&vrc4173_driver);
}

module_init(vrc4173_ac97_init);
module_exit(vrc4173_ac97_cleanup);
