/*
 * linux/arch/mips/vr41xx/vrc4173-ps2kbd.c
 *
 * Platform support for NEC Harrier.
 *
 * Copyright (C) 2000 Michael R. McDonald
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * The button code originally came from arch/mips/vr41xx/gpiobtns.c
 *
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/pc_keyb.h>
#include <linux/kbd_kern.h>
#include <linux/kbd_ll.h>
#include <asm/io.h>
#include <asm/keyboard.h>
#include <asm/vr41xx.h>
#include <asm/vrc4173.h>

#include <asm/vr41xx-platdep.h>

extern struct pci_dev *vrc4173_dev;

void vrc4173_turn_on_device(struct pci_dev *pdev, int device);
void vrc4173_turn_off_device(struct pci_dev *pdev, int device);
void vrc4173_and_or_reg(struct pci_dev *pdev, int reg, unsigned short and, unsigned short or);

typedef u64 btn_mask_t;

extern unsigned short gpio_btn_map[];

static void init_gpio(struct pci_dev *pdev)
{
    int i;
    unsigned short bits;

    for (i = 0; i < 20; i++) {
	if (!(gpio_btn_map[i+32]))
	    continue;
	bits = 1 << (i & 15);
	vrc4173_and_or_reg(pdev, (i < 16 ? VRC4173_GIUINTTYPL : VRC4173_GIUINTTYPH),
			   0xffff, bits);
	vrc4173_and_or_reg(pdev, (i < 16 ? VRC4173_GIUINTALSELL : VRC4173_GIUINTALSELH),
			   0xffff, bits);
	vrc4173_and_or_reg(pdev, (i < 16 ? VRC4173_GIUINTHTSELL : VRC4173_GIUINTHTSELH),
			   0xffff, bits);
	vrc4173_and_or_reg(pdev, (i < 16 ? VRC4173_MGIULINTREG : VRC4173_MGIUHINTREG),
			   0xffff, bits);
	vrc4173_and_or_reg(pdev, (i < 16 ? VRC4173_GIUINTENL : VRC4173_GIUINTENH),
			   0xffff, bits);
    }

    vrc4173_turn_on_device(pdev, GIU_DEV);
}

void vrc4173_init_buttons(void)
{
#ifndef CONFIG_KEYBOARD
    if (vrc4173_dev)
	init_gpio(vrc4173_dev);
#endif
}

void vrc4173_close_buttons(void)
{
    vrc4173_turn_off_device(vrc4173_dev, GIU_DEV);
}

btn_mask_t vrc4173_gpio_state(void)
{
    long ioaddr;
    btn_mask_t bits1 = 0, bits2 = 0;

    if (vrc4173_dev) {
	ioaddr = pci_resource_start(vrc4173_dev, 0);
	bits1 = inw(ioaddr + VRC4173_GIUPIODL);
	bits2 = inw(ioaddr + VRC4173_GIUPIODH);
    }
    return (bits1 << 32) | (bits2 << 48);
}

extern void gpio_handle_btn_interrupt(int gpio);

void vrc4173_handle_buttons(struct pci_dev *pdev, int index)
{
    long ioaddr;
    unsigned short status;
    int i;

    ioaddr = pci_resource_start (pdev, 0);

    status = inw(ioaddr + VRC4173_GIULINTREG);
    for (i = 0 ; i < 16 ; i++)
	if ((status & (1 << i)) != 0)
	    gpio_handle_btn_interrupt(i + 32);
    outw(status, ioaddr + VRC4173_GIUINTSTATL);
    
    status = inw(ioaddr + VRC4173_GIUHINTREG);
    for (i = 0 ; i < 5 ; i++)
	if ((status & (1 << i)) != 0)
	    gpio_handle_btn_interrupt(i + 48);
    outw(status, ioaddr + VRC4173_GIUINTSTATH);
}

