/*
 * linux/arch/mips/vr41xx/vrc4173-ps2kbd.c
 *
 * Platform support for NEC Harrier.
 *
 * Copyright (C) 2000 Michael R. McDonald
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * The PS2 code originally came from drivers/char/kc_kbd.c
 *
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/pc_keyb.h>
#include <linux/kbd_kern.h>
#include <linux/kbd_ll.h>
#include <asm/io.h>
#include <asm/keyboard.h>
#include <asm/vr41xx.h>
#include <asm/vrc4173.h>

#include <asm/vr41xx-platdep.h>

extern struct pci_dev *vrc4173_dev;

spinlock_t vrc4173_controller_lock = SPIN_LOCK_UNLOCKED;

int vrc4173_PS2_wait_for_input(struct pci_dev *pdev, int channel);
void vrc4173_PS2_write_data(struct pci_dev *pdev, int channel, unsigned short data);

/* AT scancodes to XT scanodes translation table from drivers/char/q40_keyb.c */
unsigned static char at2xt[256] =
{/* 0,   1,   2,   3,   4,   5,   6,   7,   8,   9,   a,   b,   c,   d,   e,   f, */
 0x00,0x43,0x00,0x3f,0x3d,0x3b,0x3c,0x58,0x00,0x44,0x42,0x40,0x3e,0x0f,0x29,0x00,     /* 0x00 - 0x0f */
 0x00,0x38,0x2a,0x00,0x1d,0x10,0x02,0x00,0x00,0x00,0x2c,0x1f,0x1e,0x11,0x03,0x00,     /* 0x10 - 0x1f */
 0x00,0x2e,0x2d,0x20,0x12,0x05,0x04,0x00,0x21,0x39,0x2f,0x21,0x14,0x13,0x06,0x00,     /* 0x20 - 0x2f */
 0x00,0x31,0x30,0x23,0x22,0x15,0x07,0x00,0x24,0x00,0x32,0x24,0x16,0x08,0x09,0x00,     /* 0x30 - 0x3f */
 0x00,0x33,0x25,0x17,0x18,0x0b,0x0a,0x00,0x00,0x34,0x35,0x26,0x27,0x19,0x0c,0x00,     /* 0x40 - 0x4f */
 0x00,0x00,0x28,0x00,0x1a,0x0d,0x00,0x00,0x3a,0x36,0x1c,0x1b,0x00,0x2b,0x00,0x00,     /* 0x50 - 0x5f*/
 0x00,0x56,0x00,0x00,0x00,0x00,0x0e,0x00,0x00,0x4f,0x00,0x4b,0x47,0x00,0x00,0x00,     /* 0x60 - 0x6f */
 0x52,0x53,0x50,0x4c,0x4d,0x48,0x01,0x45,0x57,0x4e,0x51,0x4a,0x37,0x49,0x46,0x00,     /* 0x70 - 0x7f */
 0x00,0x00,0x00,0x41,0x37,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,     /* 0x80 - 0x8f */
 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,     /* 0x90 - 0x9f */
 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,     /* 0xa0 - 0xaf */
 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,     /* 0xb0 - 0xbf */
 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,     /* 0xc0 - 0xcf */
 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,     /* 0xd0 - 0xdf */
 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,     /* 0xe0 - 0xef */
 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,     /* 0xf0 - 0xff */
};

char *vrc4173_PS2_init_kbd(struct pci_dev *pdev, int channel)
{
    long ioaddr;
    unsigned short status;

    ioaddr = pci_resource_start (pdev, 0);

    /* enable the port and turn off interrupts */
    outw(0x0, ioaddr + VRC4173_PS2CH1CTRL + channel);
    
//     outw(1, ioaddr + VRC4173_PS2CH1RST + channel);
//     mdelay(200);
//     status = PS2_wait_for_input(pdev, channel);
//     if ((status != 0) && (status != KBD_REPLY_POR)) {
// 	printk("VRC4173: keyboard reset status = 0x%x\n", status);
// 	return "Keyboard failed reset";
//     }

#if 0
	/*
	 * Test the keyboard interface.
	 * This seems to be the only way to get it going.
	 * If the test is successful a x55 is placed in the input buffer.
	 */
	vrc4173_PS2_write_data(pdev, channel, KBD_CCMD_SELF_TEST);
	if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	    return "Keyboard failed self test";

	/*
	 * Perform a keyboard interface test.  This causes the controller
	 * to test the keyboard clock and data lines.  The results of the
	 * test are placed in the input buffer.
	 */
	vrc4173_PS2_write_data(pdev, channel, KBD_CCMD_KBD_TEST);
	if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	    return "Keyboard interface failed self test";
#endif

    /*
     * Enable the keyboard by allowing the keyboard clock to run.
     */
    vrc4173_PS2_write_data(pdev, channel, KBD_CCMD_KBD_ENABLE);

    /*
     * Reset keyboard. If the read times out
     * then the assumption is that no keyboard is
     * plugged into the machine.
     * This defaults the keyboard to scan-code set 2.
     *
     * Set up to try again if the keyboard asks for RESEND.
     */
    do {
	vrc4173_PS2_write_data(pdev, channel, KBD_CMD_RESET);
	status = vrc4173_PS2_wait_for_input(pdev, channel);
	if (status == KBD_REPLY_ACK)
	    break;
	if (status != KBD_REPLY_RESEND)
	    return "Keyboard reset failed, no ACK";
    } while (1);

    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_POR)
	return "Keyboard reset failed, no POR";

    /*
     * Set keyboard controller mode. During this, the keyboard should be
     * in the disabled state.
     *
     * Set up to try again if the keyboard asks for RESEND.
     */
    do {
	vrc4173_PS2_write_data(pdev, channel, KBD_CMD_DISABLE);
	status = vrc4173_PS2_wait_for_input(pdev, channel);
	if (status == KBD_REPLY_ACK)
	    break;
	if (status != KBD_REPLY_RESEND)
	    return "Disable keyboard: no ACK";
    } while (1);

    vrc4173_PS2_write_data(pdev, channel, KBD_CCMD_WRITE_MODE);
    vrc4173_PS2_write_data(pdev, channel, KBD_MODE_KBD_INT
		   | KBD_MODE_SYS
		   | KBD_MODE_DISABLE_MOUSE
		   | KBD_MODE_KCC);

    /* ibm powerpc portables need this to use scan-code set 1 -- Cort */
    vrc4173_PS2_write_data(pdev, channel, KBD_CCMD_READ_MODE);
    if (!(vrc4173_PS2_wait_for_input(pdev, channel) & KBD_MODE_KCC)) {
	/*
	 * If the controller does not support conversion,
	 * Set the keyboard to scan-code set 1.
	 */
	vrc4173_PS2_write_data(pdev, channel, 0xF0);
	vrc4173_PS2_wait_for_input(pdev, channel);
	vrc4173_PS2_write_data(pdev, channel, 0x01);
	vrc4173_PS2_wait_for_input(pdev, channel);
    }
	
    vrc4173_PS2_write_data(pdev, channel, KBD_CMD_ENABLE);
    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	return "Enable keyboard: no ACK";

    /*
     * Finally, set the typematic rate to maximum.
     */
    vrc4173_PS2_write_data(pdev, channel, KBD_CMD_SET_RATE);
    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	return "Set rate: no ACK";
    vrc4173_PS2_write_data(pdev, channel, 0x00);
    if (vrc4173_PS2_wait_for_input(pdev, channel) != KBD_REPLY_ACK)
	return "Set rate: no ACK (0x00)";

    outw(0x18, ioaddr + VRC4173_PS2CH1CTRL + channel);
    return NULL; /* success */
}

void vrc4173_handle_ps2_kbd(struct pci_dev *pdev, int index)
{
    long ioaddr;
    unsigned short status, data;
    static int down = 1;
    int channel = 0x0;

    ioaddr = pci_resource_start (pdev, 0);
    
    while ((status = inw(ioaddr + VRC4173_PS2CH1CTRL + channel)) & 0x1) {
	data = inw(ioaddr + VRC4173_PS2CH1DATA + channel);
/* 	NEC_HARRIER_DISPLAY_LEDS(data); */
/* 	printk("keyboard 0x%04x %d\n", data, down); */
	if (data == 0xf0)
	    down = 0;
	else {
	    handle_scancode(at2xt[data], down);
	    down = 1;
	}
    }
    tasklet_schedule(&keyboard_tasklet);
}

