
/*
 * linux/arch/mips/vr41xx/vrc4173.c
 *
 * Platform support for NEC Harrier.
 *
 * Copyright (C) 2000 Michael R. McDonald
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * The PS2 code originally came from drivers/char/kc_kbd.c
 *
 * The button code originally came from arch/mips/vr41xx/gpiobtns.c
 *
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/pc_keyb.h>
#include <linux/kbd_kern.h>
#include <linux/kbd_ll.h>
#include <linux/delay.h>
#include <linux/random.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/buttons.h>
#include <asm/io.h>
#include <asm/keyboard.h>
#include <asm/vr41xx.h>
#include <asm/vrc4173.h>

#include <asm/vr41xx-platdep.h>

static const char *version =
"VRC4173 ver 1.0: Michael R. McDonald http://www.coolcs.com \n";

static struct pci_device_id vrc4173_pci_tbl[] __devinitdata = {
        { PCI_VENDOR_ID_NEC, PCI_DEVICE_ID_NEC_VRC4173, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
        { 0, }
};
MODULE_DEVICE_TABLE(pci, vrc4173_pci_tbl);

struct pci_dev *vrc4173_dev;

#ifdef CONFIG_PSMOUSE
void vrc4173_handle_ps2_mouse(struct pci_dev *pdev, int index);
char *vrc4173_PS2_init_mouse(struct pci_dev *pdev, int channel);
int vrc4173_psaux_init(void);
#endif
#ifdef CONFIG_PSKEYBOARD
void vrc4173_handle_ps2_kbd(struct pci_dev *pdev, int index);
char *vrc4173_PS2_init_kbd(struct pci_dev *pdev, int channel);
#endif
#ifdef CONFIG_TOUCH_PANEL
void vrc4173_handle_tpanel(struct pci_dev *pdev, int index);
int vrc4173_tpanel_init(struct pci_dev *pdev, int channel);
#endif
#ifdef CONFIG_BUTTONS_GPIO
void vrc4173_handle_buttons(struct pci_dev *pdev, int index);
#endif
#ifdef CONFIG_SOUND
void vrc4173_handle_audio(struct pci_dev *pdev, int index);
int vrc4173_audio_init(struct pci_dev *pdev, int channel);
#endif
#ifdef CONFIG_KEYBOARD
void vrc4173_handle_keyboard(struct pci_dev *pdev, int index);
int vrc4173_keyboard_init(struct pci_dev *pdev, int channel);
#endif

static unsigned short ICU_to_CMU[] = {
    1 << 5,
    1 << 7,
    1 << 6,
    1 << 4,
    1 << 3,
    1 << 0,
    1 << 2,
    1 << 1,
    0,		/* can't turn power off to the GIU */
    1 << 8,
    1 << 8,
    0,
    0,
    0,
    0,
    0,
};

void vrc4173_turn_on_device(struct pci_dev *pdev, int device)
{
    long ioaddr;
    unsigned short mask;

    ioaddr = pci_resource_start (pdev, 0);
    cli();
    /* turn on the clock */
    mask = inw(ioaddr + VRC4173_CMUCLKMSK);
    mask |= (ICU_to_CMU[device]);
    outw(mask, ioaddr + VRC4173_CMUCLKMSK);
    /* turn on the interrupt */
    mask = inw(ioaddr + VRC4173_MSYSINT1REG);
    mask |= (1 << device);
    outw(mask, ioaddr + VRC4173_MSYSINT1REG);
    sti();
#if 0
    printk("****** ON %d MSYSINT 0x%04x CMUCLK 0x%04x\n",
	   device,
	   inw(ioaddr + VRC4173_MSYSINT1REG),
	   inw(ioaddr + VRC4173_CMUCLKMSK));
#endif
}

void vrc4173_turn_off_device(struct pci_dev *pdev, int device)
{
    long ioaddr;
    unsigned short mask;

    ioaddr = pci_resource_start (pdev, 0);
    cli();
    /* turn off the interrupt */
    mask = inw(ioaddr + VRC4173_MSYSINT1REG);
    mask &= ~(1 << device);
    outw(mask, ioaddr + VRC4173_MSYSINT1REG);
    /* turn off the clock */
    mask = inw(ioaddr + VRC4173_CMUCLKMSK);
    mask &= ~(ICU_to_CMU[device]);
    outw(mask, ioaddr + VRC4173_CMUCLKMSK);
    sti();
#if 0
    printk("****** OFF %d MSYSINT 0x%04x CMUCLK 0x%04x\n",
	   device,
	   inw(ioaddr + VRC4173_MSYSINT1REG),
	   inw(ioaddr + VRC4173_CMUCLKMSK));
#endif
}

void vrc4173_and_or_reg(struct pci_dev *pdev, int reg, unsigned short and, unsigned short or)
{
    long ioaddr;
    unsigned short bits;

    ioaddr = pci_resource_start (pdev, 0);
    cli();
    bits = inw(ioaddr + reg);
    bits &= and;
    bits |= or;
    outw(bits, ioaddr + reg);
    sti();
}

void vrc4173_enable_pcmcia_irq(struct pci_dev *dev)
{
    switch (PCI_SLOT(dev->devfn)) {
    case 12:
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_MSYSINT1REG, 0xffff, 4);
	break;
    case 13:
	vrc4173_and_or_reg(vrc4173_dev, VRC4173_MSYSINT1REG, 0xffff, 2);
	break;
    default:
	printk(KERN_ERR "vrc4173_enable_pcmcia_irq: unknown slot %d\n", PCI_SLOT(dev->devfn));
    }
}

int vrc4173_PS2_wait_for_input(struct pci_dev *pdev, int channel)
{
    long timeout = KBD_INIT_TIMEOUT;
    long ioaddr;
    unsigned short status, data;

    ioaddr = pci_resource_start (pdev, 0);

    while (timeout-- && ((status = inw(ioaddr + VRC4173_PS2CH1CTRL + channel)) & 0x1) == 0) {
	mdelay(1);
    }
    if (timeout == 0) {
	printk("read timed out\n");
	return -1;
    }
    data = inw(ioaddr + VRC4173_PS2CH1DATA + channel);
//     printk("read data = 0x%x\n", data);
    return (data);
}

void vrc4173_PS2_write_data(struct pci_dev *pdev, int channel, unsigned short data)
{
    long timeout = KBD_INIT_TIMEOUT;
    long ioaddr;
    unsigned short status, old_setting;

    ioaddr = pci_resource_start (pdev, 0);

    old_setting = inw(ioaddr + VRC4173_PS2CH1CTRL + channel) & 0x18;

    /* turn reception off */
    outw(0x4, ioaddr + VRC4173_PS2CH1CTRL + channel);
    udelay(100);

    /* drain the FIFO */
    while ((status = inw(ioaddr + VRC4173_PS2CH1CTRL + channel)) & 0x1) {
	status = inw(ioaddr + VRC4173_PS2CH1DATA + channel);
	printk("dropping input - %d 0x%04x\n", channel == 0 ? 1 : 2, status);
	
    }

    /* send the data */
    outw(data, ioaddr + VRC4173_PS2CH1DATA + channel);
    udelay(100);

    /* reenable reception */
    outw(old_setting, ioaddr + VRC4173_PS2CH1CTRL + channel);

    while (timeout-- && (status = inw(ioaddr + VRC4173_PS2CH1CTRL + channel)) & 0x2) {
	udelay(1);
    }
    if (timeout == 0)
	printk("write timed out\n");
}

#ifdef CONFIG_PCMCIA
static void vrc4173_handle_pcmcia(struct pci_dev *pdev, int index)
{
//     printk("PCMCIA %d interrupt\n", index == 2 ? 1 : 2);
}
#endif

struct irq_handler {
    char *name;
    void (*handler)(struct pci_dev *pdev, int index);
} handlers[16] = {
    { "USB",		NULL},
#ifdef CONFIG_PCMCIA
    { "PCMCIA 2",	vrc4173_handle_pcmcia},
    { "PCMCIA 1",	vrc4173_handle_pcmcia},
#else
    { "PCMCIA 2",	NULL},
    { "PCMCIA 1",	NULL},
#endif
#ifdef CONFIG_PSMOUSE
    { "PS2 mouse",	vrc4173_handle_ps2_mouse},
#else
    { "PS2 mouse",	NULL},
#endif
#ifdef CONFIG_PSKEYBOARD
    { "PS2 keyboard",	vrc4173_handle_ps2_kbd},
#else
    { "PS2 keyboard",	NULL},
#endif
#ifdef CONFIG_TOUCH_PANEL
    { "PIU",		vrc4173_handle_tpanel},
#else
    { "PIU",		NULL},
#endif
#ifdef CONFIG_SOUND
    { "AIU",		vrc4173_handle_audio},
#else
    { "AIU",		NULL},
#endif
#ifdef CONFIG_KEYBOARD
    { "KIU",		vrc4173_handle_keyboard},
#else
    { "KIU",		NULL},
#endif
#ifdef CONFIG_BUTTONS_GPIO
    { "GIU",		vrc4173_handle_buttons},
#else
    { "GIU",		NULL},
#endif
    { "AC97",		NULL},
    { "AC97-1",		NULL},
    { "RFU",		NULL},
    { "RFU",		NULL},
    { "DOZEPIU",	NULL},
    { "RFU",		NULL},
    { "RFU",		NULL}
};

static void vrc4173_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    struct pci_dev *pdev = (struct pci_dev *) dev_id;
    unsigned short status;
    long ioaddr;
    int i;

    ioaddr = pci_resource_start (pdev, 0);
    status = inw(ioaddr + VRC4173_SYSINT1REG);

    for (i = 0 ; i < 16 ; i++)
	if ((status & (1 << i)) != 0) {
	    void (*func)(struct pci_dev *dev, int index);

	    func = handlers[i].handler;
	    if (func == NULL) {
		printk(KERN_DEBUG "VRC4173: spurious interrupt %s, MSYSINT1REG 0x%04x\n",
		       handlers[i].name, inw(ioaddr + VRC4173_MSYSINT1REG));
		vrc4173_turn_off_device(pdev, i);
		continue;
	    }
// 	    printk("IRQ %s\n", handlers[i].name);
	    func(pdev, i);
	}
}

static int __devinit vrc4173_pci_init_one (struct pci_dev *pdev,
					  const struct pci_device_id *ent)
{
    static unsigned version_printed = 0;
    int i, irq;
    long ioaddr;
    char *msg;

    if (version_printed++ == 0)
	printk(KERN_INFO "%s", version);

    ioaddr = pci_resource_start (pdev, 0);
    irq = pdev->irq;
    if (!ioaddr || ((pci_resource_flags(pdev, 0) & IORESOURCE_IO) == 0)) {
	printk (KERN_ERR "vrc4173: no I/O resource at PCI BAR #0\n");
	return -ENODEV;
    }
        
    i = pci_enable_device(pdev);
    if (i)
	return i;
        
    if (request_region(ioaddr, VRC4173_IO_EXTENT, "vrc4173") == NULL) {
	printk (KERN_ERR "vrc4173: I/O resource 0x%x @ 0x%lx busy\n",
		VRC4173_IO_EXTENT, ioaddr);
	return -EBUSY;
    }
    
//     i = (irq - VR41XX_IRQ_GPIO0);
//     printk("GIU status:\n");
//     printk("    VR41XX_GIUPIODL     = %d\n", ((*VR41XX_GIUPIODL) >> i) & 1);
//     printk("    VR41XX_GIUINTALSELL = %d\n", ((*VR41XX_GIUINTALSELL) >> i) & 1);
//     printk("    VR41XX_GIUINTHTSELL = %d\n", ((*VR41XX_GIUINTHTSELL) >> i) & 1);
//     printk("    VR41XX_GIUINTTYPL   = %d\n", ((* VR41XX_GIUINTTYPL) >> i) & 1);
//     printk("    VR41XX_GIUINTENL    = %d\n", ((* VR41XX_GIUINTENL) >> i) & 1);
//     printk("    VR41XX_GIUINTSTATL  = %d\n", ((* VR41XX_GIUINTSTATL) >> i) & 1);
    *VR41XX_GIUINTALSELL |= 1 << (irq - VR41XX_IRQ_GPIO0);

    if (request_irq(irq, &vrc4173_interrupt, SA_SHIRQ, "vrc4173", (void *)pdev)) {
	return -EBUSY;
    }

    /* setup how the GIU lines are used */
    outw(VRC4173_SELECTREG_MASK, ioaddr + VRC4173_SELECTREG);

    /*enable the interrupt sources */
    outw(VRC4173_SYSINT_MASK, ioaddr + VRC4173_MSYSINT1REG);

    if (vrc4173_dev == NULL) {
	vrc4173_dev = pdev;
#ifdef CONFIG_PSMOUSE
	/* register a misc file driver for /dev/mouse */
	vrc4173_psaux_init();
#endif CONFIG_PSMOUSE
    }

#ifdef CONFIG_PSKEYBOARD
    /* init the PS2 keyboard */
    vrc4173_turn_on_device(pdev, PS2KBD_DEV);
    msg = vrc4173_PS2_init_kbd(pdev, 0x0);
    if (msg != NULL) {
	printk("VRC4173 PS2 keyboard: %s\n", msg);
	vrc4173_turn_off_device(pdev, PS2KBD_DEV);
    } else {
	printk("VRC4173 PS2 keyboard found\n");
    }
#endif CONFIG_PSKEYBOARD

#ifdef CONFIG_PSMOUSE
    /* init the PS2 mouse */
    vrc4173_turn_on_device(pdev, PS2AUX_DEV);
    msg = vrc4173_PS2_init_mouse(pdev, 0x20);
    if (msg != NULL) {
	printk("VRC4173 PS2 mouse: %s\n", msg);
    } else {
	printk("VRC4173 PS2 mouse found\n");
    }
#endif CONFIG_PSMOUSE
    /* turn off the mouse until someone opens /dev/mouse */
    vrc4173_turn_off_device(pdev, PS2AUX_DEV);

#ifdef CONFIG_TOUCH_PANEL
    /* init the touch panel */
    vrc4173_tpanel_init(pdev, 0);
#endif
    /* turn off the touch panel until someone opens it */
    vrc4173_turn_off_device(pdev, PIU_DEV);

#ifdef CONFIG_SOUND
    vrc4173_audio_init(pdev, 0);
#endif

#ifdef CONFIG_KEYBOARD
    vrc4173_keyboard_init(pdev, 0);
#endif

    return 0;
}

static void __devexit vrc4173_pci_remove_one (struct pci_dev *pdev)
{
}

static struct pci_driver vrc4173_driver = {
        name:           "vrc4173",
        probe:          vrc4173_pci_init_one,
        remove:         vrc4173_pci_remove_one,
        id_table:       vrc4173_pci_tbl,
};

static int __init vrc4173_pci_init(void)
{
    int rc;

    rc = pci_module_init(&vrc4173_driver);
    return rc;
}

static void __exit vrc4173_pci_cleanup(void)
{
    printk("vrc4173_pci_cleanup called\n");
    pci_unregister_driver (&vrc4173_driver);
}

module_init(vrc4173_pci_init);
module_exit(vrc4173_pci_cleanup);
