/*
 * linux/arch/mips/vr41xx/agenda_vr3.c
 *
 * Platform support for the Agenda VR3
 *
 * Copyright (C) 2000 Bradley D. LaRonde
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/vr41xx.h>
#include <asm/system.h>
#include <linux/delay.h>

void debug_display(int n) {
	*(volatile unsigned short*)0xb7ffffc0 = n;
}

static void cont_clock(int n) {
	unsigned short d = *VR41XX_GPDATLREG;
	if(n)
		d |= VR41XX_GPDATLREG_GPIO5;
	else
		d &= ~VR41XX_GPDATLREG_GPIO5;
	*VR41XX_GPDATLREG = d;
}

static void cont_strobe(int n) {
	unsigned short d = *VR41XX_GPDATLREG;
	if(n)
		d |= VR41XX_GPDATLREG_GPIO3;
	else
		d &= ~VR41XX_GPDATLREG_GPIO3;
	*VR41XX_GPDATLREG = d;
}

static void cont_data(int n) {
	unsigned short d = *VR41XX_GPDATLREG;
	if(n)
		d |= VR41XX_GPDATLREG_GPIO4;
	else
		d &= ~VR41XX_GPDATLREG_GPIO4;
	*VR41XX_GPDATLREG = d;
}

#if 0
static void cont_reset(void) {
	// Reset contrast voltage regulator.
	cont_clock(1);
	cont_strobe(1);
	udelay(20);
	cont_clock(0);
	cont_strobe(0);
	udelay(20);
}
#endif

static void cont_set(unsigned char n) {
	unsigned char mask = 0x80;

	// Get ready.
	cont_strobe(0);

	// Just in case VSTRB was high (it's not supposed to be though),
	// wait some time (ths).
	udelay(20);

	// Clock the data in.
	while(mask) {
		// Clock low.
		cont_clock(0);

		// Set the data.
		if(mask & n)
			cont_data(1);
		else
			cont_data(0);

		// Let the data settle (tsc) and give the clock it's off width (ttc - tdc).
		udelay(20);

		// Clock it in.
		cont_clock(1);

		// Give the clock it's on width (tdc).
		udelay(20);

		// Get ready for next bit.
		mask >>= 1;
	}

	// Done setting the data.
	cont_clock(0);
	cont_data(0);

	// Rest (tss) after the final clock and before strobing.
	udelay(20);

	// Strobe high.
	cont_strobe(1);

	// Let it take (tda).
	udelay(20);

	// Strobe low.
	cont_strobe(0);

	// Give the strobe a rest (ths).
	udelay(20);
}

int display_last_contrast;

int display_get_contrast(void)
{
	return display_last_contrast;
	return 0;
}

int display_set_contrast(int n)
{
	cont_set(n);
	display_last_contrast = n;
	return 0;
}

int display_get_backlight(void)
{
	return((*VR41XX_GPDATHREG & 0x0002) ? 0 : 1);
}

int display_set_backlight(int n)
{
	if(n)
		*VR41XX_GPDATHREG &= ~0x0002;
	else
		*VR41XX_GPDATHREG |= 0x0002;

	return 0;
}

int display_setup(void)
{
	// Use GP12-15 for color LCD FPD[4-7] output.
	// *VR41XX_GPMD1REG |= 0xff00;

	// Set gclk to Hpck ratio.
	// *VR41XX_LCDCFGREG1 = 0x0707;

// BUG: Sequence works for power-on, but is not observed for power-off.
// All turn off at the same time.
#define IFSIG_FRAME  0x0
#define BIAS_FRAME   0x1  // LCD_BIAS and VEE
#define VCC_FRAME    0x2  // #DOFF

	// Set LCD power control.
	*VR41XX_PWRCONREG1 = BIAS_FRAME;
	*VR41XX_PWRCONREG2 = (0 << 15) | (1 << 14) | (1 << 11) | (0 << 10)
		| (IFSIG_FRAME << 5) | VCC_FRAME;

	return 0;
}

int display_power(int on)
{
	// printk(KERN_DEBUG "vr3: display_power %d\n", on);

	if(on) {
		// Turn on the LCD panel.
		*VR41XX_PWRCONREG2 |= (1 << 10);
	}
	else {
		// Turn off the backlight.
		display_set_backlight(0);
		// Turn off the LCD panel.
		*VR41XX_PWRCONREG2 &= ~(1 << 10);
	}
	return 0;
}

void irda_power(int n)
{
	unsigned long flags;

	save_and_cli(flags);

	if(n)
		// *VR41XX_GPDATLREG &= ~VR41XX_GPDATLREG_GPIO12;
		*VR41XX_GPDATHREG &= ~VR41XX_GPDATHREG_GPIO21;
	else
		// *VR41XX_GPDATLREG |= VR41XX_GPDATLREG_GPIO12;
		*VR41XX_GPDATHREG |= VR41XX_GPDATHREG_GPIO21;

	restore_flags(flags);
}

void platdep_setup(void)
{
	// TODO: get the initial contrast value from someplace configurable.
	display_setup();
	display_set_contrast(100);
	display_set_backlight(0);
}
