/* $Id: ide-vr41xx.c,v 1.1.1.1 2001/05/02 13:47:08 mfloeer Exp $
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * IDE routines for typical pc-like standard configurations.
 *
 * Copyright (C) 1998, 1999 by Ralf Baechle
 */
#include <linux/sched.h>
#include <linux/ide.h>
#include <linux/ioport.h>
#include <linux/hdreg.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <asm/ptrace.h>
#include <asm/hdreg.h>
#include <mmac/types.h>
#include <mmac/rtos.h>
#include <mmac/hardware/hwata.h>
#include <mmac/hardware/hwxdma.h>
#include <mmac/hardware/hwide.h>
#include <mmac/hardware/hwbhif.h>
#include <mmac/hardware/hwzpio.h>

// static volatile EMMA_CU 	*emmaCU_Ptr 	= (EMMA_CU*)	0xb0012900;

#define NEC_HD_BASE	0x2000

    /*
     *  Offsets from the above base (scaling 4)
     */

// offsets for 16-bit access
#define NEC_ATA_CONTROL 0x2
#define NEC_ATA_INT_EN_ITOP_STATUS 0x62
#define NEC_ATA_INT_STATUS 0x66
#define NEC_ATA_MODE 0x6a

// standard ata registers
#define NEC_HD_DATA	0xd22
#define NEC_HD_ERROR	0xd26		/* see err-bits */
#define NEC_HD_NSECTOR	0xd2a		/* nr of sectors to read/write */
#define NEC_HD_SECTOR	0xd2e		/* starting sector */
#define NEC_HD_LCYL	0xd32		/* starting cylinder */
#define NEC_HD_HCYL	0xd36		/* high byte of starting cyl */
#define NEC_HD_SELECT	0xd3a		/* 101dhhhh , d=drive, hhhh=head */
#define NEC_HD_STATUS	0xd3e		/* see status-bits */
#define NEC_HD_CONTROL	0xd5a		/* control/altstatus */


#define NEC_HD_ISR	0x66
#define NEC_HD_IRQ  13

/* DMA defines */
/* #defines for List DMA Reset */
#define EMMA_BIOS_DMA_LIST_DMA_RST_MEM_BIT_MASK		(0x00000040)
#define EMMA_BIOS_DMA_LIST_DMA_RST_GIOW_BIT_MASK	(0x00000020)
#define EMMA_BIOS_DMA_LIST_DMA_RST_GIOR_BIT_MASK	(0x00000010)
#define EMMA_BIOS_DMA_LIST_DMA_RST_IDEW_BIT_MASK	(0x00000008)
#define EMMA_BIOS_DMA_LIST_DMA_RST_IDER_BIT_MASK	(0x00000004)
#define EMMA_BIOS_DMA_LIST_DMA_RST_PCMW_BIT_MASK	(0x00000002)
#define EMMA_BIOS_DMA_LIST_DMA_RST_PCMR_BIT_MASK	(0x00000001)
#define EMMA_BIOS_DMA_LIST_DMA_RSTALL_BIT_MASK		(0x0000007F)

/* #defines for PRI_CHSEL channel priorities - DMA channel IDs */
#define EMMA_BIOS_DMA_VIDMAIN_CHAN_ID	(0x00000000)
#define EMMA_BIOS_DMA_VIDSUB_CHAN_ID	(0x20000000)
#define EMMA_BIOS_DMA_AUDCH0_CHAN_ID	(0x40000000)
#define EMMA_BIOS_DMA_AUDCH1_CHAN_ID	(0x60000000)
#define EMMA_BIOS_DMA_PCMR_CHAN_ID		(0x80000000)
#define EMMA_BIOS_DMA_PCMW_CHAN_ID		(0x90000000)
#define EMMA_BIOS_DMA_IDER_CHAN_ID		(0xA0000000)
#define EMMA_BIOS_DMA_IDEW_CHAN_ID		(0xB0000000)
#define EMMA_BIOS_DMA_MEM_CHAN_ID		(0xC0000000)
#define EMMA_BIOS_DMA_GIORW_CHAN_ID		(0xE0000000)

/* #defines for REQ_ENA_LIST bits */
#define EMMA_BIOS_DMA_REQ_ENA_MEM_BIT_MASK	(0x00000020)
#define EMMA_BIOS_DMA_REQ_ENA_GIO_BIT_MASK	(0x00000010)
#define EMMA_BIOS_DMA_REQ_ENA_IDEW_BIT_MASK	(0x00000008)
#define EMMA_BIOS_DMA_REQ_ENA_IDER_BIT_MASK	(0x00000004)
#define EMMA_BIOS_DMA_REQ_ENA_PCMW_BIT_MASK	(0x00000002)
#define EMMA_BIOS_DMA_REQ_ENA_PCMR_BIT_MASK	(0x00000001)

/* #defines for REQ_MASK bits */
#define EMMA_BIOS_DMA_REQ_MSK_MEM_BIT_MASK			(0x00004000)
#define EMMA_BIOS_DMA_REQ_MSK_GIOW_BIT_MASK			(0x00002000)
#define EMMA_BIOS_DMA_REQ_MSK_GIOR_BIT_MASK			(0x00001000)
#define EMMA_BIOS_DMA_REQ_MSK_IDEW_BIT_MASK			(0x00000800)
#define EMMA_BIOS_DMA_REQ_MSK_IDER_BIT_MASK			(0x00000400)
#define EMMA_BIOS_DMA_REQ_MSK_PCMW_BIT_MASK			(0x00000200)
#define EMMA_BIOS_DMA_REQ_MSK_PCMR_BIT_MASK			(0x00000100)
#define EMMA_BIOS_DMA_LIST_DMA_ENAREQALL_BIT_MASK	(0x00007F0F)

/* #define for LIST DMA disable bit */
#define EMMA_BIOS_DMA_NOLIST_BIT_MASK				(0x00000001)
													
/* #defines for non-List DMA start and done bits */
#define EMMA_BIOS_DMA_DMA_START_BIT_MASK			(0x00000001)
#define EMMA_BIOS_DMA_DMA_DONE_BIT_MASK				(0x00000002)

/* #defines for INTERRUPT bits */
#define EMMA_BIOS_DMA_INT_LIST_ERROR_BIT_MASK	(0x04000000)
#define EMMA_BIOS_DMA_INT_REQ_ERROR_BIT_MASK	(0x02000000)




/* #defines for ATA COMMAND CODES */
#define MMAC_ATA_ATA_ATA_IDENTIFY_DEVICE	(0x000000EC)
#define MMAC_ATA_ATA_ATA_READ_DMA			(0x000000C8)
#define MMAC_ATA_ATA_ATA_WRITE_DMA			(0x000000CA)
#define MMAC_ATA_ATA_ATA_WRITE_SECTORS		(0x00000030)
#define MMAC_ATA_ATA_ATA_READ_SECTORS		(0x00000020)
#define MMAC_ATA_ATA_ATA_READ_MULTIPLE		(0x000000C4)
#define MMAC_ATA_ATA_ATA_WRITE_MULTIPLE	    (0x000000C5)
#define MMAC_ATA_ATA_SEEK					(0x00000070)
#define	MMAC_ATA_ATA_SET_FEATURES			(0x000000EF)

#define MMAC_ATA_ATA_SET_TRANSFER_MODE		(0x00000003)
#define MMAC_ATA_ATA_UDMA_MODE				(0x00000040)
#define MMAC_ATA_ATA_DMA_MODE				(0x00000020)
#define MMAC_ATA_ATA_PIO_MODE_IORDY		    (0x00000000)
#define MMAC_ATA_ATA_PIO_MODE_NO_IORDY		(0x00000001)

/* #defines for ATA_CONTROL register */
#define MMAC_ATA_ATA_UDMA_TRANSFER_BIT_MASK	(0x00000006)
#define MMAC_ATA_ATA_DIR_BIT_MASK			(0x00000008)

/* #defines for DMA/UDMA Timing Modes */
#define MMAC_ATA_ATA_CFG_DMA_MODE_0		    (0x00007777)
#define MMAC_ATA_ATA_CFG_DMA_MODE_1		    (0x00002222)
#define MMAC_ATA_ATA_CFG_DMA_MODE_2		    (0x00002020)
#define MMAC_ATA_ATA_CFG_DMA_DEFAULT		(0x00009999)
#define MMAC_ATA_ATA_CFG_UDMA_MODE_0		(0x00000333)
#define MMAC_ATA_ATA_CFG_UDMA_MODE_1		(0x00000322)
#define MMAC_ATA_ATA_CFG_UDMA_MODE_2		(0x00000311)
#define MMAC_ATA_ATA_CFG_UDMA_DEFAULT		(0x00000377)

/* #defines for PIO control register */
#define MMAC_ATA_ATA_S_POE_BIT_MASK		    (0x00000010)
#define MMAC_ATA_ATA_M_POE_BIT_MASK		    (0x00000020)
#define MMAC_ATA_ATA_S_PRE_BIT_MASK		    (0x00000040)
#define MMAC_ATA_ATA_M_PRE_BIT_MASK		    (0x00000080)
#define MMAC_ATA_ATA_MAX_PIO_CNT_MASK		(0x0000FF00)

/* #defines for PIO Mode register */
#define MMAC_ATA_ATA_PIO_MASK				(0x0000000F)
#define MMAC_ATA_ATA_RDY_EN_BIT_MASK		(0x00008000)

/* #defines for DMA/UDMA registers */
#define MMAC_ATA_ATA_DMA_M_CMD_MASK		    (0x0000F000)
#define MMAC_ATA_ATA_DMA_M_RCV_MASK		    (0x00000F00)
#define MMAC_ATA_ATA_DMA_S_CMD_MASK		    (0x000000F0)
#define MMAC_ATA_ATA_DMA_S_RCV_MASK		    (0x0000000F)
#define MMAC_ATA_ATA_MUDM_BIT_MASK			(0x00000100)
#define MMAC_ATA_ATA_SUDM_BIT_MASK			(0x00000200)
#define MMAC_ATA_ATA_UDMA_M_COM_MASK		(0x00000070)
#define MMAC_ATA_ATA_UDMA_S_COM_MASK		(0x00000007)
#define MMAC_ATA_ATA_DMA_ON_BIT_MASK		(0x00000001)

/* #defines for LBA register(s) */
#define MMAC_ATA_ATA_BITS_7TO0_MASK		    (0x000000FF)
#define MMAC_ATA_ATA_BITS_15TO8_MASK		(0x0000FF00)
#define MMAC_ATA_ATA_BITS_23TO16_MASK		(0x00FF0000)
#define MMAC_ATA_ATA_BITS_27TO24_MASK		(0x0F000000)
#define MMAC_ATA_ATA_LBA_27TO24_MASK		(0x0000000F)
#define MMAC_ATA_ATA_DRV_BIT_MASK			(0x00000010)
#define MMAC_ATA_ATA_LBA_BIT_MASK			(0x00000040)

/* #defines for Bus Master register */
#define MMAC_ATA_ATA_BM_EN_BIT_MASK		    (0x00000040)
#define MMAC_ATA_ATA_BM_IO_AI_BIT_MASK		(0x00000020)
#define MMAC_ATA_ATA_BM_MEMR_MODE_MASK		(0x00000018)
#define MMAC_ATA_ATA_BM_MIO_BIT_MASK		(0x00000004)
#define MMAC_ATA_ATA_DISC_EN_BIT_MASK		(0x00000002)
#define MMAC_ATA_ATA_BM_SSC_BIT_MASK		(0x00000001)

/* #defines for byte/word order swap register */
#define MMAC_ATA_ATA_BYTE_SWAP_PB_BIT_MASK	(0x00000200)
#define MMAC_ATA_ATA_WORD_SWAP_PB_BIT_MASK	(0x00000100)
#define MMAC_ATA_ATA_BYTE_SWAP_REC_BIT_MASK	(0x00000002)
#define MMAC_ATA_ATA_WORD_SWAP_REC_BIT_MASK	(0x00000001)

/* #define for ATA Soft Reset bit */
#define MMAC_ATA_ATA_ATA_SOFT_RESET_SRST_BIT_MASK	(0x00000001)

/* #define for ATA interrupt bit */
#define MMAC_ATA_ATA_ATA_INT_STATUS_INT_BIT_MASK	(0x00000001)
#define MMAC_ATA_ATA_ATA_INT_ENABLE_BIT_MASK		(0x00004000)

/* #define for ATA IDE Status/ Alt Status error bit */
#define MMAC_ATA_ATA_ATA_IDE_STATUS_ERR_BIT_MASK	(0x00000001)

/* #define for ATA IDE Features/ Error ABRT bit */
#define MMAC_ATA_ATA_ATA_IDE_FEATURES_ERROR_ABRT_BIT_MASK	(0x00000004)

/* #defines for driveInfo words */
#define MMAC_ATA_ATA_DRIVE_INFO_PIO_MODE_BIT_MASK 			(0xFF00)
#define MMAC_ATA_ATA_DRIVE_INFO_IORDY_SUPPORTED_BIT_MASK	(0x0800)
#define MMAC_ATA_ATA_DRIVE_INFO_DMA_MODE_2_BIT_MASK		    (0x0004)
#define MMAC_ATA_ATA_DRIVE_INFO_DMA_MODE_1_BIT_MASK		    (0x0002)
#define MMAC_ATA_ATA_DRIVE_INFO_DMA_MODE_0_BIT_MASK		    (0x0001)
#define MMAC_ATA_ATA_DRIVE_INFO_UDMA_MODE_2_BIT_MASK		(0x0004)
#define MMAC_ATA_ATA_DRIVE_INFO_UDMA_MODE_1_BIT_MASK		(0x0002)
#define MMAC_ATA_ATA_DRIVE_INFO_UDMA_MODE_0_BIT_MASK		(0x0001)
#define	MMAC_ATA_ATA_DRIVE_INFO_ADV_PIO_MODE_4_MASK		    (0x0002)
#define	MMAC_ATA_ATA_DRIVE_INFO_ADV_PIO_MODE_3_MASK		    (0x0001)

/* #defines for BHIF ICLK and I2CLK */
#define MMAC_ATA_ICLK_SEL			(0x01000000)
#define MMAC_ATA_I2CLK_SEL			(0x02000000)
#define MMAC_ATA_ICLK_MSK			(0x00002000)
#define MMAC_ATA_I2CLK_MSK			(0x00004000)

static int vr41xx_ide_default_irq(ide_ioreg_t base)
{
	return NEC_HD_IRQ;	// don't specify IRQ.  autoprobe instead
}

static ide_ioreg_t vr41xx_ide_default_io_base(int index)
{
	switch (index) {
		case 0:	return NEC_HD_BASE + NEC_HD_DATA;
		case 1:	return 0;
		case 2: return 0;
		case 3: return 0;
		case 4: return 0;
		case 5: return 0;
		default:
			return 0;
	}
}

static void vr41xx_ide_init_hwif_ports (hw_regs_t *hw, ide_ioreg_t data_port,
                                     ide_ioreg_t ctrl_port, int *irq)
{
	ide_ioreg_t reg = data_port;
	int i;

	for (i = IDE_DATA_OFFSET; i <= IDE_STATUS_OFFSET; i++) {
		hw->io_ports[i] = reg;
		reg += 4;
	}
	if (ctrl_port) {
		hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
	} else {
		hw->io_ports[IDE_CONTROL_OFFSET] = hw->io_ports[IDE_DATA_OFFSET] + (NEC_HD_CONTROL-NEC_HD_DATA);
	}
	if (irq != NULL)
		*irq = NEC_HD_IRQ;
}

static int vr41xx_ide_request_irq(unsigned int irq,
                                void (*handler)(int,void *, struct pt_regs *),
                                unsigned long flags, const char *device,
                                void *dev_id)
{
	return request_irq(irq, handler, SA_SHIRQ, device, dev_id);
}			

static void vr41xx_ide_free_irq(unsigned int irq, void *dev_id)
{
	free_irq(irq, dev_id);
}

static int vr41xx_ide_check_region(ide_ioreg_t from, unsigned int extent)
{
	return check_region(from, extent);
}

static void vr41xx_ide_request_region(ide_ioreg_t from, unsigned int extent,
                                    const char *name)
{
	request_region(from, extent, name);
}

static void vr41xx_ide_release_region(ide_ioreg_t from, unsigned int extent)
{
	release_region(from, extent);
}

struct ide_ops vr41xx_ide_ops = {
	&vr41xx_ide_default_irq,
	&vr41xx_ide_default_io_base,
	&vr41xx_ide_init_hwif_ports,
	&vr41xx_ide_request_irq,
	&vr41xx_ide_free_irq,
	&vr41xx_ide_check_region,
	&vr41xx_ide_request_region,
	&vr41xx_ide_release_region
};

int ide_nec_setup_hw()
{
    /* switch on harddisk on discvision hardware */
    emmaZPIO.ZPIO_PORT_DIRECTION_0 |= (1<<6);
    emmaZPIO.ZPIO_PORT_DATA_0 |= (1<<6); /* now, this pin stays up */
	emmaBHIF.BHIF_STRAP_WRITE_2 |= (1<<21);  // SEP 2


    /* Set IDE Clock bits in BHIF registers */
    /* Mask ICLK and I2CLK so they may be changed */
    emmaBHIF.BHIF_CLK_MSK |= (MMAC_ATA_ICLK_MSK | MMAC_ATA_I2CLK_MSK);
    /* Set ICLKSEL bit */
    emmaBHIF.BHIF_CLK_SEL_1 |= MMAC_ATA_ICLK_SEL;
    /* Ensure I2CLKSEL is cleared */
    emmaBHIF.BHIF_CLK_SEL_1 &= ~MMAC_ATA_I2CLK_SEL;
    /* Un-mask ICLK and I2CLK */
    emmaBHIF.BHIF_CLK_MSK &= ~(MMAC_ATA_ICLK_MSK | MMAC_ATA_I2CLK_MSK);

    /* Peform Soft Reset... */
    emmaATA.ATA_SOFT_RESET = MMAC_ATA_ATA_ATA_SOFT_RESET_SRST_BIT_MASK;
    udelay(2000);
    emmaATA.ATA_SOFT_RESET = 0x00000000;

	//emmaIDE.ATA_IDE_DEV_CONTROL_ALT_STATUS = 0x00000004;
    /* Must wait at least 2ms before checking that reset is complete... */
    udelay(2000);
    //ATA_WaitNotBsyDrdy();
//	emmaIDE.ATA_IDE_DEV_CONTROL_ALT_STATUS = 0x00000000;
	/* Enable interrupts at the ATA interface */
    emmaATA.ATA_INT_EN_ITOP_STATUS = MMAC_ATA_ATA_ATA_INT_ENABLE_BIT_MASK;
    /* Enable device interrupts */
    emmaIDE.ATA_IDE_DEV_CONTROL_ALT_STATUS = 0x00000000;

    // ATA_SetAccessMode( ATA_LOGICAL_MODE );
    
    /* Bring ATA up to MODE 1... */

    /* Set to PIO Mode0 and clear IORDY */
    emmaATA.ATA_MODE = 0x00000000;

    /* Set special control register */
    emmaATA.ATACFG_SPEC_CONTROL = 0x00000060;
 
    /* Disable pre-fetch and post-enable for both Master & Slave drives */
    emmaATA.ATACFG_PIO_CONTROL = 0x00000000;

    /* Set DMA timing config */
    emmaATA.ATACFG_DMA_TIMING_MODE = MMAC_ATA_ATA_CFG_DMA_DEFAULT;
    /* Set UDMA timing config */
    emmaATA.ATACFG_UDMA_TIMING_MODE = MMAC_ATA_ATA_CFG_UDMA_DEFAULT;

    /* Set special control register */
    emmaATA.ATACFG_SPEC_CONTROL = 0x00000060;

    // ATA_SetAccessMode( ATA_LOGICAL_MODE );
    
    /* Bring ATA up to MODE 1... */

    /* Set to PIO Mode4  */
    emmaATA.ATA_MODE = 0x00000004;
    // emmaATA.ATA_MODE = 0x00000000;
    return 0;
}
#ifdef MODULE

void xdma_handler()
{
    printk("xdma irq: %08x\n",emmaXDMA.XDMA_INT);
    emmaXDMA.XDMA_INT = 0;

}
int init_module()
{
    volatile int dummy;
    int retval;
    ide_nec_setup_hw();
    emmaXDMA.XDMA_SOFT_RST = 0x00000001;
    emmaXDMA.XDMA_VBUF_RST = 0x00000101;
    emmaXDMA.XDMA_ABUF_RST = 0x00000101;
    emmaXDMA.XDMA_LIST_RST = EMMA_BIOS_DMA_LIST_DMA_RSTALL_BIT_MASK;
    /* Short delay */
    for( dummy=0; dummy<255; dummy++ )
    {
	    ;
    }
    emmaXDMA.XDMA_SOFT_RST = 0x00000000;
    emmaXDMA.XDMA_VBUF_RST = 0x00000000;
    emmaXDMA.XDMA_ABUF_RST = 0x00000000;
    emmaXDMA.XDMA_LIST_RST = 0x00000000;

    /* Wait for 12*HCLK periods minimum - we'll wait for 1ms  */
    udelay(1000);
	
    /* Enable List DMA in each direction for all channels */
    emmaXDMA.XDMA_REQ_MASK = EMMA_BIOS_DMA_LIST_DMA_ENAREQALL_BIT_MASK;
    /* Disable REQ_ENA for IDE r/w */
    emmaXDMA.XDMA_REQ_MASK &= ~EMMA_BIOS_DMA_REQ_ENA_IDEW_BIT_MASK;
    emmaXDMA.XDMA_REQ_MASK &= ~EMMA_BIOS_DMA_REQ_ENA_IDER_BIT_MASK;
    emmaXDMA.XDMA_IDE_R_COMMAND = 1 /*nolist */ | (1<<25) /*int en*/ | (1<<28) /* burst len */ ;
    emmaXDMA.XDMA_IDE_W_COMMAND = 1 /*nolist */ | (1<<25) /*int en*/ | (1<<28) /* burst len */ ;

    udelay(1000);
	
/*	
    retval = request_irq( 9 , xdma_handler, SA_SHIRQ,
			 "xdma", NULL);
    emmaXDMA.XDMA_INT_EN = (1<<18) | (1<<19) | (1<<26) | (1<<27) | (1<<28);
    */

    ide_ops = &vr41xx_ide_ops;
    return 0;
}

void cleanup_module()
{
    // free_irq(9, NULL);
    /* switch off harddisk on discvision hardware */
    emmaZPIO.ZPIO_PORT_DIRECTION_0 |= (1<<6);
    emmaZPIO.ZPIO_PORT_DATA_0 &= ~(1<<6); /* now, this pin stays up */
}
#endif

