/* $Id: power.c,v 1.1.1.1 2001/05/02 13:47:08 mfloeer Exp $
 *
 * VR41xx reset and power management interface
 * PM User interface loosely based on Andrew Henroid's i386 ACPI driver
 *
 * Copyright (C) 2000 Michael Klar
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/delay.h>
#include <linux/mm.h>
#include <linux/sysctl.h>
#include <linux/pm.h>
#include <linux/acpi.h>
#include <linux/irq.h>	/* for disable_irq, enable_irq, to be removed */
#include <asm/vr41xx.h>
#include <asm/cacheops.h>
#include <asm/mipsregs.h>
#include <asm/pgalloc.h>
#include <asm/power.h>

/*
 * vr41xx_dma_sem is needed because we cannot suspend while a DMA transfer
 * is in progress, otherwise the CPU will be left in an undefined state.
 */
extern struct semaphore vr41xx_dma_sem;

void vr41xx_restart(char* c)
{
	void* reset_addr = (void*)0xbfc00000;

	printk("try deadman reset\n");
#if 0 // DEBUG_JAN
	/*
	* try deadman's reset first.
	*/
	*(volatile int *)VR41XX_DSUCLRREG = 1;  /* clear last one */
	*(volatile int *)VR41XX_DSUSETREG = 1;  /* 1sec */
	*(volatile int *)VR41XX_DSUCNTREG = 1;  /* enable deadman switch */
	mdelay(2000);                           /* wait 2 second */
#endif

	printk("try jump start address\n");
	/* jump start address */

	goto *reset_addr;
}

void vr41xx_hibernate(void)
{
#ifdef CONFIG_CPU_VR4181
	// this loads the hibernate code into icache
	// before putting the dram to sleep
	asm volatile (
		" .set push\n"
		" .set noreorder\n"
		" .set mips3\n"
		" la $8,1f\n"                // Set up some registers ahead of time.
		" lui $9,%%hi(%0)\n"
		" lui $10,%%hi(%1)\n"
		" li $11,0xffff\n"
		" li $12,2\n"
		" li $13,3\n"
		" li $14,1\n"
		" li $15,5\n"
		" cache %3,0($8)\n"          // Load the rest of the hibernate code into the cache.
		" cache %3,16($8)\n"
		" cache %3,32($8)\n"
		" b 1f\n"
		" cache %3,48($8)\n"
		" .align 4\n"
		"1: sh $0,%%lo(%0)($9)\n"    // Mask all of the internal peripheral clocks (CMUCLKMSK).
		" sh $11,%%lo(%2)($9)\n"     // Max out the DRAM refresh time.
		" sh $12,%%lo(%1)($10)\n"    // Begin a self-refresh request.
		"2: andi $9,$12,8\n"         // Spin until it is OK to stop the MCU clocks.
		" bnez $9,2b\n"
		" lhu $12,%%lo(%1)($10)\n"
		" sh $13,%%lo(%1)($10)\n"    // Latch the DRAM control lines (DRAM_EN).
		" sh $14,%%lo(%1)($10)\n"    // End the self-refresh request (SUSPEND).
		" sh $15,%%lo(%1)($10)\n"    // Stop the MCU clocks.
		" nop\n"
		" nop\n"
		" nop\n"
		" hibernate\n"
		" nop\n"
		" nop\n"
		" nop\n"
		" .set pop\n"
		:
		:"i" (VR41XX_CMUCLKMSK),"i" (VR41XX_DRAMHIBCTL),
		 "i" (VR41XX_BCURFCNTREG),"i" (Fill)
		:"$8","$9","$10","$11","$12","$13","$14","$15"
	);
#else // not VR4181
#ifndef CONFIG_CPU_VR4120A
	asm volatile (
		" .set noreorder\n"
		" .align 4\n"
		" hibernate\n"
		" nop\n"
		" nop\n"
		" nop\n"
		" .set reorder\n"
	);
#endif
#endif
	// never reached...
}

#ifdef CONFIG_PM

#ifdef CONFIG_PM_POWERED_SUSPEND

#endif // CONFIG_PM_POWERED_SUSPEND

#ifdef CONFIG_PM_SUSPEND_WAKEUP

//
// On VR41xx CPUs, the hibernate instruction will cut power to most of
// the device.  On a wakeup event (usually power button), the CPU
// restarts at the reset vector in ROM, so the bootloader has to run
// again in order to pass control back to the kernel.  As a result,
// the bootloader (and/or some level of code in the ROM) must not
// disturb the contents of RAM before passing control to the kernel
// when waking from hibernate.
//
// This code uses the power management kernel interface to the
// peripheral drivers.  It is not real ACPI support, which would require
// some rather impractical (for this class of devices) hardware and
// firmware extensions.  Also, for now, the SUSPEND and RESUME functions
// in the peripheral drivers just need to do any necessary hardware
// state saving and restoring.  Power, IRQ mask, clock mask, and GPIO
// state are handled globally.  This will probably change in the future
// to allow for power-state control of individual peripheral units.
//
// Right now, the data arg passed to the pm_request callback for SUSPEND
// and RESUME means:
//   0: restore state and return to powered-on
//   1: save state and prepare for hibernate, no need to power off because
//      power will be removed globally
//   2: save state and prepare for suspend, global power will remain, so
//      the peripheral driver may want to power down its local power
//   3: prepare for suspend, peripherals should appear to user to be
//      running, most interrupts can wake from this state
// The numbers above should get replaced with an enum evetually, and
// the requirements for each mode will probably change, since the above
// doesn't really make sense (eg, no need to save state for 2, except
// that we need something to restore to for 0).  The pm_request callbacks
// currently ignore the data value anyway....
//
// This may seem like a bad idea to put the state information into the
// data segment where the bootloader may overwrite it, but if the
// bootloader rewrites data (and/or bss), it won't support wake from
// hibernate, anyway.
//
unsigned int powerevent_queued;
unsigned int hibernation_state = LOAD_MAGIC;
static unsigned short clkmsk_state;
extern void do_pm_irq_request(pm_request_t rqst);

#ifdef CONFIG_CPU_VR4181
#define MIN_GPIOREG VR41XX_GPMD0REG
#define MAX_GPIOREG VR41XX_LCDGPMODE
#else
#define MIN_GPIOREG VR41XX_GIUIOSELL
#define MAX_GPIOREG VR41XX_GIUINTHTSELH
#endif
#define NR_GPIOREGS (((long)MAX_GPIOREG - (long)MIN_GPIOREG) / sizeof(short) + 1)

unsigned short gpio_state[NR_GPIOREGS];

asmlinkage void do_hibernate(void *sp)
{
	int i;

	cli();
	powerevent_queued = 0;

	if (pm_send_all(PM_SUSPEND, (void *)1)) {
		sti();
		up(&vr41xx_dma_sem);
		return;
	}
	do_pm_irq_request(PM_SUSPEND);
	for (i = 0; i < NR_GPIOREGS; i++)
		gpio_state[i] = *(MIN_GPIOREG + i);
	clkmsk_state = *VR41XX_CMUCLKMSK;

	// we may want to put in some checksum or CRC on all of RAM, too

	hibernation_state = HIB_MAGIC;
	flush_cache_all();
	vr41xx_hibernate();
	// never reached...
}

asmlinkage int do_wakeup(void)
{
	int i, retval;

	// start out with something reasonable in CP0_STATUS, this will get
	// replaced when context is restored.  ints disbaled at this point
	write_32bit_cp0_register(CP0_STATUS, ST0_CU0);

	// Do what loadmmu would have done if cold init
	set_cp0_config(CONF_CM_CMASK, CONF_CM_CACHABLE_NONCOHERENT);
	flush_cache_all();
	write_32bit_cp0_register(CP0_WIRED, 0);
	write_32bit_cp0_register(CP0_PAGEMASK, PM_4K);
	flush_tlb_all();

	// this is about the point where we would check CRC if we did one

	*VR41XX_CMUCLKMSK = clkmsk_state;
	for (i = 0; i < NR_GPIOREGS; i++)
		*(MIN_GPIOREG + i) = gpio_state[i];
	do_pm_irq_request(PM_RESUME);
	retval = pm_send_all(PM_RESUME, (void *)0);
	if (!retval) {
		sti();
		up(&vr41xx_dma_sem);
	}
	return retval;
}

/*
 * Enter system sleep state (hibernate, really soft-off)
 *
 * Actually, all this does is to set a flag, so the tail end of the syscall
 * handler jumps to do_hibernate above, rather than executing
 * o32_ret_from_sys_call.  That just makes for a convenient place to restore
 * context from, since context is already saved on the stack at that point,
 * and bottom-half processing already done.  The wakeup part is done in
 * kernel_entry, since waking up from hibernate will restart at reset vector.
 */
static int pm_do_sleep(ctl_table *ctl,
			 int write,
			 struct file *file,
			 void *buffer,
			 size_t *len)
{
	if (!write) {
		if (file->f_pos) {
			*len = 0;
			return 0;
		}
	} else {
		if (file->f_flags & O_NONBLOCK) {
                	if (down_trylock(&vr41xx_dma_sem))
                	        return -EAGAIN;
	        } else {
	                if (down_interruptible(&vr41xx_dma_sem))
        	                return -ERESTARTSYS;
	        }

		powerevent_queued = 1;
	}
	file->f_pos += *len;
	return 0;
}

#endif // CONFIG_PM_SUSPEND_WAKEUP

#ifdef CONFIG_PM_POWERED_SUSPEND

static inline void vr41xx_suspend(void)
{
#ifdef CONFIG_CPU_VR4181
// I have tested suspend on the VR4181, and is only partially broken.
// It actually does put the VR4181 into suspend.

// Broken: this code does not put the DRAM into self-refresh.
// From what I can tell, in order to put the DRAM into self-refresh, DRAM_EN has to be
// used to hold the MCU clocks in a fixed state.

// Broken: occasionally, the system refuses to come out of suspend.
// Only RTC reset will revive the system.  A RESET does jump to the reset
// vector, but it appears that the DRAM is hung.

// I am wondering how reasonable it is to put the DRAM into self-refresh during
// suspend.  The VR4181 documentation is confusing.  It gives instructions on putting
// the DRAM into self-refresh before suspend, but then later it says not to
// fix the MCU clocks (DRAM_EN) before suspend.

// It appears that masking and restoring CMUCLKMSK is unnecessary.  Maybe it is just me,
// but the DRAM hang problem seems to occur less frequently without the CMUCLKMSK
// mask/restore.  Also, not masking CMUCLKMSK allows touch-panel wake-up.

// bdl, 5-Aug-2000

// DRAM wake-up code has to run after the suspend sequence returns, probably
// by the interrupt exception handler.  yuck.  also, the case where an int
// happens after DRAM is put to sleep but before the suspend instruction
// runs has to be handled.  double yuck.  or can ints be disabled during
// suspend on VR4181?
// I'm wondering if zeroing out CLKMSK is right, too.

	// unsigned short clkmsk_state = *VR41XX_CMUCLKMSK;
	// barrier();

	// same as hibernate, but execute suspend instead,
	// and no DRAM_EN tweaking
	asm volatile (
		" .set push\n"
		" .set noreorder\n"
		" .set mips3\n"
		" la $8,1f\n"                 // Set up some registers ahead of time.
		" lui $9,%%hi(%0)\n"
		" lui $10,%%hi(%1)\n"
		" li $11,0xffff\n"
		" li $12,2\n"
		" li $13,4\n"
		" cache %3,0($8)\n"           // Load the rest of the suspend code into the cache.
		" cache %3,16($8)\n"
		" cache %3,32($8)\n"
		" b 1f\n"
		" cache %3,48($8)\n"
		" .align 4\n"
		"1:\n"
		// " sh $0,%%lo(%0)($9)\n"       // Mask all the internal peripheral clocks (CMUCLKMSK).
		" sh $11,%%lo(%2)($9)\n"      // Max out the DRAM refresh time (BCURFCNTREG).
		" sh $12,%%lo(%1)($10)\n"     // Begin a self-refresh request (DRAMHIBCTL).
		"2: andi $9,$12,8\n"          // Spin until it is ok to stop the MCU clocks (DRAMHIBCTL).
		" bnez $9,2b\n"
		" lhu $12,%%lo(%1)($10)\n"
		" sh $0,%%lo(%1)($10)\n"      // End the self-refresh request (DRAMHIBCTL).
		" sh $13,%%lo(%1)($10)\n"     // Stop the MCU clocks (DRAMHIBCTL).
		" nop\n"
		" nop\n"
		" nop\n"
		" suspend\n"
		" nop\n"
		" nop\n"
		" nop\n"
		" .set pop\n"
		:
		:"i" (VR41XX_CMUCLKMSK),"i" (VR41XX_DRAMHIBCTL),
		 "i" (VR41XX_BCURFCNTREG),"i" (Fill)
		:"$8","$9","$10","$11","$12","$13"
	);
	// barrier();
	// *VR41XX_CMUCLKMSK = clkmsk_state;
#else // not VR4181
	asm volatile (
		" .set noreorder\n"
		" .align 4\n"
		" suspend\n"
		" nop\n"
		" nop\n"
		" nop\n"
		" .set reorder\n"
	);
#endif
}

/*
 * Enter system "suspend" state, CPU remains powered but most/all clocks off
 */
static int pm_do_suspend(ctl_table *ctl,
			 int write,
			 struct file *file,
			 void *buffer,
			 size_t *len)
{
	if (!write) {
		if (file->f_pos) {
			*len = 0;
			return 0;
		}
	} else {
		int retval;

		// Must not suspend while DMA in progress
		if (file->f_flags & O_NONBLOCK) {
                	if (down_trylock(&vr41xx_dma_sem))
                	        return -EAGAIN;
	        } else {
	                if (down_interruptible(&vr41xx_dma_sem))
        	                return -ERESTARTSYS;
	        }

		retval = pm_send_all(PM_SUSPEND, (void *)2);
		if (retval) {
			up(&vr41xx_dma_sem);
			return retval;
		}
		vr41xx_suspend();
		retval = pm_send_all(PM_RESUME, (void *)0);
		up(&vr41xx_dma_sem);
		if (retval)
			return retval;
	}
	file->f_pos += *len;
	return 0;
}

#endif // CONFIG_PM_POWERED_SUSPEND

#ifdef CONFIG_PM_STANDBY

static inline void vr41xx_standby(void)
{
	asm volatile (
		" .set noreorder\n"
		" .align 4\n"
		" standby\n"
		" nop\n"
		" nop\n"
		" nop\n"
		" .set reorder\n"
	);
}

/*
 * Enter system "standby" state, CPU remains powered, clocks on
 */
static int pm_do_standby(ctl_table *ctl,
			 int write,
			 struct file *file,
			 void *buffer,
			 size_t *len)
{
	if (!write) {
		if (file->f_pos) {
			*len = 0;
			return 0;
		}
	} else {
// MFK: replace with send PM_SUSPEND request to pm_time_request:
		disable_irq(VR41XX_IRQ_INT1);
		vr41xx_standby();
// MFK: replace with send PM_RESUME request to pm_time_request:
		enable_irq(VR41XX_IRQ_INT1);
	}
	file->f_pos += *len;
	return 0;
}

#endif // CONFIG_PM_STANDBY

/*
 * NOTE1: We're hijacking the ACPI binary ctl_name's for now.  Don't expect
 *        this to necessarily behave as ACPI does, some are blatently wrong.
 * NOTE2: Don't get used to the text names, either, they are subject to change
 *        if we ever come up with a real naming convention....
 * NOTE3: For that matter, don't get used to this being done via sysctl,
 *        that's subject to change, too.
 */
static struct ctl_table pm_table[] =
{
#ifdef CONFIG_PM_SUSPEND_WAKEUP
	{ACPI_SLEEP, "sleep", NULL, 0, 0600, NULL, &pm_do_sleep},
#endif
#ifdef CONFIG_PM_POWERED_SUSPEND
	{ACPI_S1_SLP_TYP, "suspend", NULL, 0, 0600, NULL, &pm_do_suspend},
#endif
#ifdef CONFIG_PM_STANDBY
	{ACPI_S0_SLP_TYP, "standby", NULL, 0, 0600, NULL, &pm_do_standby},
#endif
	{0}
};

static struct ctl_table pm_dir_table[] =
{
	{CTL_ACPI, "pm", NULL, 0, 0555, pm_table},
	{0}
};

/*
 * Initialize the power management user interface
 */
static int __init pm_init(void)
{
	register_sysctl_table(pm_dir_table, 1);
	return 0;
}

__initcall(pm_init);

#endif // CONFIG_PM
