/*
 * linux/arch/mips/vr41xx/vrc4172lcd.c
 *
 * Platform support for the machine which LCD control by VRC4172 PWM.
 *
 * Copyright (C) 2000 SATO Kazumi
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/kernel.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/vr41xx.h>
#include <asm/vrc4172.h>

#if defined(CONFIG_NEC_MOBILEGEAR2_R430) || \
    defined(CONFIG_NEC_MOBILEGEAR2_R530) || \
    defined(CONFIG_NEC_MOBILEGEAR2_R700) || \
    defined(CONFIG_NEC_MOBILEGEAR2_R730) || \
    defined(CONFIG_DOCOMO_SIGMARION)
#define LCDON   	VRC4172_PWM_LCD_EN
#define LCDON_PORT	VRC4172_PWM_LCDDUTYEN
#define LIGHTON 	VRC4172_PWM_LCD_EN
#define LIGHTON_PORT	VRC4172_PWM_LCDDUTYEN
#define MAXLIGHT	8
#define LIGHTVAL(n)	((n) >= MAXLIGHT? 0x3f: (n) == 7? 0x37: 0x11 + (n)*5)
#define DEFLIGHT	6
#define DEFCONTRAST	6
#else
#error no LCD PWM Infomation....
#endif

static int backlight, contrast, lcdpower;

void vrc4172pwm_init_backlight(void)
{
#ifdef STANDALONE
	backlight = DEFLIGHT;
	*VRC4172_PWM_LCDDUTY = LIGHTVAL(DEFLIGHT);
	*VRC4172_PWM_LCDFREQ = VRC4172_PWM_LCDFREQ_DEF;
#else /* STANDALONE */
	int i = 1;

	/* save boot time status (maybe set by Windows CE) */
	if (*LIGHTON_PORT&LIGHTON) {
		for (i = 1; i <= MAXLIGHT; i++) 
			if ((*VRC4172_PWM_LCDDUTY&VRC4172_PWM_LCDDUTY_MASK)
				 <= LIGHTVAL(i))
				break;
		if (i >= MAXLIGHT)
			backlight = MAXLIGHT;
		else	
			backlight = i;
	} else
		backlight = 0;
#endif /* STANDALONE */
}

int get_vrc4172pwm_backlight(void)
{
	return backlight;
}

int vrc4172pwm_backlight(int n)
{
	int flags;

	if (n > MAXLIGHT)
		backlight = MAXLIGHT;
	else
		backlight = n;
	if(n == 0) {
		save_and_cli(flags);
		*VRC4172_PWM_LCDDUTY = LIGHTVAL(1); /* XXX why need ? */
		*LIGHTON_PORT &= ~LIGHTON;
		restore_flags(flags);
		// Turn the backlight off.
	} else {
		save_and_cli(flags);
		*LIGHTON_PORT |= LIGHTON;
		*VRC4172_PWM_LCDDUTY = LIGHTVAL(n);
		restore_flags(flags);
	}
	return 0;
}

void vrc4172pwm_init_contrast(void)
{
#ifdef STANDALONE
	contrast = DEFCONTRAST;
#else /* STANDALONE */
	/* save boot time status (maybe set by Windows CE) */
	/* but we don't know current value methods, so set constant */
	contrast = DEFCONTRAST;
#endif /* STANDALONE */
}

int get_vrc4172pwm_contrast(void)
{
	return contrast;
}

int vrc4172pwm_contrast(int n)
{
	contrast = n;
	// Turn the lcd on and some contrast.
	return 0;
}

int get_vrc4172pwm_lcdpower(void)
{
	return lcdpower;
}

int vrc4172pwm_lcdpower(int n)
{
	int flags;

	lcdpower = n;
	if(n == 0) {
		save_and_cli(flags);
		*LCDON_PORT &=  ~LCDON;
		restore_flags(flags);
	} else {
		save_and_cli(flags);
		*LCDON_PORT |=  LCDON;
		restore_flags(flags);
	}
	return 0;
}

void vrc4172pwm_setup(void)
{
	lcdpower = 1; /* boot time always on */
	/* backlight & contrast inherit by WinCE */
	vrc4172pwm_init_backlight();
	vrc4172pwm_init_contrast();
	vrc4172pwm_lcdpower(lcdpower);
	vrc4172pwm_contrast(contrast);
	vrc4172pwm_backlight(backlight);
}
