/*****************************************************************************/

/*
 *      kmemmap.c  --  Kernel Memory Access routines.
 *
 *      Copyright (C) 2000
 *        Thomas Sailer (sailer@ife.ee.ethz.ch)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  $Id: kmemmap.c,v 1.2 2000/01/11 11:07:43 tom Exp $	 
 *
 * THIS IS NOT PORTABLE right now. It makes too many assumptions about
 * i386 page tables. It does not even work with PAE mode kernels!!
 *
 */

/*****************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define _XOPEN_SOURCE 500
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

/* ---------------------------------------------------------------------- */

static void *kernelmap = NULL;

struct vm_area_struct;
typedef struct {
	unsigned int pmd;
} pgd_t;

#define PAGE_OFFSET (0xc0000000)
#define KERNEL_SIZE (0x40000000)
#define PAGE_SIZE   4096

#define virt_to_phys(x)  ((unsigned long)(x)-PAGE_OFFSET)
#define phys_to_virt(x)  ((unsigned long)(x)+PAGE_OFFSET)

#define virt_to_bus(x)  virt_to_phys(x)
#define bus_to_virt(x)  phys_to_virt(x)

#define phys_to_map(x) ((typeof (x))((unsigned long)(x)+(unsigned long)kernelmap))
#define map_to_phys(x) ((typeof (x))((unsigned long)(x)-(unsigned long)kernelmap))

#define virt_to_map(x) ((typeof (x))(virt_to_phys(x)+(unsigned long)kernelmap))
#define map_to_virt(x) ((typeof (x))(phys_to_virt(x)-(unsigned long)kernelmap))

/* ---------------------------------------------------------------------- */

struct mm_struct {
        struct vm_area_struct * mmap;           /* list of VMAs */
        struct vm_area_struct * mmap_avl;       /* tree of VMAs */
        struct vm_area_struct * mmap_cache;     /* last find_vma result */
        pgd_t * pgd;
	/* and so on... */
};

static void init(void)
{
	static struct symtab {
		const char *name;
		unsigned long val;
	} stab[] = {
		{ "jiffies", 0 },
		{ "high_memory", 0 },
		{ "init_mm", 0 },
		{ NULL, 0 }
	};
	struct symtab *st;
	char line[128];
	FILE *f;
	char *p;
	unsigned long sym = 0;
	unsigned int slen;
	unsigned long jiffies, memsize;
	void *high_memory;
	unsigned int pg[1024], pt[1024];
	unsigned int *pgd, *pte, pfa, pta, addr;
	struct mm_struct *init_mm;
	struct uhci *devs;
	int fd;

	if (!(f = fopen("/proc/ksyms", "r"))) {
		fprintf(stderr, "cannot open /proc/ksyms\n");
		exit(1);
	}
	while (fgets(line, sizeof(line), f)) {
		sym = strtoul(line, &p, 16);
		if (!p || p == line || p[0] != ' ')
			continue;
		for (st = stab; st->name; st++) {
			slen = strlen(st->name);
			if (strncmp(p+1, st->name, slen))
				continue;
			if (p[1+slen] != 0 && p[1+slen] != ' ' && p[1+slen] != '\t' && (p[1+slen] != '_' || p[2+slen] != 'R'))
				continue;
			if (st->val) {
				fprintf(stderr, "WARNING: Symbol %s found more than once, previous 0x%x, current 0x%x\n", 
					st->name, st->val, sym);
				continue;
			}
			st->val = sym;
			printf("Symbol %s = 0x%x found\n", st->name, st->val);
			break;
		}
	}
	fclose(f);
	for (st = stab; st->name; st++) {
		if (!st->val) {
			fprintf(stderr, "symbol \"%s\" not found\n", st->name);
			exit(1);
		}
	}
	fd = open("/dev/mem", O_RDONLY);
	if (fd == -1) {
		perror("open: /dev/mem");
		exit(1);
	}
	/* read high_memory pointer and init_mm->pgd */
	if (pread(fd, &high_memory, sizeof(high_memory), virt_to_phys(stab[1].val)) != sizeof(high_memory)) {
		perror("pread");
		exit(1);
	}
	if (pread(fd, &pgd, sizeof(pgd), virt_to_phys(stab[2].val+(unsigned long)&((struct mm_struct *)0)->pgd)) != sizeof(pgd)) {
		perror("pread");
		exit(1);
	}
	if ((unsigned long)high_memory & (PAGE_SIZE-1)) {
		fprintf(stderr, "high_memory (%p) not page aligned!\n", high_memory);
		exit(1);
	}
	memsize = virt_to_phys(high_memory);
	printf("Your kernel has detected %uM %uk Bytes of memory\n", memsize >> 20, (memsize >> 10) & 0x3ff);
	/* read page directory */
	if (pread(fd, &pg, sizeof(pg), virt_to_phys(pgd)) != sizeof(pg)) {
		perror("pread");
		exit(1);
	}
	/* get kernelmap address */
	if ((kernelmap = mmap(NULL, KERNEL_SIZE, PROT_READ, MAP_SHARED, fd, 0)) == MAP_FAILED) {
		perror("mmap: /dev/mem");
		exit(1);
	}
	munmap(kernelmap, KERNEL_SIZE);
	/* now walk the page directory/tables */
	for (pfa = PAGE_OFFSET; pfa != 0; pfa += (1 << 22)) {
		addr = pg[pfa >> 22];
		/* check for page not present */
		if (!(addr & 1))
			continue;
		/* check for 4M page */
		if (addr & 0x80) {
			printf("Mapping 4M virt 0x%08x phys 0x%08x\n", pfa, addr);
			if (mmap((void *)virt_to_map(pfa), (1<<22), PROT_READ, MAP_SHARED | MAP_FIXED, fd, addr & ~(PAGE_SIZE-1)) == MAP_FAILED) {
				perror("mmap: /dev/mem");
				exit(1);
			}
			continue;
		}
		/* read page table */
		if (pread(fd, &pt, sizeof(pg), addr & ~(PAGE_SIZE-1)) != sizeof(pt)) {
			perror("pread");
			exit(1);
		}
		for (pta = 0; pta < (1<<22); pta += PAGE_SIZE) {
			addr = pt[pta >> 12];
			/* check for page not present */
			if (!(addr & 1))
				continue;
			printf("Mapping 4k virt 0x%08x phys 0x%08x\n", pfa | pta, addr);
			if (mmap((void *)virt_to_map(pfa | pta), PAGE_SIZE, PROT_READ, MAP_SHARED | MAP_FIXED, fd, addr & ~(PAGE_SIZE-1)) == MAP_FAILED) {
				perror("mmap: /dev/mem");
				exit(1);
			}
		}
	}
	close(fd);
	printf("kernelmap %p jiffies_mapped %p\n", kernelmap, virt_to_map((unsigned long *)stab[1].val));
	jiffies = *virt_to_map((unsigned long *)stab[0].val);
	printf("Dump at time (jiffies) %lu = 0x%x\n", jiffies, jiffies);
	devs = *virt_to_map((struct uhci **)stab[3].val);
	uhci_devs = virt_to_map(devs);
}

