/* $Id: vr4120A.h,v 1.1 2001/07/18 12:48:37 jrathert Exp $
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1999 by Michael Klar
 * Copyright (C) 2000 by Michael R. McDonald
 */
#ifndef __ASM_MIPS_VR4122_H 
#define __ASM_MIPS_VR4122_H 

#include <asm/addrspace.h>

/* CPU interrupts */
#define VR41XX_IRQ_SW1       0  /* IP0 - Software interrupt */
#define VR41XX_IRQ_SW2       1  /* IP1 - Software interrupt */
#define VR41XX_IRQ_INT0      2  /* IP2 - All but battery, high speed modem, and real time clock */
#define VR41XX_IRQ_INT1      3  /* IP3 - RTC Long1 (system timer) */
#define VR41XX_IRQ_INT2      4  /* IP4 - RTC Long2 */
#define VR41XX_IRQ_INT3      5  /* IP5 - High Speed Modem */
#define VR41XX_IRQ_INT4      6  /* IP6 - Unused */
#define VR41XX_IRQ_TIMER     7  /* IP7 - Timer interrupt from CPO_COMPARE (Note: RTC Long1 is the system timer.) */

/* Cascaded from VR41XX_IRQ_INT0 (ICU mapped interrupts) */
#define VR41XX_IRQ_ZTMR0     44
#define VR41XX_IRQ_ZTMR1     45
#define VR41XX_IRQ_ZRTC      46
#define VR41XX_IRQ_SIU	    56 
#define VR41XX_IRQ_SIU1	    57 

#define VR41XX_IRQ_MAX       71

#ifndef _LANGUAGE_ASSEMBLY
#define __preg8		(volatile unsigned char*)
#define __preg16	(volatile unsigned short*)
#define __preg32	(volatile unsigned int*)
#else
#define __preg8
#define __preg16
#define __preg32
#endif

/* Embedded CPU peripheral registers */

/* Interrupt Control Unit (ICU) */
#define VR41XX_SYSINT1REG    __preg32(KSEG1 + 0x10000030)
#define VR41XX_SYSINT2REG    __preg32(KSEG1 + 0x10000034)
#define VR41XX_MSYSINT1REG   __preg32(KSEG1 + 0x10000040)
#define VR41XX_MSYSINT2REG   __preg32(KSEG1 + 0x10000044)
#define VR41XX_BHIF_NMI_ST   __preg32(KSEG1 + 0x10000050)
#define VR41XX_BHIF_NMI_EN   __preg32(KSEG1 + 0x10000060)
#if 0
#define VR41XX_GIUINTLREG	__preg16(KSEG1 + 0x0F000088)	/* Level 2 GIU interrupt register Low */
#define VR41XX_DSIUINTREG	__preg16(KSEG1 + 0x0F00008A)	/* Level 2 DSIU interrupt register */
#define VR41XX_MGIUINTLREG	__preg16(KSEG1 + 0x0F000094)	/* Level 2 mask GIU interrupt register Low */
#define VR41XX_MDSIUINTREG	__preg16(KSEG1 + 0x0F000096)	/* Level 2 mask DSIU interrupt register */
#define VR41XX_NMIREG		__preg16(KSEG1 + 0x0F000098)	/* NMI register */
#define VR41XX_SOFTINTREG	__preg16(KSEG1 + 0x0F00009A)	/* Software interrupt register */
#define VR41XX_GIUINTHREG	__preg16(KSEG1 + 0x0F0000A2)	/* Level 2 GIU interrupt register High */
#define VR41XX_FIRINTREG	__preg16(KSEG1 + 0x0F0000A4)	/* Level 2 FIR interrupt register */
#define VR41XX_MGIUINTHREG	__preg16(KSEG1 + 0x0F0000A8)	/* Level 2 mask GIU interrupt register High */
#define VR41XX_MFIRINTREG	__preg16(KSEG1 + 0x0F0000AA)	/* Level 2 mask FIR interrupt register */
#define VR41XX_PCIINTREG	__preg16(KSEG1 + 0x0F0000AC)	/* Level 2 PCI interrupt register */
#define VR41XX_SCUINTREG	__preg16(KSEG1 + 0x0F0000AE)	/* Level 2 SCU interrupt register */
#define VR41XX_CSIINTREG	__preg16(KSEG1 + 0x0F0000B0)	/* Level 2 CSI interrupt register */
#define VR41XX_MPCIINTREG	__preg16(KSEG1 + 0x0F0000B2)	/* Level 2 mask PCI interrupt register */
#define VR41XX_MSCUINTREG	__preg16(KSEG1 + 0x0F0000B4)	/* Level 2 mask SCU interrupt register */
#define VR41XX_MCSIINTREG	__preg16(KSEG1 + 0x0F0000B6)	/* Level 2 mask CSI interrupt register */
#endif

/* Real Time Clock Unit (RTC) */

#define VR41XX_ZTMR0_TIMER_COMMAND __preg32(KSEG1 + 0x12010200)
#define VR41XX_ZTMR0_TIMER_MODE    __preg32(KSEG1 + 0x12010204)
#define VR41XX_ZTMR0_TIMER_VALUE   __preg32(KSEG1 + 0x12010208)
#define VR41XX_ZTMR0_INT_STATUS    __preg32(KSEG1 + 0x12010210)
#define VR41XX_ZTMR0_EVENT_ACTION  __preg32(KSEG1 + 0x12010214)
#define VR41XX_ZTMR0_COMPARE       __preg32(KSEG1 + 0x12010218)

/* General Purpose I/O Unit (GIU) */
#define VR41XX_GIUIOSELL	__preg16(KSEG1 + 0x0F000140)	/* GPIO Input/Output Select Register L */
#define VR41XX_GIUIOSELH	__preg16(KSEG1 + 0x0F000142)	/* GPIO Input/Output Select Register H */
#define VR41XX_GIUPIODL		__preg16(KSEG1 + 0x0F000144)	/* GPIO Port Input/Output Data Register L */
#define VR41XX_GIUPIODL_GPIO15  0x8000
#define VR41XX_GIUPIODL_GPIO14  0x4000
#define VR41XX_GIUPIODL_GPIO13  0x2000
#define VR41XX_GIUPIODL_GPIO12  0x1000
#define VR41XX_GIUPIODL_GPIO11  0x0800
#define VR41XX_GIUPIODL_GPIO10  0x0400
#define VR41XX_GIUPIODL_GPIO9  0x0200
#define VR41XX_GIUPIODL_GPIO8  0x0100
#define VR41XX_GIUPIODL_GPIO7  0x0080
#define VR41XX_GIUPIODL_GPIO6  0x0040
#define VR41XX_GIUPIODL_GPIO5  0x0020
#define VR41XX_GIUPIODL_GPIO4  0x0010
#define VR41XX_GIUPIODL_GPIO3  0x0008
#define VR41XX_GIUPIODL_GPIO2  0x0004
#define VR41XX_GIUPIODL_GPIO1  0x0002
#define VR41XX_GIUPIODL_GPIO0  0x0001
#define VR41XX_GIUPIODH		__preg16(KSEG1 + 0x0F000146)	/* GPIO Port Input/Output Data Register H */
#define VR41XX_GIUPIODH_GPIO31  0x8000
#define VR41XX_GIUPIODH_GPIO30  0x4000
#define VR41XX_GIUPIODH_GPIO29  0x2000
#define VR41XX_GIUPIODH_GPIO28  0x1000
#define VR41XX_GIUPIODH_GPIO27  0x0800
#define VR41XX_GIUPIODH_GPIO26  0x0400
#define VR41XX_GIUPIODH_GPIO25  0x0200
#define VR41XX_GIUPIODH_GPIO24  0x0100
#define VR41XX_GIUPIODH_GPIO23  0x0080
#define VR41XX_GIUPIODH_GPIO22  0x0040
#define VR41XX_GIUPIODH_GPIO21  0x0020
#define VR41XX_GIUPIODH_GPIO20  0x0010
#define VR41XX_GIUPIODH_GPIO19  0x0008
#define VR41XX_GIUPIODH_GPIO18  0x0004
#define VR41XX_GIUPIODH_GPIO17  0x0002
#define VR41XX_GIUPIODH_GPIO16  0x0001
#define VR41XX_GIUINTSTATL	__preg16(KSEG1 + 0x0F000148)	/* GPIO Interrupt Status Register L */
#define VR41XX_GIUINTSTATH	__preg16(KSEG1 + 0x0F00014A)	/* GPIO Interrupt Status Register H */
#define VR41XX_GIUINTENL	__preg16(KSEG1 + 0x0F00014C)	/* GPIO Interrupt Enable Register L */
#define VR41XX_GIUINTENH	__preg16(KSEG1 + 0x0F00014E)	/* GPIO Interrupt Enable Register H */
#define VR41XX_GIUINTTYPL	__preg16(KSEG1 + 0x0F000150)	/* GPIO Interrupt Type (Edge or Level) Select Register */
#define VR41XX_GIUINTTYPH	__preg16(KSEG1 + 0x0F000152)	/* GPIO Interrupt Type (Edge or Level) Select Register */
#define VR41XX_GIUINTALSELL	__preg16(KSEG1 + 0x0F000154)	/* GPIO Interrupt Active Level Select Register L */
#define VR41XX_GIUINTALSELH	__preg16(KSEG1 + 0x0F000156)	/* GPIO Interrupt Active Level Select Register H */
#define VR41XX_GIUINTHTSELL	__preg16(KSEG1 + 0x0F000158)	/* GPIO Interrupt Hold/Through Select Register L */
#define VR41XX_GIUINTHTSELH	__preg16(KSEG1 + 0x0F00015A)	/* GPIO Interrupt Hold/Through Select Register H */

#define VR41XX_GIUPODATEN	__preg16(KSEG1 + 0x0F00015C)	/* GPIO Port Output Data Enable Register */
#define VR41XX_GIUPODATL	__preg16(KSEG1 + 0x0F00015E)	/* GPIO Port Output Data Register L */
#define VR41XX_GIUPODATL_GPIO47  0x8000
#define VR41XX_GIUPODATL_GPIO46  0x4000
#define VR41XX_GIUPODATL_GPIO45  0x2000
#define VR41XX_GIUPODATL_GPIO44  0x1000
#define VR41XX_GIUPODATL_GPIO43  0x0800
#define VR41XX_GIUPODATL_GPIO42  0x0400
#define VR41XX_GIUPODATL_GPIO41  0x0200
#define VR41XX_GIUPODATL_GPIO40  0x0100
#define VR41XX_GIUPODATL_GPIO39  0x0080
#define VR41XX_GIUPODATL_GPIO38  0x0040
#define VR41XX_GIUPODATL_GPIO37  0x0020
#define VR41XX_GIUPODATL_GPIO36  0x0010
#define VR41XX_GIUPODATL_GPIO35  0x0008
#define VR41XX_GIUPODATL_GPIO34  0x0004
#define VR41XX_GIUPODATL_GPIO33  0x0002
#define VR41XX_GIUPODATL_GPIO32  0x0001
#define VR41XX_GIUPODATL_PODAT15  0x8000
#define VR41XX_GIUPODATL_PODAT14  0x4000
#define VR41XX_GIUPODATL_PODAT13  0x2000
#define VR41XX_GIUPODATL_PODAT12  0x1000
#define VR41XX_GIUPODATL_PODAT11  0x0800
#define VR41XX_GIUPODATL_PODAT10  0x0400
#define VR41XX_GIUPODATL_PODAT9  0x0200
#define VR41XX_GIUPODATL_PODAT8  0x0100
#define VR41XX_GIUPODATL_PODAT7  0x0080
#define VR41XX_GIUPODATL_PODAT6  0x0040
#define VR41XX_GIUPODATL_PODAT5  0x0020
#define VR41XX_GIUPODATL_PODAT4  0x0010
#define VR41XX_GIUPODATL_PODAT3  0x0008
#define VR41XX_GIUPODATL_PODAT2  0x0004
#define VR41XX_GIUPODATL_PODAT1  0x0002
#define VR41XX_GIUPODATL_PODAT0  0x0001
#define VR41XX_SECIRQMASKL	VR41XX_GIUINTENL
#define VR41XX_SECIRQMASKH	VR41XX_GIUINTENH

/* Serial Interface Unit (SIU / SIU1 and SIU2) */
#define VR41XX_SIURB		__preg32(KSEG1 + 0x12001000)	/* Receiver Buffer Register (Read) DLAB = 0 */
#define VR41XX_SIUTH		__preg32(KSEG1 + 0x12001000)	/* Transmitter Holding Register (Write) DLAB = 0 */
#define VR41XX_SIUDLL		__preg32(KSEG1 + 0x12001000)	/* Divisor Latch (Least Significant Byte) DLAB = 1 */
#define VR41XX_SIUIE		__preg32(KSEG1 + 0x12001010)	/* Interrupt Enable DLAB = 0 */
#define VR41XX_SIUDLM		__preg32(KSEG1 + 0x12001010)	/* Divisor Latch (Most Significant Byte) DLAB = 1 */
#define VR41XX_SIUIID		__preg32(KSEG1 + 0x12001020)	/* Interrupt Identification Register (Read) */
#define VR41XX_SIUFC		__preg32(KSEG1 + 0x12001020)	/* FIFO Control Register (Write) */
#define VR41XX_SIULC		__preg32(KSEG1 + 0x12001030)	/* Line Control Register */
#define VR41XX_SIUMC		__preg32(KSEG1 + 0x12001030)	/* MODEM Control Register */
#define VR41XX_SIULS		__preg32(KSEG1 + 0x12001050)	/* Line Status Register */
#define VR41XX_SIUMS		__preg32(KSEG1 + 0x12001060)	/* MODEM Status Register */
#define VR41XX_SIUSC		__preg32(KSEG1 + 0x12001070)	/* Scratch Register */
#if 0
#define VR41XX_SIUIRSEL		__preg32(KSEG1 + 0x12001080)	/* SIU/FIR IrDA Selector */
#define VR41XX_SIURESET		__preg32(KSEG1 + 0x12001090)	/* SIU Reset Register */
#define VR41XX_SIUCSEL		__preg32(KSEG1 + 0x120010A0)	/* SIU Echo-Back Control Register */
#endif
#define VR41XX_SIURB1		__preg32(KSEG1 + 0x12002000)	/* Receiver Buffer Register (Read) DLAB = 0 */

#ifdef CONFIG_NEC_UPD61130X
#define VR41XX_PORT_BASE	(KSEG1 + 0x10010000 )   // for ide interface access
#define VR41XX_ISAMEM_BASE	(KSEG1 + 0x10000000 )  // not used
#endif

#ifdef CONFIG_NEC_HARRIER

/* physical address spaces */
#define VR41XX_LCD             0x0a000000
#define VR41XX_INTERNAL_IO_2   0x0b000000
#define VR41XX_INTERNAL_IO_1   0x0c000000
#define VR41XX_ISA_MEM         0x10000000
#define VR41XX_ISA_IO          0x16000000
#define VR41XX_ROM             0x18000000

#define NEC_HARRIER_SIO1	__preg16(KSEG1 + 0x0A000000)
#define NEC_HARRIER_SIO2	__preg16(KSEG1 + 0x0A000010)

#define NEC_HARRIER_FLSHCNTREG	__preg8(KSEG1 + 0x0DFFFFA0)
#define NEC_HARRIER_FLSHBANKREG	__preg8(KSEG1 + 0x0DFFFFA4)
#define NEC_HARRIER_SWSETREG	__preg8(KSEG1 + 0x0DFFFFA8)
#define NEC_HARRIER_LED1REG	__preg8(KSEG1 + 0x0DFFFFC0)
#define NEC_HARRIER_LED2REG	__preg8(KSEG1 + 0x0DFFFFC4)
#define NEC_HARRIER_SDBINTREG	__preg8(KSEG1 + 0x0DFFFFD0)
#define NEC_HARRIER_SDBINTMASK	__preg8(KSEG1 + 0x0DFFFFD4)
#define NEC_HARRIER_RSTREG	__preg8(KSEG1 + 0x0DFFFFD8)
#define NEC_HARRIER_PCIINTREG	__preg8(KSEG1 + 0x0DFFFFDC)
#define NEC_HARRIER_PCIINTMASK	__preg8(KSEG1 + 0x0DFFFFE0)
#define NEC_HARRIER_PCICLKREG	__preg32(KSEG1 + 0x0DFFFFE4)

#define NEC_HARRIER_DISPLAY_LEDS(x) \
    do { \
	*NEC_HARRIER_LED1REG = (unsigned char) ((x)       & 0xff); \
        *NEC_HARRIER_LED2REG = (unsigned char)(((x) >> 8) & 0xff); \
    } while (0)
 

/* This is the base address for IO port decoding to which the 16 bit IO port address */
/* is added.  Defining it to 0 will usually cause a kernel oops any time port IO is */
/* attempted, which can be handy for turning up parts of the kernel that make */
/* incorrect architecture assumptions (by assuming that everything acts like a PC), */
/* but we need it correctly defined to use the PCMCIA/CF controller: */
#define VR41XX_PORT_BASE	(KSEG1 + VR41XX_ISA_IO)
#define VR41XX_ISAMEM_BASE	(KSEG1 + VR41XX_ISA_MEM)

#else  /* CONFIG_NEC_HARRIER */

#define NEC_HARRIER_DISPLAY_LEDS(x) {}

#endif /* CONFIG_NEC_HARRIER */

#endif /* __ASM_MIPS_VR4122_H */
